/* Seamless3d

Copyright (C) 2001-2011 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include "seamless3d.h"
#include "script.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"

extern M_<Ptr<Function> > g_functionList;
void resetDefList();
extern bool g_forceSaveAs;
void changeExtensionToSmls(String& str);
void makeExistingNameUniqueUsingDefList(Node* node);

void addOIAssignment(Function* f, OrientationInterpolator* oi, Art* part);
void addOIAssignment(Function* f, PositionInterpolator* pi, Art* part);
void commonEndForSetupAnim();
void insertRecentFile(const char* fullFileName);
void render();
Rotation convRot4dto4f(const Rotation4d& par);
void createDefList(M_Node& defLookUp);

static const double toInches = 39.37;
static const double toMeters = 1/toInches;


static const float geomRad = .05*.5;
extern Ptr<AnimControlPanel> g_acp;;
int piCount = 0;

Art* addPart(M_SArt& m_part, Art* part, String& str, int& i){
	part->name = str.extractName(i);
	makeExistingNameUniqueUsingDefList(part);
	m_part+=part;
	str.stb(i);
	return part;
}

int addChannelRotData(String& str, int& i){
	int valOrder = 0;
	for(int j = 3; --j>=0;){
		if(str.cmpKWord("Xrotation", i)){
			valOrder<<=8;
		}
		else if(str.cmpKWord("Yrotation", i)){
			valOrder<<=8;
			valOrder|=1;
		}
		else if(str.cmpKWord("Zrotation", i)){
			valOrder<<=8;
			valOrder|=2;
		}
		str.stb(i);
	}
	return valOrder;
}

int addChannelPosData(String& str, int& i){
	int valOrder = 0;
	for(int j = 3; --j>=0;){
		if(str.cmpKWord("Xposition", i)){
			valOrder<<=8;
		}
		else if(str.cmpKWord("Yposition", i)){
			valOrder<<=8;
			valOrder|=1;
		}
		else if(str.cmpKWord("Zposition", i)){
			valOrder<<=8;
			valOrder|=2;
		}
		str.stb(i);
	}
	return valOrder;
}

void addChannelData(String& str, int& i, Art* part, M_Node& interpolator, Function* f, M_Int& valOrder){
	int remI = i;
	i++;
	if(str.cmpKWord("rotation", i)){
		i = remI;
		valOrder+=addChannelRotData(str, i);
		OrientationInterpolator* oi = new OrientationInterpolator;
		oi->name = "oi_"+part->name;
		interpolator+=oi;

		addOIAssignment(f, oi, part);
	}
	else if(str.cmpKWord("position", i)){
		i = remI;
		valOrder+=addChannelPosData(str, i);
		PositionInterpolator* pi = new PositionInterpolator;
		pi->name = "pi_"+part->name;
		interpolator+=pi;

		if(piCount<1){
			addOIAssignment(f, pi, part);
			piCount++;
		}
	}
	else throw;
}

NurbsLathe* addNewLathe(Art* part, const Vec3f accumOffset){
	//if(child->pivotPoint.refValue().length() < .0001){
	//	return 0;
	//}
	if(part->show == false || part->name.length == 0){
		return 0;
	}
	NurbsLathe* nurbsLathe = new NurbsLathe;
	nurbsLathe->part.refValue() = part;
	nurbsLathe->weight.refValue().length = 7;
	nurbsLathe->knot.refValue() = M_Float(0, 0, 0, .333333, .333333, .666666, .666666, 1, 1, 1);
	nurbsLathe->translation.refValue() = accumOffset;
	M_Vec2f& point = nurbsLathe->point.refValue();
	point.length = 7;
	
	nurbsLathe->processFields();


	//part->pivotPoint.refValue() = fromOffset;
	return nurbsLathe;
}

void getOffset(String& str, int& i, Art* part, Art* child, M_Build& geomGroup, const Vec3f fromOffset, Vec3f& toOffset, const Vec3f accumOffset){

	NurbsLathe* nurbsLathe = addNewLathe(part, accumOffset);
	part->pivotPoint.refValue() = fromOffset;
	

	if(child){
		addPart(part->part.refValue(), child, str, i);
	}

	str.stb(i);
	if(str[i++]=='{'){
		str.stb(i);
		if(str.cmpKWord("OFFSET", i)){
			toOffset.getField(str, i);
			toOffset*=toMeters;

			if(child){
				child->pivotPoint.refValue() = toOffset;
			}

			float diff = toOffset.length();
			if(nurbsLathe){
				geomGroup += nurbsLathe;
				nurbsLathe->rotation.refValue() = Rotation(Vec3f(0, 1, 0), toOffset);
				M_Vec2f& point = nurbsLathe->point.refValue();

				point[0] = Vec2f();
				point[1] = Vec2f(geomRad, 0);
				point[6] = Vec2f(0, diff);
				point[5] = Vec2f(geomRad, diff);
				point[3] = (point[1] + point[5]) / 2;
			}
		}
	}
}

void recurseBVH(String& str, int& i, Art* part, M_Build& geomGroup, Vec3f fromOffset, Vec3f accumOffset, M_Node& interpolator, Function* f, M_Int& valOrder){
	Vec3f toOffset;

	str.stb(i);
	while(str[i]!='}'){
		if(str.cmpKWord("CHANNELS", i)){
			str.stb(i);
			int channelLen = getInt(str, i);
			str.stb(i);
			addChannelData(str, i, part, interpolator, f, valOrder);
			str.stb(i);
			if(channelLen == 6){
				addChannelData(str, i, part, interpolator, f, valOrder);
			}
		}
		else if(str.cmpKWord("JOINT", i)){
			Art* child = new Part;
			getOffset(str, i, part, child, geomGroup, fromOffset, toOffset, accumOffset);
			recurseBVH(str, i, child, geomGroup, toOffset, accumOffset+toOffset, interpolator, f, valOrder);
		}
		else if(str.cmpKWord("End Site", i)){
			getOffset(str, i, part, 0, geomGroup, fromOffset, toOffset, accumOffset);
			str.stb(i);
			if(str[i++]!='}'){
				throw;
			}
		}
		else throw;
		str.stb(i);
	}
	i++;
}

Rotation4d getRot(String& str, int& i, int valOrder){
	valOrder &= 0x0ff;
	double v = radians(getFloat2(str, i));
	Rotation rot;
	switch(valOrder){
		case 0:{
			return Rotation4d(1, 0, 0 ,v);
		}
		case 1:{
			return Rotation4d(0, 1, 0 ,v);
		}
		case 2:{
			return Rotation4d(0, 0, 1 ,v);
		}
	}
	throw;
}

void getPos(String& str, int& i, int valOrder, Vec3f& position){
	valOrder &= 0x0ff;
	float v = getFloat2(str, i);
	v*=toMeters;
	switch(valOrder){
		case 0:{
			position.x = v;
		}
		case 1:{
			position.y = v;
		}
		case 2:{
			position.z = v;
		}
	}
	return;
}

void getMotionData2(String& str, int& i, M_Node& interpolator, const int numOfFrames, const double period, M_Int& valOrder){
	for(int frame = 0; frame < numOfFrames; frame++){
		for(int k = 0; k < interpolator.length; k++){
			str.stb(i);
			if(OrientationInterpolator* oi = dynamic_cast<OrientationInterpolator*>(interpolator[k].p)){

				Rotation4d rotC = getRot(str, i, valOrder[k]>>16);
				Rotation4d rotB = getRot(str, i, valOrder[k]>>8);
				Rotation4d rotA = getRot(str, i, valOrder[k]);

				Rotation4d rot = rotA * rotB * rotC;

				oi->keyValue.refValue() += convRot4dto4f(rot);
				oi->key.refValue()+=static_cast<double>(frame) / (numOfFrames - 1);
			}
			else if(PositionInterpolator* pi = dynamic_cast<PositionInterpolator*>(interpolator[k].p)){
				Vec3f position;
				getPos(str, i, valOrder[k]>>16, position);
				getPos(str, i, valOrder[k]>>8, position);
				getPos(str, i, valOrder[k], position);

				pi->keyValue.refValue()+=position;
				pi->key.refValue()+=static_cast<double>(frame) / (numOfFrames - 1);
			}
		}
	}
}
void getMotionData(String& str, int& i, M_Node& interpolator, Anim* anim, M_Int& valOrder){
	
	str.stb(i);

	if(str.cmpKWord("MOTION", i)){
		str.stb(i);
		if(str.cmpKWord("Frames:", i)){
			str.stb(i);
			int numOfFrames = getInt(str, i);
			str.stb(i);
			if(str.cmpKWord("Frame Time:", i)){
				str.stb(i);
				double frameTime = getFloat2(str, i);
				int framesPerSecond = 1.0/frameTime + .5;
				anim->stepsPerSecond = framesPerSecond;
				double period = frameTime * (numOfFrames - 1);
				period = quantize(period, frameTime);
				anim->period = period;
				getMotionData2(str, i, interpolator, numOfFrames, period, valOrder);
			}
		}
	}
}

void getRoots(String& str, int& i, Seamless* smls, GroupBuild* groupBuild, Function* function, M_Int& valOrder, M_Node& interpolator){

	while(str.cmpKWord("ROOT", i)){

		Art* part = new Part;
		addPart(smls->skeleton.refValue(), part, str, i);

		if(str[i++]=='{'){
			str.stb(i);
			if(str.cmpKWord("OFFSET", i)){
				Vec3f offset;
				offset.getField(str, i);
				offset*=toMeters;
				part->pivotPoint.refValue() = offset;

				recurseBVH(str, i, part, groupBuild->build.refValue(), offset, offset, interpolator, function, valOrder);
			}
		}
		str.stb(i);
	}
}

bool loadBVH(const String& filename, ScenePtr& importScene, M_<SceneTreeListItem>& importSceneTreeList, M_Node& importDefList){
	setMathHiPresision();
	Function::reset_m_word=false;
	String& str=g_loadStr;
	str = readFile(filename);
	if(str=="")return false;

	piCount=0;

	int i = 0;
	str.stb(i);
	if(str.cmpKWord("HIERARCHY", i)){
		setMathHiPresision();
		importScene = new Scene;
		Function::selectedLine=0;//create new code lines


		str.stb(i);
		//changeExtensionToSmls(g_currentFilename);
		g_forceSaveAs = true;
		//insertRecentFile(filename.c_str());

		M_Node& children = importScene->children.refValue();
		children.length=0;
		importDefList.length=0;
		Seamless* smls = new Seamless;
		GroupBuild* groupBuild = new GroupBuild;
		smls->build.refValue() += groupBuild;

		smls->effect = new ColorEffect;
		children += smls;
		Anim* anim = new Anim;
		children += anim;

		Function* function = new MemberFunction;
		anim->onFrame=function;
		g_functionList.length=0;
		g_functionList+=function;
		M_Int valOrder;
		M_Node& interpolator = anim->children.refValue();

		getRoots(str, i, smls, groupBuild, function, valOrder, interpolator);

		getMotionData(str, i, interpolator, anim, valOrder);

		{
			int dst = 0;
			piCount = 0;
			for(int src = 0; src < interpolator.length; src++){
				if(dynamic_cast<PositionInterpolator*>(interpolator[src].p)){
					if(piCount>0)continue;
					piCount++;
				}
				interpolator[dst++] = interpolator[src];
			}
			interpolator.length = dst;
		}

		g_acp->setupNodeWindow(anim);

		importScene->wOpen.refValue().length=0;
		importScene->wPos.refValue().length=0;
		importScene->wSize.refValue().length=0;
		g_colorOwnership = false;
		g_paletteMode = false;
		return true;
	}
	return false;
}

void openBVH(const String& filename){

	if(loadBVH(filename, g_scene, g_sceneTreeList, g_defLookUp)){
		g_stw->selLine=0;
		g_scene->genNodeList2();
		postLoadProcessNodes(g_sceneTreeList);

		resetLooseEnds();
		commonEndForSetupAnim();

		Window::postLoadProcessWin(*g_scene);

		BackUp::deleteBackupFiles();
		BackUp::createUniqueBackupName();

		g_stw->paint();
		Window::activeWindow=g_acp.p;
		Window::paintAllControlPanels();
		setSmlsSaved();
		setTitleName();

		restoreMathPresision();
	}
}

bool importBVHGesture(const String& filename, ScenePtr& importScene, M_<SceneTreeListItem>& importSceneTreeList, M_Node& importDefList){

	if(loadBVH(filename, importScene, importSceneTreeList, importDefList)){
		//importScene->genNodeList2();
		//postLoadProcessNodes(g_sceneTreeList);

		//resetLooseEnds();
		//commonEndForSetupAnim();

		//Window::postLoadProcessWin(*importScene);

		//BackUp::deleteBackupFiles();
		//BackUp::createUniqueBackupName();

		//g_stw->paint();
		//Window::activeWindow=g_acp.p;
		//Window::paintAllControlPanels();
		//setSmlsSaved();
		//setTitleName();
/////////////////
		//decodeFunctions(str);
		importScene->genNodeList3();
		importSceneTreeList = g_sceneTreeList;

		createDefList(importDefList);
		postLoadProcessNodes(g_sceneTreeList);

		importScene->postLoadProcessCollapse();
		g_scene->genNodeList2();
		restoreMathPresision();
		return true;
	}
	restoreMathPresision();
	return false;
}

Seamless* Function::getSeamlessReference(){

	W_Field* w_dstField=0;
	W_Field* w_srcField=0;

	for(int i=codeLine.length; --i>=0;){
		if(getNodeFieldEqualsSomething(m_word, w_dstField, codeLine, i)){
			if(Art* part = dynamic_cast<Art*>(w_dstField->node)){
				return part->smls;
			}
		}
	}
	return 0;
}

bool hasIKChild(M_SArt& part){
	for(int k = part.length; --k>=0;){
		if(part[k]->ik){
			return true;
		}
	}
	return false;
}

void addInterpolatorPartToList(const bool pos, Art* part, M_Bool& m_pos, M_Art& m_interpolatorPart){
	int stopAt = m_interpolatorPart.length - 2;
	if(stopAt < 0){
		stopAt = 0;
	}
	for(int k = m_interpolatorPart.length; --k >= stopAt;){
		if(m_pos[k] == pos && m_interpolatorPart[k] == part){
			return;
		}
	}
	m_pos += pos;
	m_interpolatorPart += part;
}

void Function::addInterpToListForExportBVH(Art* part, M_Bool& pos, M_Art& m_interpolatorPart){

	W_Field* w_dstField=0;
	W_Field* w_srcField=0;
	bool addToList = false;

	for(int i = 0; i < codeLine.length; i++){
		if(getNodeFieldEqualsSomething(m_word, w_dstField, codeLine, i)){
			if(part == w_dstField->node){
				addToList = true;
				Field* dstField = w_dstField->node->getField(w_dstField->fieldIdx);
				if(&part->translation == dstField || (&part->ikPos == dstField && part->parent == 0)){
					addInterpolatorPartToList(true, part, pos, m_interpolatorPart);
				}
				else{
					addInterpolatorPartToList(false, part, pos, m_interpolatorPart);
				}
			}
		}
	}
	if(addToList == false && hasIKChild(part->part.refValue())){
		addInterpolatorPartToList(false, part, pos, m_interpolatorPart);
	}
}

void Art::addPartToTreeForExportBVH(String& str, Function* function, M_Bool& pos, M_Art& m_interpolatorPart){

	int k = pos.length;
	function->addInterpToListForExportBVH(this, pos, m_interpolatorPart);
	int j = pos.length - k;
	if(j){
		str += "\nCHANNELS ";
		str += j * 3;
		while(k < pos.length){
			if(pos[k]){
				str+=" Xposition Yposition Zposition";
			}
			else{
				str+=" Yrotation Zrotation Xrotation";
			}
			k++;
		}
	}

	M_SArt& m_part = part.refValue();


	if(m_part.length){
		for(int k = 0; k < m_part.length; k++){
			Art* part = m_part[k];
			str += "\nJOINT ";
			str += part->name;
			str += "\n{\nOFFSET ";
			str += toInches * part->pivotPoint;
			part->addPartToTreeForExportBVH(str, function, pos, m_interpolatorPart);
		}
	}
	else{
		str += "\nEnd Site\n{\nOFFSET ";
		float furthest = .001;

		for(int i = coord.length; --i>=0;){
			float a = (coord[i] - absolutePivotPoint).length();
			if(a > furthest){
				furthest = a;
			}
		}
		str += pivotPoint.refValue().normalize() * furthest * toInches;
		str += "\n}";
	}
	str += "\n}";
}

void Anim::saveBvh(const char* filename){
	setMathHiPresision();
	float multiply = 1;
	float period = 1;
	float step = 0;

	multiply = speed;
	period = this->period;
	step = 1.0 / (stepsPerSecond * period);
	slider = 0;


	Function* function = dynamic_cast<Function*>(onFrame.p);

	if(function == 0){
		return;
	}

	Seamless* smls = function->getSeamlessReference();

	M_SArt& skeleton = smls->skeleton.refValue();
	M_Art& partList = smls->partList;

	String str;
	str = "HIERARCHY\n";
	
	M_Bool pos;
	M_Art m_interpolatorPart;

	for(int k = 0; k < skeleton.length; k++){
		Art* part = skeleton[k];
		if(part->show){
			str+= "ROOT ";
			str+= part->name;
			str+= "\n{\nOFFSET ";
			str+= toInches * part->pivotPoint;
			part->addPartToTreeForExportBVH(str, function, pos, m_interpolatorPart);
		}
	}
	str += "\nMOTION\nFrames: ";
	int frames = stepsPerSecond * period + 1.5;
	str += frames;
	str += "\nFrame Time: ";
	str += period / (frames - 1);
	Rotation rot;
	Rotation4d answer;

	for(int f = 0; f < frames; f++){
		str += '\n';
		float frac = 1.0 / (frames - 1) * f;
		slider = frac; //step * f * multiply;
		render();
		for(int k = 0; k < pos.length; k++){
			if(pos[k]){
				Vec3f trans = m_interpolatorPart[k]->translation.refValue() * toInches;
				str += ' ';
				str += trans;
			}
			else{
				//Art* part = m_interpolatorPart[k];
				//rot = part->rotation;
				//if(part->name == "l_shoulder" && frac > .47){
				//	Quaternion quat = m_interpolatorPart[k]->rotation.refValue().quaternion();
				//	Vec3d euler = quat.eulerYZX() / PI * 180;

				//	Rotation4d rotA = Rotation4d(1, 0, 0 ,radians(euler.x));
				//	Rotation4d rotB = Rotation4d(0, 0, 1 ,radians(euler.y));
				//	Rotation4d rotC = Rotation4d(0, 1, 0 ,radians(euler.z));
				//	answer = rotA * rotB * rotC;
				//	
				//	int a = 0;
				//}
				Quaternion quat = m_interpolatorPart[k]->rotation.refValue().quaternion();
				Vec3d euler = quat.eulerYZX() / PI * 180;
				str += ' ';
				str += euler;
			}
		}
	}
//	for(int f = 0; f < frames; f++){
//		str += '\n';
//		float frac = 1.0 / (frames - 1) * f;
//		slider = frac; //step * f * multiply;
//		render();
//
//		for(int i = 0; i < partList.length; i++){
//			Art* part = partList[i];
//			Quaternion quat = part->rotation.refValue().quaternion();
//			Vec3f euler = quat.eulerYZX() / PI * 180;
//			str += ' ';
//			str += euler;
//
//			if(i == 0){
//				Vec3f pos = part->translation * toInches;
//				str += ' ';
//				str += pos;
//			}
////			if(i > 0 && dynamic_cast<Art*>(part->parent) == 0){ //dont bother exporting non skeleton parts
//		}
//	}


	str += '\n';

	g_writeGzipped = false;
	str.write(filename);
	ExportVrmlX3d* exportVrmlX3d = dynamic_cast<ExportVrmlX3d*>(g_scene->exportVrmlX3d.p);
	g_writeGzipped = exportVrmlX3d->gZip;
	restoreMathPresision();
	update = false;
}

NurbsLathe* Art::latheFromTwoPoints(const Vec3f& from, const Vec3f& to){
	if(to.length() < .0001){
		return 0;
	}
	if(NurbsLathe* nurbsLathe = addNewLathe(this, from)){
		M_Vec2f& point = nurbsLathe->point.refValue();
		float diff = to.length();
		nurbsLathe->rotation.refValue() = Rotation(Vec3f(0, 1, 0), to);
		point[0] = Vec2f();
		point[1] = Vec2f(geomRad, 0);
		point[6] = Vec2f(0, diff);
		point[5] = Vec2f(geomRad, diff);
		point[3] = (point[1] + point[5]) / 2;
		if(isRightMirror()){
			nurbsLathe->mirrorPoints();
		}
		nurbsLathe->processFields();

		return nurbsLathe;
	}
	return 0;
}

SphereBuild* Art::sphereFromTwoPoints(const Vec3f& from, const Vec3f& to){
	if(to.length() < .0001){
		return 0;
	}
	SphereBuild* sphereBuild = new SphereBuild;
	sphereBuild->part.refValue() = this;
	sphereBuild->translation.refValue() = from;
	sphereBuild->yCentred.refValue() = true;
	
	sphereBuild->radius.refValue() = to.length();
	sphereBuild->rotation.refValue() = Rotation(Vec3f(0, 1, 0), to);
	sphereBuild->processFields();

	return sphereBuild;
}

void Art::lathesFromPivotPoints(M_Build& m_build, const Vec3f accumOffset){

	M_SArt& m_part = part.refValue();

	if(m_part.length){
		for(int k = 0; k < m_part.length; k++){
			Art* part = m_part[k];
			if(NurbsLathe* nurbsLathe = latheFromTwoPoints(accumOffset, part->pivotPoint)){
				m_build += nurbsLathe;
			}
			part->lathesFromPivotPoints(m_build, accumOffset + part->pivotPoint);
		}
	}
	else{
		float furthest = .001;

		for(int i = coord.length; --i>=0;){
			float a = (coord[i] - absolutePivotPoint).length();
			if(a > furthest){
				furthest = a;
			}
		}
		Vec3f toOffset = pivotPoint.refValue().normalize() * furthest;

		if(NurbsLathe* nurbsLathe = latheFromTwoPoints(accumOffset, toOffset)){
			m_build += nurbsLathe;
		}
	}
}

void convertAllToWhiteParts();

void createLathesFromPivotPoints(){
	if(Seamless* smls=dynamic_cast<Seamless*>(g_selNode.getRaw())){
		BackUp::monostable.beginCycle();

		M_SArt& skeleton = smls->skeleton.refValue();
		GroupBuild* group = new GroupBuild;
		smls->build.refValue().length = 0;
		smls->build.refValue() += group;
		M_Build& m_build = group->build.refValue();

		for(int k = 0; k < skeleton.length; k++){
			Art* part = skeleton[k];
			part->lathesFromPivotPoints(m_build, part->pivotPoint);
		}
		g_scene->genNodeList2();
		convertAllToWhiteParts();
	}
}

