/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#define STRICT
#include <windows.h>
#include <windowsx.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "seamless3d_gui_render.h"
#include "gui.h"
#include "ms-gui.h"


#define BUFFER_SIZE           16
#define RECORD_3D_POINTER 1

HANDLE                  g_hMouseEvent   = NULL;



extern HINSTANCE               g_hInst;
extern HWND hWndMain;
extern SettingsPtr g_settings;

extern bool g_winAPI3DMode;
bool g_zDragMode = false;
bool g_viewPosDragMode = false;
bool g_touchViewPosDragMode = false;
Vec3f g_touchBeginMousePos;
Vec3f g_mouseFrom;
bool g_viewRotDragMode = false;
bool g_zoomSpin = false;
bool g_radiusControl = false;
float g_beginRadius = .5;
bool g_leftButtonZoom = false;
bool g_viewZRotDragMode = false;
bool g_insertControlPoint = false;
bool g_deleteControlPoint = false;
Vec3f g_viewDragMousePos;
bool MouseFlag = false;
bool tugFlag = false;
bool g_tugRotFlag = false;
bool orientationTugFlag = false;
bool dragPartFlag = false;
bool g_dragMorphFlag = false;
bool g_dragActuallyHappened = false;

bool	g_mirrorModeVertexFound=false;
bool	mouse3dMode = false;
static int cursorCounter = 0;
Rotation g_beginViewRot;
Vec3f g_beginViewPos;
Vec3f g_touchBeginViewPos;
Vec2i g_rightDragCounter;
Vec2i g_beginMousePos;
void get2dMousePosFrom3d();
void restoreRobotFrom3d();
Vec2i get2dMousePosFrom3d(Vec3f par);

void updateBuild3dWindow(bool ifPartBuild=false);
void backUpSmlsCoord(const M_Art& partList);
extern Ptr<PartClue> g_vcArtTriBase;
extern Ptr<PartCageClue> g_partCageClue;
extern TransformPtr g_vcBezierBase;
extern Ptr<Palette> g_palette;
extern bool g_joinActive;
void commonIncDecColorChange(Seamless* smls, Art* part, M_Color3f& color, const int colorIdx);
extern Ptr<Region> g_region;
extern BrushPtr g_brush;
extern Ptr<Grid> g_grid;
extern Material* g_vcEllipticalMaterial;
void createNewBuildNode(Build* buildNode, Seamless* smls, Art* part);
void mouse2dSync();
Vec3f mouse2dSync(Vec2i pixelPos);

BOOL   g_MousebActive      = TRUE;
BOOL   g_bSwapMouseButtons;





NurbsPatch* getSelNodePatch(){
	Node* node = g_selNode.getRaw();
	NurbsPatch* patch  = dynamic_cast<NurbsPatch*>(node);
	if(g_transformCP->selectObject){
		if(Build* bld = dynamic_cast<Build*>(node)){
			patch = bld->getPatch();
		}
	}
	return patch;
}

bool g_directInputMode = false;

void release3dDragFinal(){
#ifdef PROG_ROB
	g_storeCrossPoint.active = false;
#endif
	NurbsPatch* patch = getSelNodePatch();

	if(dynamic_cast<Lock*>(g_selNode.getRaw()) && (g_scene->mode & HAIR_REGION_MODE)){
		if(Lock::selBorderLock){
			g_region->borderPoint = Lock::selBorderLock->border[Lock::selBorder];
		}
	}
	else if(patch){
		//if(g_transformCP->dragRotPointMode){
		//	g_transformCP->quantizeRotationLever();
		//}
		if(g_transformCP->dragPivotPointMode){
			g_transformCP->pivotPoint.refValue() += g_transformCP->translation.refValue();
			g_grid->snapIt(g_transformCP->pivotPoint.refValue());
			g_transformCP->refreshRemPoints();
		}
		if(Seamless* smls = patch->smls){
			if(g_dragActuallyHappened){
				//if(smls->animCP == false){
				if(Anim* anim = Window::getAnimBar()){
					if(anim->soundTrack.length){
						if(Composition* composition = dynamic_cast<Composition*>(anim->soundTrack[0].p)){
							composition->updateAndPlay(anim, patch);
						}
					}
				}
				//}
			}
		}
	}
	else if(Oscillator* osc = dynamic_cast<Oscillator*>(g_selNode.getRaw())){
		if(MouseFlag){
			if(Anim* anim = Window::getAnimBar()){
				if(anim->soundTrack.length){
					if(Composition* composition = dynamic_cast<Composition*>(anim->soundTrack[0].p)){
						composition->updateAndPlay(anim, osc);
					}
				}
			}
		}
	}
	else if(Chords* chords = dynamic_cast<Chords*>(g_selNode.getRaw())){
		if(MouseFlag){
			if(Anim* anim = Window::getAnimBar()){
				if(anim->soundTrack.length){
					if(Composition* composition = dynamic_cast<Composition*>(anim->soundTrack.refValue()[0].p)){
						composition->updateAndPlay(anim, chords);
					}
				}
			}
		}
	}
	if(MultiSelect::selectRectMode){
		MultiSelect::selectRectMode = false;
		g_visualClue = g_vcBezierBase;
		if(BezierPatch* bezierPatch = dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
			bezierPatch->joinEdge();
		}
		if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
			MultiSelect::selectTriangles(part);
			if(g_palette->rectangle){
				commonEndVertexChange();
			}
		}
		else if(Build* build = dynamic_cast<Build*>(g_selNode.getRaw())){
			if(Lock* lock = dynamic_cast<Lock*>(build)){
				if((g_scene->mode & HAIR_REGION_MODE)){
					lock->selectRegions();
					return;
				}
			}
			Seamless* smls = getSeamless2();
			smls->selectControlPoints();
			houseKeepVisualClue();
		}
	}
	else if(g_viewZRotDragMode == false && g_viewRotDragMode == false && g_viewPosDragMode == false && g_insertControlPoint == false && g_deleteControlPoint == false){
		if(g_directInputMode){
			get2dMousePosFrom3d();
		}
		//SetCursorPos(screenPos.x,screenPos.y);
	}
	else {
		g_viewZRotDragMode = false;
		g_viewRotDragMode = false;
		g_viewPosDragMode = false;
		g_touchViewPosDragMode = false;
		g_insertControlPoint = false;
		g_deleteControlPoint = false;
		g_radiusControl = false;
	}
	//restoreRobotFrom3d();

	showCursor(true);
	g_transformCP->mode = TransformCP::NOTHING;
	g_zDragMode = false;
	g_zoomSpin = false;
	g_leftButtonZoom = false;
	g_viewPosDragMode = false;
	g_viewRotDragMode = false;
	g_radiusControl = false;
	g_viewZRotDragMode = false;
	g_insertControlPoint = false;
	g_deleteControlPoint = false;
	MouseFlag = false;
	tugFlag = false;
	g_tugRotFlag = false;
	orientationTugFlag = false;
	dragPartFlag = false;
	g_dragMorphFlag = false;
	g_partCageClue->mode = 0;

	g_mirrorModeVertexFound = false;
	mouse3dMode = false;
}

Vec3f localDrag;
unsigned g_state=0;

bool checkRobotActive();

void commonPostTug(){
	g_scene->updateBuild();
	BackUp::monostable.beginCycle();
}
void EC2::drag(Vec3f pt){
	M_Vec2f& point=this->point;
	Vec3f svec3f= g_scene->getIViewMatrix() * translationMatrix(-location) * rotationMatrix(-orientation) * pt;
	//Vec3f svec3f=getIViewMatrix() * pt;
	int idx=index[0];
	if(orientationTugFlag==true){
		float diffX=point[idx].x-svec3f.x;
		float diffY=point[idx].y-svec3f.y;

		tangent[idx]=atan2(diffY,diffX);
	}
	else{
		point[idx]=Vec2f(svec3f.x,svec3f.y);
	}
	g_selNode->processFields();
	ecHouseKeep();
	commonPostTug();
}
void EC3::drag(Vec3f pt){
	M_Vec2f& point=this->point;
	Vec3f svec3f= g_scene->getIViewMatrix() * translationMatrix(-location) * rotationMatrix(-orientation) * pt;
	//Vec3f svec3f=getIViewMatrix() * pt;
	int idx=index[0];
	if(orientationTugFlag==true){
		float diffX=point[idx].x-svec3f.x;
		float diffY=point[idx].y-svec3f.y;

		tangent[idx]=atan2(diffY,diffX);
	}
	else{
		point[idx].x=svec3f.x;
		point[idx].y=svec3f.y;
	}
	g_selNode->processFields();
	ecHouseKeep();
	commonPostTug();
}
void Bezier::drag(Vec3f pt){
	M_Vec2f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * imatrix * pt;
	matchBorderTangents(Vec2f(vec.x,vec.y));
	g_selNode->processFields();
	bezHouseKeep();
	commonPostTug();
}
void Bezier3f::drag(Vec3f pt){
	M_Vec3f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * imatrix * pt;
	matchBorderTangents(Vec3f(vec.x,vec.y,vec.z));
	g_selNode->processFields();
	bezHouseKeep();
	commonPostTug();
}
void BezierTape::drag(Vec3f pt){
	M_Vec2f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * rotationMatrix(Rotation(0,1,0,PI*.5)) * pt;
	matchBorderTangents(Vec2f(vec.x,vec.y));
	g_selNode->processFields();
	bezHouseKeep();
	commonPostTug();
}
int nextLink(int bp){
	bp+=4;
	if(bp==16){
		bp=0;
	}
	return bp;
}
Vec3f setOppositePoint(Vec3f hinge, Vec3f swingA, Vec3f swingB){
	float disA = (swingA - hinge).length();
	float disB = (swingB - hinge).length();
	float ratio = disB / disA;
//	return hinge + (swingA - hinge) * Vec3f(-ratio, -ratio, -ratio);

	return hinge + (hinge - swingA) * Vec3f(ratio, ratio, ratio);;
}
void setCorner(M_Vec3f& point, const M_Int& cornerLink, const int bp, const Vec3f vecA){
	Vec3f vecB = point[cornerLink[bp+1]];
	Vec3f vecC = point[cornerLink[bp+3]];

//	point[cornerLink[bp + 1]] = vecB;
//	point[cornerLink[bp + 3]] = vecC;

	point[cornerLink[bp + 2]] = vecB + (vecC - vecA);

//	point[cornerLink[nextLink(bp) + 1]] = point[cornerLink[bp + 3]];
}
void setCorner0(M_Vec3f& point, const M_Int& cornerLink, const int bp, const Vec3f vecA){
	Vec3f vecB = point[cornerLink[bp+2]];
	Vec3f vecC = point[cornerLink[bp+3]];

//	point[cornerLink[bp + 1]] = vecB;
//	point[cornerLink[bp + 3]] = vecC;

	point[cornerLink[bp + 1]] = vecB + (vecA - vecC);

//	point[cornerLink[nextLink(bp) + 1]] = point[cornerLink[bp + 3]];
}

void setCorners(M_Vec3f& point, const M_Int& cornerLink, const Vec3f pivot){
	setCorner(point, cornerLink, 0, pivot);
	setCorner(point, cornerLink, 4, pivot);
	setCorner(point, cornerLink, 8, pivot);
	setCorner(point, cornerLink, 12, pivot);
}
void ensureCoplaner(M_Vec3f& point, const M_Int& cornerLink){
	const int b0=0;
	const int b1=4;
	const int b2=8;
	const int b3=12;

	Vec3f normal = calcNormal(point[cornerLink[b0+3]], point[cornerLink[b0+1]], point[cornerLink[b0+0]]);
	Vec3f rightAngle = calcNormal(point[cornerLink[b0+3]], point[cornerLink[b0+0]], point[cornerLink[b0+0]]+normal);

	const Vec3f centerPoint = point[cornerLink[0]];

	{
		float dis = (point[cornerLink[b0+1]] - centerPoint).length();
		point[cornerLink[b0+1]]= rightAngle * dis + centerPoint;
	}
	{
		float dis = (point[cornerLink[b1+3]] - centerPoint).length();
		point[cornerLink[b1+3]] = -rightAngle * dis + centerPoint;
	}
	{
		point[cornerLink[b2+3]] = setOppositePoint(centerPoint, point[cornerLink[b0+3]], point[cornerLink[b2+3]]);
		point[cornerLink[b2+1]] = point[cornerLink[b1+3]];
	}
	{
		point[cornerLink[b3+1]] = point[cornerLink[b2+3]];
		point[cornerLink[b3+3]] = point[cornerLink[b0+1]];
	}

	setCorners(point, cornerLink, centerPoint);
}

void rotatePoint(M_Vec3f& point, const M_Int& cornerLink, int bp, Rotation rot, Vec3f pivot){
	point[cornerLink[bp+3]] = rot * (point[cornerLink[bp+3]] - pivot) + pivot;
}
void dragJoinedCornerEdge(int idx, Vec3f mousePos, M_Vec3f& point, const M_Int& cornerLink, int b0){
	int b1 = nextLink(b0);
	int b2 = nextLink(b1);
	int b3 = nextLink(b2);

	Vec3f pivot = point[cornerLink[b0+0]];
	Rotation rot = Rotation(point[idx]-pivot, mousePos-pivot);
	point[cornerLink[b3+3]] = mousePos;
	point[cornerLink[b0+1]] = mousePos;

	rotatePoint(point, cornerLink, b0, rot, pivot);
	rotatePoint(point, cornerLink, b1, rot, pivot);
	rotatePoint(point, cornerLink, b2, rot, pivot);

	point[cornerLink[b1+1]] = point[cornerLink[b0+3]];
	point[cornerLink[b2+1]] = point[cornerLink[b1+3]];
	point[cornerLink[b3+1]] = point[cornerLink[b2+3]];
	point[cornerLink[b3+3]] = point[cornerLink[b0+1]];

	ensureCoplaner(point, cornerLink);
}
void setPoint(M_Vec3f& point, const M_Int& cornerLink, const int bp, const Vec3f& mousePos, const Vec3f diff){
	point[cornerLink[bp+0]] = mousePos;
	//point[cornerLink[bp+2]] += diff;
	point[cornerLink[bp+3]] += diff;
	point[cornerLink[nextLink(bp) + 1]] = point[cornerLink[bp + 3]];
}
bool BezierPatch::doCornerLink(M_Vec3f& point, const int idx, const Vec3f mousePos){
	static M_Int cornerLink;
	const M_Int& link = this->link;
	
	cornerLink.length=4*4;

	for(int l = 0; l<link.length; ){
		for(int k = 0; k < 4; k++){
			int i = linkToPatchIdx[l];
			int noU=dim[i].x;
			int noV=dim[i].y;
			int pointLen=noU*noV;

			int j = link[l++];
			int h = j - m_bp[i];
			cornerLink[k*4+0] = j;
			if(h == 0){
				cornerLink[k*4+1] = j+1;
				cornerLink[k*4+2] = j+1+noU;
				cornerLink[k*4+3] = j+noU;
			}
			else if(h == noU - 1){
				cornerLink[k*4+1] = j+noU;
				cornerLink[k*4+2] = j+noU-1;
				cornerLink[k*4+3] = j-1;
			}
			else if(h == pointLen - noU){
				cornerLink[k*4+1] = j-noU;
				cornerLink[k*4+2] = j-noU+1;
				cornerLink[k*4+3] = j+1;
			}
			else if(h == pointLen - 1){
				cornerLink[k*4+1] = j-1;
				cornerLink[k*4+2] = j-1-noU;
				cornerLink[k*4+3] = j-noU;
			}
		}


		const int b0=0;
		const int b1=4;
		const int b2=8;
		const int b3=12;
		Vec3f diff;

		if(	cornerLink[b0] == idx){
			diff = mousePos-point[cornerLink[b0]];
			goto here;
		}
		else if(cornerLink[b1] == idx){
			diff = mousePos-point[cornerLink[b1]];
			goto here;
		}
		else if(cornerLink[b2] == idx){
			diff = mousePos-point[cornerLink[b2]];
			goto here;
		}
		else if(cornerLink[b3] == idx){
			diff = mousePos-point[cornerLink[b3]];
here:		
			point[cornerLink[b0+0]] = mousePos;
			point[cornerLink[b1+0]] = mousePos;
			point[cornerLink[b2+0]] = mousePos;
			point[cornerLink[b3+0]] = mousePos;


			point[cornerLink[b0+1]] += diff;
			point[cornerLink[b0+3]] += diff;
			point[cornerLink[b1+3]] += diff;
			point[cornerLink[b2+3]] += diff;

			point[cornerLink[nextLink(b0) + 1]] = point[cornerLink[b0 + 3]];

			ensureCoplaner(point, cornerLink);

			//point[cornerLink[4*2+3]] = setOppositePoint(point[cornerLink[0]], point[cornerLink[3]], point[cornerLink[4*2+3]]);
			return true;
		}
		else if(cornerLink[b0+1] == idx || cornerLink[b3+3] == idx){
			dragJoinedCornerEdge(idx, mousePos, point, cornerLink, b0);
			return true;
		}
		else if(cornerLink[b0+3] == idx || cornerLink[b1+1] == idx){
			dragJoinedCornerEdge(idx, mousePos, point, cornerLink, b1);
			return true;
		}
		else if(cornerLink[b1+3] == idx || cornerLink[b2+1] == idx){
			dragJoinedCornerEdge(idx, mousePos, point, cornerLink, b2);
			return true;
		}
		else if(cornerLink[b2+3] == idx || cornerLink[b3+1] == idx){
			dragJoinedCornerEdge(idx, mousePos, point, cornerLink, b3);
			return true;
		}
	/*
		else if(cornerLink[2] == idx){
			Vec3f pivot = point[cornerLink[b0+0]];
			Rotation rot = Rotation(point[idx]-pivot, mousePos-pivot);
			point[idx] = mousePos;

			rotatePoint(point, cornerLink, b0, rot, pivot);
			rotatePoint(point, cornerLink, b1, rot, pivot);
			rotatePoint(point, cornerLink, b2, rot, pivot);
			rotatePoint(point, cornerLink, b3, rot, pivot);

			point[cornerLink[b0+1]] = point[cornerLink[b3+3]];
			point[cornerLink[b1+1]] = point[cornerLink[b0+3]];
			point[cornerLink[b2+1]] = point[cornerLink[b1+3]];
			point[cornerLink[b3+1]] = point[cornerLink[b2+3]];

			Vec3f normal = calcNormal(point[cornerLink[b0+3]], point[cornerLink[b0+2]], pivot);
			Rotation rotB = Rotation(normal, Vec3f(0,0,1));

	//		point[cornerLink[1]] = rot * (point[cornerLink[1]] - pivot) + pivot;
			point[cornerLink[2]] = rotB * (point[cornerLink[2]] - pivot) + pivot;
			point[cornerLink[3]] = rotB * (point[cornerLink[3]] - pivot) + pivot;

			Rotation rotC = Rotation(point[cornerLink[b0+3]]-pivot, Vec3f(0,-1,0));
			point[cornerLink[2]] = rotC * (point[cornerLink[2]] - pivot) + pivot;
			point[cornerLink[3]] = rotC * (point[cornerLink[3]] - pivot) + pivot;

			point[cornerLink[b0+3]].y=point[cornerLink[2]].y;

			point[cornerLink[2]] = -rotC * (point[cornerLink[2]] - pivot) + pivot;
			point[cornerLink[3]] = -rotC * (point[cornerLink[3]] - pivot) + pivot;

			point[cornerLink[2]] = -rotB * (point[cornerLink[2]] - pivot) + pivot;
			point[cornerLink[3]] = -rotB * (point[cornerLink[3]] - pivot) + pivot;
			
			setCorner0(point, cornerLink, 0, pivot);

			//		rotatePoint(point, cornerLink, b0, rot, pivot);
	//		rotatePoint(point, cornerLink, b1, rot, pivot);
	//		rotatePoint(point, cornerLink, b2, rot, pivot);
	//		rotatePoint(point, cornerLink, b3, rot, pivot);

	//		point[cornerLink[b0+1]] = point[cornerLink[b3+3]];
	//		point[cornerLink[b1+1]] = point[cornerLink[b0+3]];
	//		point[cornerLink[b2+1]] = point[cornerLink[b1+3]];
	//		point[cornerLink[b3+1]] = point[cornerLink[b2+3]];

			//Vec3f rightAngle = calcNormal(point[cornerLink[b0+2]], pivot+normal, pivot);
			//point[cornerLink[b1+2]]
			//Vec3f v = (Vec3f(1,1,1)-rightAngle)*((point[cornerLink[b0+2]]-pivot).length())+pivot;
			//Vec3f v = rightAngle*(point[cornerLink[b0+2]]-pivot)+pivot;
			//point[cornerLink[b1+2]]=v;
			//Vec3f diff = (v - point[cornerLink[b0+2]]) * .5;
			//point[cornerLink[b0+3]] = point[cornerLink[b1+1]] = point[cornerLink[b0+2]] + diff;
			//Vec3f vecA = pivot;
			//Vec3f vecB = point[cornerLink[b0+2]];
			//Vec3f vecC = point[cornerLink[b0+3]];
			//point[cornerLink[b0 + 1]] = vecB + (vecA - vecC);
	//		setCorners(point, cornerLink, pivot);
			return true;
		}
	*/
	}
	return false;
}
void BezierPatch::doLink(M_Vec3f& point, const M_Int& link, int idx, Vec3f diff, int exclude){

	for(int i=link.length; (i-=4)>=0;){
		if(i==exclude)continue;

		if(link[i+hingeB]==idx){
			int sA=link[i+swingA];
			int hA=link[i+hingeA];
			int hB=link[i+hingeB];
			int sB=link[i+swingB];

			point[hA]=point[hB];
			point[sA]+=diff;
			point[sB]=point[hA]+(point[hA]-point[sA]);

			//doLink(point, link, sA, Vec3f(), hA);
			//doLink(point, link, sB, Vec3f(), hA);
		}
		else if(link[i+hingeA]==idx){
			int sA=link[i+swingA];
			int hA=link[i+hingeA];
			int hB=link[i+hingeB];
			int sB=link[i+swingB];

			point[hB]=point[hA];
			point[sB]+=diff;
			point[sA]=point[hB]+(point[hB]-point[sB]);

			//doLink(point, link, sA, Vec3f(), hB);
			//doLink(point, link, sB, Vec3f(), hB);
		}
		else if(link[i+swingA]==idx){
			int sB=link[i+swingB];
			int sA=link[i+swingA];
			int h=link[i+hingeB];

			point[sB]=point[h]+(point[h]-point[sA]);
//			doLink(point, link, sA, Vec3f(), sB);
		}
		else if(link[i+swingB]==idx){
			int sB=link[i+swingB];
			int sA=link[i+swingA];
			int h=link[i+hingeB];

			point[sA]=point[h]+(point[h]-point[sB]);
//			doLink(point, link, sB, Vec3f(), sA);
		}
	}
}
void BezierPatch::refreshLink(M_Int& link){
	return;
	for(int i=link.length; (i-=4)>=0;){
		for(int j=i; (j-=4)>=0;){
			if(link[j+hingeA]==link[i+hingeA] && link[j+hingeB] != link[i+hingeA]){
				link[j+hingeA]=link[i+hingeB];
			}
			if(link[j+hingeB]==link[i+hingeA] && link[j+hingeA] != link[i+hingeA]){
				link[j+hingeB]=link[i+hingeB];
			}
		}
	}
}
/*
void SineTug::drag(Vec3f pt){
	int i=index[0];
	const M_Vec3f& location = this->location;
	M_Rotation& orientation = this->orientation.refValue();
	M_Float& distance = this->distance.refValue();

	Vec3f displacement = g_scene->getIViewMatrix() * pt - location[i];
	distance[i] = displacement.length() /8;
	orientation[i] = Rotation(Vec3f(0,1,0), displacement);

	commonPostTug();
	houseKeep();
}
*/
void SineTug::drag(Vec3f pt){
	int i=index[0];
	const M_Vec3f& location = this->location;
	M_Rotation& orientation = this->orientation.refValue();
	M_Float& distance = this->distance.refValue();

	Matrix matrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * g_scene->getViewMatrix();

	Vec3f pivotPoint = matrix * Vec3f();

	Vec3f controlPoint;
	{
		Vec3f vec;
		float* ptr=0;
		ptr=&distance[i];
		controlPoint = matrix * Vec3f(0, *ptr*8, 0);

		Vec3f loc = location[i];
		Rotation orien = orientation[i];

		Matrix imatrix = translationMatrix(Vec3f(-loc.x, -loc.y, -loc.z)) * rotationMatrix(-orien);

		Vec3f newPt = g_scene->getIViewMatrix() * imatrix * pt;

		if(distance[i]>=0){
			*ptr = sqrt(newPt.x*newPt.x + newPt.y*newPt.y + newPt.z*newPt.z)/8;
		}
		else{
			*ptr = -sqrt(newPt.x*newPt.x + newPt.y*newPt.y + newPt.z*newPt.z)/8;
		}

		{

			//would there not be an easier way to calculate the orientation with the information we have at this point in time?

			Vec3f toDiff = Vec3f(pt.x-pivotPoint.x, pt.y-pivotPoint.y, pt.z-pivotPoint.z);
			Vec2f fromDiff = Vec2f(controlPoint.x-pivotPoint.x, controlPoint.y-pivotPoint.y);
			float toAngle = Vec2f(toDiff.x, toDiff.y).getAngle();
			float fromAngle = fromDiff.getAngle();
			float angle = toAngle - fromAngle;

			Vec3f axis = (-g_scene->rotation * Vec3f(0, 0, 1)).normalize();
			orientation[i] = orientation[i] * Rotation(axis, angle);
		}
		Vec3f sr;
		{
			sr= Vec3f(0, distance[i]*8, 0);
		}
		Matrix postMatrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * g_scene->getViewMatrix();
		g_pointer.translation= postMatrix * sr;
	}
	houseKeep();
	commonPostTug();
}


void BezierPatch::drag(Vec3f pt){
	M_Vec3f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx=index[0];

	if(doCornerLink(point, idx, vec) == false){
		point[idx] = vec;
	}

	processFields();
	commonPostTug();
	houseKeep();
}
void Subdivision::drag(Vec3f pt){
	M_Vec3f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx=index[0];
	point[idx]=vec;
	commonPostTug();
	houseKeep();
}
int findMirrorVertex(const M_Vec3f& cPtr, Vec3f svec);

inline bool NurbsPatch::doSameViewLocation(const Vec3f& pt, const Vec3f& svec, const int& c2, const int& r, const int& xDim, const int& yDim, M_Vec3f& point, const M_Vec3f& dpoint, const M_Float& weight, const Matrix& imatrix, const Matrix& mtrx){
	int idx2 = r * xDim + c2;
	Vec3f svec2 = mtrx * dpoint[idx2];
	if(0.001 > abs(svec.x - svec2.x) && 0.001 > abs(svec.y - svec2.y)){
		Vec3f pt2 = pt;
		pt2.z = svec2.z;
		Vec3f vec2 = g_scene->getIViewMatrix() * imatrix * pt2;

		if(mirror){
			Vec3f svec = dpoint[idx2];
			if(svec.x==0){
				vec2.x=0;
			}
			else{
				svec.x=-svec.x;
				int mirrorIdx = findMirrorVertex(dpoint, svec);
				if(mirrorIdx>=0){
					svec=vec2;
					svec.x=-svec.x;
					//point[mirrorIdx]=svec*weight[mirrorIdx];
					g_transformCP->setFilteredDimension(point, smls->cp.refValue(), cpIndex.refValue(), mirrorIdx, svec*weight[mirrorIdx]);
				}
			}
		}


		//vec2 = Vec3f();
		//point[idx2] = vec2 * weight[idx2];
		g_transformCP->setFilteredDimension(point, smls->cp.refValue(), cpIndex.refValue(), idx2, vec2 * weight[idx2]);
//		String str;
//		vec2.toString(str);
//		dLog(str);
		return true;
	}
	return false;
}


void processAllPatches(M_<NurbsPatch*>& cpNodeList);

Vec2f getScreenCoordinate(const bool perspective, Vec3f sr){
	Vec2f v;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);
	}
	else{
		v.x = sr.x;
		v.y = sr.y;
	}
	return v;
}

extern M_Vec3f g_backUpCoord;


//	Vec3f vec= g_scene->getIViewMatrix() * imatrix * pt;
//
//	if(g_visualClue == g_vcRotateEditBase){
//		g_vcRotateEditBase->drag(pt);
//		return;
//	}
//	else{
//		if(sameViewPositionTug){
//			int incAmount = 1;
//			if(showAll == false){
//				incAmount++;
//			}
//			Matrix mtrx = g_viewMatrix * matrix;
//			Vec3f svec = mtrx * dpoint[idx];
//			int xDim = dim.x;
//			int yDim = dim.y;
//			int c = idx%xDim;
//			int r = idx/xDim;
//			int c2 = c + incAmount;
//			if(c2 > xDim){
//				c2 = xDim - c2; 
//			}
//			if(c2 < 0){
//				c2 += xDim - 1;
//			}
//			if(!doSameViewLocation(pt, svec, c2, r, xDim, yDim, point, dpoint, weight, imatrix, mtrx)){
//				int c2 = c - incAmount;
//				if(c2 < 0){
//					c2 += xDim - 1;
//				}
//				doSameViewLocation(pt, svec, c2, r, xDim, yDim, point, dpoint, weight, imatrix, mtrx);
//			}
//		}
//		//if(mirrorTug){
//		//	Vec3f svec = dpoint[idx];
//		//	if(mirrorPatch){
//		//	}
//		//	else{
//		//		if(svec.x==0){
//		//			vec.x=0;
//		//		}
//		//		else{
//		//			svec.x=-svec.x;
//		//			int mirrorIdx = findMirrorVertex(dpoint, svec);
//		//			if(mirrorIdx>=0){
//		//				svec=vec;
//		//				svec.x=-svec.x;
//		//				//point[mirrorIdx]=svec*weight[mirrorIdx];
//		//				setFilteredDimension(point, mirrorIdx, svec*weight[mirrorIdx]);
//		//			}
//		//		}
//		//	}
//		//}
//		//point[idx]=vec*weight[idx];
////		setFilteredDimension(point, idx, vec*weight[idx]);
//
//
//		if(MultiSelect::pointList.length){
//			Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * imatrix * pt;
//			Vec3f diff = newPt - MultiSelect::beginDragPos[MultiSelect::clickedIndex];
//			for(int i = MultiSelect::pointList.length; --i>=0;){
//				int idx = MultiSelect::pointList[i].i;
//				NurbsPatch* patch = MultiSelect::pointList[i].p;
//				Vec3f svec = MultiSelect::beginDragPos[i] + diff;
//				patch->setFilteredDimension(patch->point.refValue(), idx, svec * patch->weight[idx]);
//				patch->dpoint.refValue()[idx] = patch->point[idx] / patch->weight[idx];
//				if(patch->mirror){
//					if(NurbsPatch* mirrorPatch = patch->mirrorPatch){
//						Vec3f vec = patch->point[idx];
//						vec.x = -vec.x;
//						mirrorPatch->point.refValue()[idx] = vec;
//						vec = patch->dpoint[idx];
//						vec.x = -vec.x;
//						mirrorPatch->dpoint.refValue()[idx] = vec;
//					}
//					else{
//						int uDim = patch->dim.x;
//						int vDim = patch->dim.y;
//						int u = idx % uDim;
//						int v = idx / uDim;
//						int middleIdx = (uDim >> 1);
//
//						if(u == middleIdx){
//							patch->point.refValue()[idx].x = 0;
//							patch->dpoint.refValue()[idx].x = 0;
//						}
//						else{
//							int midx = v * uDim + uDim - 1 - u;
//							Vec3f vec = patch->point[idx];
//							vec.x = -vec.x;
//							patch->point.refValue()[midx] = vec;
//							vec = patch->dpoint[idx];
//							vec.x = -vec.x;
//							patch->dpoint.refValue()[midx] = vec;
//						}
//					}
//				}
//			}
//			smls->calcAllPolyMidPoints();
//			commonPostTug();
//			updateCage(smls);
//			houseKeep();
//			return;
//		}
//	}
//
//	processFields();
//	commonPostTug();
//	houseKeep();
//}

void Oscillator::drag(Vec3f pt){
	M_Vec2f& fPoint = this->note;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx = index[0];
	if(idx == 0){
		vec.x = 0;
	}
	float w = 1;
	if(fWeight.length){
		w = fWeight[idx];
	}
	fPoint[idx] = Vec2f(vec.x, vec.y*10) * w;
	processFields();
	commonPostTug();
	houseKeep();
	g_dragActuallyHappened = true;
}

void Envelope::drag(Vec3f pt){
	M_Vec2f& point = this->point;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx = index[0];
	if(idx == 0){
		vec.x = 0;
	}
	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = Vec2f(vec.x, vec.y*10) * w;
	processFields();
	commonPostTug();
	houseKeep();
	g_dragActuallyHappened = true;
}

void AmplitudeModulator::drag(Vec3f pt){
	M_Vec2f& point = this->wet;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx = index[0];
	if(idx == 0){
		vec.x = 0;
	}
	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = Vec2f(vec.x, vec.y*10) * w;
	processFields();
	commonPostTug();
	houseKeep();
	g_dragActuallyHappened = true;
}

void FrequencyModulator::drag(Vec3f pt){
	M_Vec2f& point = this->wet;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx = index[0];
	if(idx == 0){
		vec.x = 0;
	}
	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = Vec2f(vec.x, vec.y*10) * w;
	processFields();
	commonPostTug();
	houseKeep();
	g_dragActuallyHappened = true;
}

void Chords::drag(Vec3f pt){
	M_Vec2f& point = this->spectrum;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx = index[0];
	if(idx == 0){
		vec.x = 0;
	}
	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = Vec2f(vec.x, vec.y*10) * w;
	processFields();
	commonPostTug();
	houseKeep();
	g_dragActuallyHappened = true;
}


NurbsLathe* getMirrorLatheLoop(M_Build& build, const String& mirrorName){
	for(int k = 0; k < build.length; k++){
		Build* bld = build[k];
		if(NurbsLathe* lathe = dynamic_cast<NurbsLathe*>(bld)){
			if(Art* part = lathe->part.refValue().p){
				if(part->name == mirrorName){
					return lathe;
				}
			}
		}
		else if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(bld)){
			if(NurbsLathe* lathe = getMirrorLatheLoop(groupBuild->build.refValue(), mirrorName)){
				return lathe;
			}
		}
	}
	return 0;
}

NurbsLathe* NurbsLathe::getMirrorLathe()const{
	if(const Art* part = this->part.refValue().p){
		if(part->mirrorPart){
			return getMirrorLatheLoop(part->smls->build.refValue(), part->mirrorPart->name);
		}
	}
	return 0;
}

SphereBuild* getMirrorSphereLoop(M_Build& build, const String& mirrorName){
	for(int k = 0; k < build.length; k++){
		Build* bld = build[k];
		if(SphereBuild* sphere = dynamic_cast<SphereBuild*>(bld)){
			if(Art* part = sphere->part.refValue().p){
				if(part->name == mirrorName){
					return sphere;
				}
			}
		}
		else if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(bld)){
			if(SphereBuild* sphere = getMirrorSphereLoop(groupBuild->build.refValue(), mirrorName)){
				return sphere;
			}
		}
	}
	return 0;
}

SphereBuild* SphereBuild::getMirrorSphere()const{
	if(const Art* part = this->part.refValue().p){
		if(part->mirrorPart){
			return getMirrorSphereLoop(part->smls->build.refValue(), part->mirrorPart->name);
		}
	}
	return 0;
}


void NurbsLathe::drag(Vec3f pt){
	g_dragActuallyHappened = true;
	M_Vec2f& point = this->point.refValue();
	int idx = index[0];
	Vec3f vec= g_scene->getIViewMatrix() * imatrix * pt;
	vec.z = 0;
	if(beginSP && idx == 0){
		vec.x = 0;
	}
	if(endSP && idx == point.length - 1){
		vec.x = 0;
	}

	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = Vec2f(vec.x, vec.y) * w;

	if(mirrorContainer){
		mirrorContainer->point.refValue()[idx] = Vec2f(vec.x, vec.y).mirrorX() * w;
		mirrorContainer->processFields();
	}
	if(point.length > 2){
		processFields();
	}

	g_pointer.translation = matrix * g_scene->getViewMatrix() * vec;
	g_cursorOverPointIlsBase->translation = g_pointer.translation;


	commonPostTug();
	houseKeep();
}

void BendyLathe::drag(Vec3f pt){
	g_dragActuallyHappened = true;

	Vec3f vec = g_scene->getIViewMatrix() * imatrix * pt;
	M_Vec3f& point=this->point;

	if(showRingControlPoints){
		int idx=index[2];
		if(radIdx<0){
			dragRingBase(pt, idx, point, weight, knot, minT(), maxT(), ringBase[idx]);
		}
		else{
			dragRadius(pt, idx);
		}
//		M_Float& radius = this->radius;
//		M_Float& at = this->at;
	}
	else{
		int idx=index[0];
		float w = 1;
		if(weight.length){
			w = weight[idx];
		}
		point[idx]=vec*w;
	}
	processFields();
	commonPostTug();
	houseKeep();
}

void NikInterpolator::drag(Vec3f pt){
	M_Vec3f& focusPoint = focusOffset ? shortPoint : longPoint;

	if(Anim* anim = Window::getAnimBar()){
		g_dragActuallyHappened = true;

		Vec3f vec = g_scene->getIViewMatrix() * pt;

		float w = 1;
		if(weight.length){
			w = weight[selPointIdx];
		}
	
		focusPoint[selPointIdx] = vec;
		vec = vec * w;

		take(anim->slider, vec, selPointIdx, focusPoint.length * 2, focusOffset);
	

		g_pointer.translation = pt;
		g_cursorOverPointIlsBase->translation = pt;

		processFields();
		commonPostTug();
		houseKeep();

		anim->update = true;
		anim->renderNodeFirst();
		anim->update = false;
	}
}

void NurbsTrack::drag(Vec3f pt){
	g_dragActuallyHappened = true;

	Vec3f vec = g_scene->getIViewMatrix() * pt;
	M_Vec3f& point = this->keyValue.refValue();

	int idx = index[1];
	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = vec * w;

	g_pointer.translation = pt;
	g_cursorOverPointIlsBase->translation = pt;

	processFields();
	commonPostTug();
	houseKeep();

	if(Anim* anim = Window::animMode()){
		anim->update = true;
		anim->renderNodeFirst();
		anim->update = false;
	}


}

void Harmonic::drag(Vec3f pt){
}

void HarmonicPatch::drag(Vec3f pt){
	M_Vec3f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx=index[0];
	point[idx]=vec*weight[idx];
	processFields();
	commonPostTug();
	houseKeep();
}
void ColorSweep::drag(Vec3f pt){
	int idx=index[0];
	M_Float& distance=this->distance;
	if(idx==0){
		Vec3f vec= g_scene->getIViewMatrix() * pt;
		location.refValue()=vec;
	}
	else{
		Vec3f vec= g_scene->getIViewMatrix() * imatrix * pt;
		for(int i=0; i<idx-1; i++){
			vec.y-=distance[i];
		}
		distance[idx-1]=vec.y;
	}
	processFields();
	commonPostTug();
	houseKeep();
}
void CCLathe::drag(Vec3f pt) 
{
	int idx = index[0];
	Vec3f vec = g_scene->getIViewMatrix() * pt;
	Vec2f pos = Vec2f(vec.x, vec.y);

	// adjust angle  
	if ( angleDrag )  {
		if ( idx == 0 ) {
			pos -= seg[idx].Pf;
			double a = pos.angle() / (PI / 4);
			if ( 0x8000 & GetKeyState(VK_LSHIFT) ) {
				double ia = floor(a + 0.5);
				if ( fabs(a - ia) * controlLen < 0.05 ){
					a = ia;
				}
			}
			curve.refValue()[idx] = a * (PI / 4);
		}
		else {
			bool notches = (0x8000 & GetKeyState(VK_LSHIFT)) != 0;
			double ro = seg[idx].Ro.length();
			double ao = seg[idx].Ro.angle();
			Vec2f Rf = pos - seg[idx].C;
			Rf *= (ro / Rf.length());
			double a = fmod(Rf.angle() - ao + 2*PI, 2*PI);

			if ( notches ) {
				double zoom = g_scene->getZoom();
				double s = sin(Rf.angle());
				double ae = 0.05 / (ro * zoom + controlLen);

				if ( a - ae < 0 || a + ae > 2 * PI ) {
					a = 2 * PI;
				}
				else if ( fabs(Rf.x + seg[idx].C.x) * zoom < 0.05 * fabs(s) ) {
					Rf.x = -seg[idx].C.x;
					Rf.y = sqrt(ro * ro - Rf.x * Rf.x) * (s < 0 ? -1 : 1);
					a = fmod(Rf.angle() - ao + 2*PI, 2*PI);
				}
				else {
					double aa = Rf.angle() / (PI / 4);
					double ia = floor(aa + 0.5);
					if ( fabs(aa - ia) < ae ){
						a = fmod(ia * (PI / 4) - ao + 2*PI, 2*PI);						
					}
				}
			}

			if ( curve[idx] < 0 )
				a -= 2 * PI;

			if ( fabs(a) < 0.0001 )
				a = (curve[idx] < 0) ? -2 * PI : 2 * PI;

			curve.refValue()[idx] = a;
			distance.refValue()[idx - 1] = abs(a * ro);
		}
	}
	else
	{
		if ( 0x8000 & GetKeyState(VK_LSHIFT) ) {
			float zoom = g_scene->getZoom();
			if ( abs(pos.x) * zoom < 0.05 ) {
				pos.x = 1e-4;			
			}
		}
		setPoint(idx, pos);
	}

	g_selNode->processFields();
	commonPostTug();
	houseKeep();
}

void CCLathe::addControlPoint( Vec3f pt ) {

	Vec3f vec = g_scene->getIViewMatrix() * pt;
   Vec2f pos = Vec2f(vec.x, vec.y);

	// don't add a control point if it is very close to an existing one
	for(int i = 0; i < seg.length; i++ ){
		if((seg[i].Pf - pos).length() < 0.0001 ){
			return;
		}
	}


   Stem* stem = getStemOwner();
	if (stem!=nullPtr){
		if (stem->bands.length == 0){
			stem->sync.refValue() = true;
		}
	}

	int idx = max(0, index[0]);
	int oldLen = curve.length;
	int newLen = max(2, oldLen + 1);

   curve.resize(newLen);
   curve.set(&curve);
   distance.resize(newLen - 1);
   distance.set(&distance);

   if ( (idx < oldLen - 1) && (0x8000 & GetKeyState(VK_LSHIFT)) ) {
      idx++;
      for ( int i = newLen; --i > idx; )
         curve.refValue()[i] = curve[i - 1];
      for ( int i = newLen - 1; --i > idx - 1; )
         distance.refValue()[i] = distance[i - 1];
   }
   else
   {
   	idx = newLen - 1;

      autoClose = false;
      if ( oldLen == 0 )
      {
         curve.refValue()[0] = - PI / 2;
      }
   }
	index[0] = idx;
   setPoint(idx, pos);

	if ( stem && stem->sync == true ){
		int stemIndex = max(0, idx - 1);
		stem->index[0] = stemIndex;

		M_Int& bands = stem->bands;
		getBandAndYDist(bands, Stem::bandDist, stem->yDistance);
		bands.length = stem->yDistance.length;

      for ( int i = bands.length; --i > stemIndex; ) {
         bands[i] = bands[i - 1];
      }
		int lastVal = 5;
		for (int i = 0; i < bands.length; i++){
			if (i == stemIndex || bands[i] == 0) {
				bands[i] = lastVal;
			}
			lastVal = bands[i];
		}
	}

	stem->convertYDistanceToHeight();
	processFields();
	commonPostTug();
	updateBuild3dWindow();
}

void CCLathe::deleteControlPoint()
{
	int oldLen = curve.length;
	if ( oldLen == 0 )
	{
		return;
	}

	int newLen = oldLen > 2 ? oldLen - 1 : 0;
	int idx = min(index[0], oldLen - 1);

	if ( (idx > 0) && (idx < oldLen - 1) ) {
		for ( int i = idx; i < oldLen - 1; i++ )
			curve.refValue()[i] = curve[i + 1];
		for ( int i = idx - 1; i < oldLen - 2; i++ )
			distance.refValue()[i] = distance[i + 1];
	}
	else {
		idx = max(0, newLen - 1);
	}

	curve.resize(newLen);
	curve.set(&curve);
	distance.resize(max(0, newLen - 1));
	distance.set(&distance);

	Stem* stem = getStemOwner();
	if ( stem && stem->sync == true ){
		int stemIndex = max(0, index[0] - 1);
		stem->index[0] = max(0, stemIndex - 1);
		M_Int& bands = stem->bands;

		if ( stemIndex < bands.length - 1 ) {
			for ( int i = stemIndex; i < bands.length - 1; i++ ) {
				bands[i] = bands[i + 1];
			}
		}

		getBandAndYDist(bands, Stem::bandDist, stem->yDistance);
		bands.length = stem->yDistance.length;
	}

	index[0] = idx;
	stem->convertYDistanceToHeight();
	processFields();
	commonPostTug();
	updateBuild3dWindow();
}


Art* RingBuild::getSelPart(){

	Art* prt = part[0];
	int idx = index[0];

	for(int i = 0, r = 0; i < ownership.length; i++){
		if(i < ownership.length){
			r+=ownership[i];
			if(i<part.length){
				prt = part[i];
			}
		}
		if(r >= idx){
			break;
		}
	}
	return prt;
}

void RingBuild::senseMirrorRing(){

	Art* part = RingBuild::getSelPart();
	Seamless* smls = part->smls;
	M_Vec3f& coord = part->coord.refValue();

	if(part->mirror && part->mirrorPart == 0){
		int idx = index[0];
		if(quadrant == 1){
			scaleX.refValue()[idx].y = scaleX.refValue()[idx].x;
		}
		else if(quadrant == 3){
			scaleX.refValue()[idx].x = scaleX.refValue()[idx].y;
		}
	}
}

void RingBuild::drag(Vec3f pt){
	M_Vec2f& scaleX=this->scaleX.refValue();
	M_Vec2f& scaleZ=this->scaleZ.refValue();
	M_Rotation& orientation=this->orientation.refValue();
	M_Vec3f& location=this->location.refValue();
	const Rotation& rotation=this->rotation.refValue();
	const Vec3f& translation=this->translation.refValue();
	const Vec3f& scale = this->scale.refValue();

	int i=index[0];

	Matrix matrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * 
	scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * g_scene->getViewMatrix();


	Vec3f pivotPoint = matrix * Vec3f();

	Vec3f controlPoint;
	if(quadrant==4){
		Matrix pivotPointIMatrix = translationMatrix(Vec3f(-translation.x, -translation.y, -translation.z)) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z));
		Vec3f newPt=g_scene->getIViewMatrix() * pivotPointIMatrix * pt;

		if(dragLocationY){
			location[i].y = newPt.y;
		}
		if(dragLocationXZ){
			location[i].x = newPt.x;
			location[i].z = newPt.z;
		}
	}
	else{
		Vec3f vec;
		float* ptr=0;
		switch(quadrant){
			case 0:
				ptr=&scaleZ[i].x;
				controlPoint = matrix * Vec3f(0, 0, -*ptr);
				vec = Vec3f(0, 0, -*ptr);
				break;
			case 1:
				ptr=&scaleX[i].x;
				controlPoint = matrix * Vec3f(-*ptr, 0, 0);
				vec = Vec3f(-*ptr, 0, 0);
				break;
			case 2:
				ptr=&scaleZ[i].y;
				controlPoint = matrix * Vec3f(0, 0, *ptr);
				vec = Vec3f(0, 0, *ptr);
				break;
			case 3:
				ptr=&scaleX[i].y;
				controlPoint = matrix * Vec3f(*ptr, 0, 0);
				vec = Vec3f(*ptr, 0, 0);
				break;
		}
		Vec3f loc = this->location.refValue()[i];
		Rotation orien = this->orientation.refValue()[i];

		Matrix imatrix=translationMatrix(Vec3f(-translation.x, -translation.y, -translation.z)) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z)) *
			translationMatrix(Vec3f(-loc.x, -loc.y, -loc.z)) * rotationMatrix(-orien);

		Vec3f newPt=g_scene->getIViewMatrix() * imatrix * pt;

		*ptr = sqrt(newPt.x*newPt.x + newPt.y*newPt.y + newPt.z*newPt.z);

//		{
//			Matrix baseMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * g_viewMatrix;
//			Matrix matrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * baseMatrix;

		//}

		senseMirrorRing();


		if(dragOrientation){

			//would there not be an easier way to calculate the orientation with the information we have at this point in time?

			Vec3f toDiff = Vec3f(pt.x-pivotPoint.x, pt.y-pivotPoint.y, pt.z-pivotPoint.z);
			Vec2f fromDiff = Vec2f(controlPoint.x-pivotPoint.x, controlPoint.y-pivotPoint.y);
			float toAngle = Vec2f(toDiff.x, toDiff.y).getAngle();
			float fromAngle = fromDiff.getAngle();
			float angle = toAngle - fromAngle;

			Rotation unwind = -rotation;
			Vec3f axis = (-g_scene->rotation * unwind * Vec3f(0, 0, 1)).normalize();
			orientation[i] = orientation[i] * Rotation(axis, angle);
		}
		Vec3f sr;
		switch(quadrant){
			case 0:
				sr= Vec3f(0, 0, -scaleZ[i].x);
				break;
			case 1:
				sr= Vec3f(-scaleX[i].x, 0, 0);
				break;
			case 2:
				sr= Vec3f(0, 0, scaleZ[i].y);
				break;
			case 3:
				sr= Vec3f(scaleX[i].y, 0, 0);
				break;
		}
		Matrix postMatrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * 
		scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * g_scene->getViewMatrix();
		g_pointer.translation = postMatrix * sr;
	}
	houseKeep();
	commonPostTug();
}

void Art::drag(Vec3f pt){
	if(PartAnimClue* partAnimClue = dynamic_cast<PartAnimClue*>(g_visualClue)){
		partAnimClue->dragPart(pt);
	}
}

extern bool tugFlag2;
void openTransformCP();


void release3dRightDrag(){
#ifdef PROG_ROB
	g_storeCrossPoint.active = false;
	dLog("new Move("+ String(g_mousePos.x) + ',' + String(g_mousePos.y) + "),");
	dLog("new RightButtonUp(),");
#endif

	disablePen();
	g_viewZRotDragMode = false;
	g_viewRotDragMode = false;
	g_radiusControl = false;
	g_winAPI3DMode = false;

	ReleaseCapture();
}

void release3dDrag(){


#ifdef PROG_ROB
	dLog("new Move("+ String(g_mousePos.x) + ',' + String(g_mousePos.y) + "),");
	dLog("new LeftButtonUp(),");
#endif
	disablePen();
	ReleaseCapture();
	static String str;
	g_winAPI3DMode = false;
	if(MultiSelect::selectRectMode){
		if(shiftKeyState()){
			if(controlKeyState()){
			}
			else{
			}
		}
	}
	if(g_drawNurbsLine->mousePhase == DrawNurbsLine::DRAW){
		if(tugFlag && g_dragActuallyHappened){
			//g_vcEllipticalMaterial->emissiveColor = Color3f(0,1,1);
			//g_vcEllipticalMaterial->diffuseColor = Color3f(0,1,1);

			g_drawNurbsLine->mousePhase = DrawNurbsLine::READY_TO_DRAG;
			
			//openSceneTree();

			g_drawNurbsLine->convertDrawnLineToNurbsLine();

			//g_stw->cp->setupNodeWindow(g_drawNurbsLine);
			//Window::activeWindow = g_stw->cp.p;
		}
		else{
			resetTemporaryModes();
			MultiSelect::reset();
			changeModellingCursor(g_crossCursor);
			g_visualClue = nullPtr;
		}
	}
	else if(g_dragActuallyHappened){
		bool backup = true;

		if(NurbsLathe* lathe = dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
			lathe->postTug();
		}
		else if(Lock* lock = dynamic_cast<Lock*>(g_selNode.getRaw())){
		}
		else if(Build* build = dynamic_cast<Build*>(g_selNode.getRaw())){
			if(NurbsPatch* patch = build->getPatch()){
				if(g_transformCP->selectObject || dynamic_cast<NurbsPatch*>(build)){
					backup = patch->postTug();
				}
				//}
				//else{
				//	dLog("new TugVertex(" + String(patch->index[0]) + "," +String(g_pointer.translation.x)+','+String(g_pointer.translation.y) + "),");
				//}
			}
		}
		else if(Art* art = dynamic_cast<Art*>(g_selNode.getRaw())){
			Seamless* smls = art->smls;
			if(Window::animMode()){
			}
		}
		else if(RingBuild* ringBuild = dynamic_cast<RingBuild*>(g_selNode.getRaw())){
			int i = ringBuild->index[0];
		}

		if(Window::getAnimBar()){
			if(AnimLever* animClue = dynamic_cast<AnimLever*>(SenseTugGlobals::node)){
				Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
				if(g_dragActuallyHappened){
					if(g_partCageClue->mode == PartCageClue::BONE_PIVOT_POINT){
						if(Anim::autoCapture){
							if(part->ik){
								animClue->anim->captureSinglePartPose(part, &part->ikPos);
							}
							else{
								animClue->anim->captureSinglePartPose(part, &part->translation);
							}
						}
						g_cursorOverPointIlsBase->translation = (g_pointer.translation + g_transformCP->mouseOffset);
						part->houseKeep();
					}
					else{
						if(Anim::autoCapture){
							if(part->ik){
								animClue->anim->captureSinglePartPose(part, &part->ikRot);
							}
							else{
								animClue->anim->captureSinglePartPose(part, &part->rotation);
							}
						}
					}
					Window::paintAllControlPanels();
				}
			}
			if(tugFlag == true){
				Seamless* smls = 0;
				if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
					smls = part->smls;
				}

				if(AnimLever* animClue = dynamic_cast<AnimLever*>(g_visualClue)){
					if(smls->animCP == false && smls->workAnimCP == false && g_dragMorphFlag && g_dragActuallyHappened && !(controlKeyState()) && !(altKeyState())){
	//								smls->genDynamicMatricesAndPopulate();
						animClue->anim->capturePolyMorph(smls);
						tugFlag2 = tugFlag = false;
					}
					else if(Part* part = dynamic_cast<Part*>(g_selNode.getRaw())){
						Seamless* smls = part->smls;
						if(smls->workAnimCP && g_partCageClue->mode == PartCageClue::MORPH_CP){
							animClue->anim->captureMorph(smls);
						}
					}
				}
				else if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
					tugFlag2=tugFlag = false;
					//if(g_dragActuallyHappened && part->smls->build.length){
					//	g_joinActive = true;
					//	part->smls->doBuild();
					//	int idx = part->index[0];
					//	if(idx >= part->coord.length){
					//		idx = part->coord.length - 1;
					//		MultiSelect::coordList.length = 0;
					//		if(idx < 0){
					//			idx = 0;
					//		}
					//		part->index[0] = idx;
					//	}
					//}
					//g_dragActuallyHappened = false;
					//part->smls->ctrf = true;
				}
				if(backup){
					BackUp::monostable.beginCycle();
				}
			}
		}
		if(g_scene->mode & SINGLE_ACTIVE_MODE){
			g_transformCP->refreshRemPoints();
		}
	}
	tugFlag = false;
	g_partCageClue->mode = 0;
	orientationTugFlag = false;
	dragPartFlag = false;
	g_mirrorModeVertexFound = false;
	mouse3dMode = false;
	release3dDragFinal();
	g_dragActuallyHappened = false;
	//MultiSelect::withinSights = false;
	MultiSelect::selectRectMode = false;
	showCursor(true);
	g_stw->cp->paint();
}

extern Vec2i g_newMousePos;

void dragZoom(float beginZoom){
	float diff = g_newMousePos.y - g_beginMousePos.y;
	diff = diff * .002;

	if(diff > 0){
		g_scene->setZoom(beginZoom * (1.0 + diff));
	}
	else{
		g_scene->setZoom(beginZoom / (1.0 - diff));
	}

	houseKeepVisualClue();
}

void doRadiusControl(){
	g_dragActuallyHappened = true;
	float diff = g_newMousePos.y - g_beginMousePos.y;
	diff = diff * -.002 * (1 / g_scene->getZoom());

	float rad = abs(g_beginRadius + diff);

	if(g_scene->mode & MODELLING_MODE){
		g_radial_Tug->radius = rad;
	}
	else if(g_scene->mode & (PAINT_MODE | BRUSH_MODE)){
		g_brush->radius = rad;
	}

	houseKeepVisualClue();
}

void moveMouse3d(){
	static float beginZoom = 0;

	if(g_mousePos == g_prevMousePos){
		return;
	}
	Vec2i mouseRel;

	Seamless* smls = 0;
	mouse2dSync();
	Art* part = dynamic_cast<Art*>(g_selNode.getRaw());

	if(g_radiusControl){
		doRadiusControl();
		return;
	}
	if(g_zoomSpin){
		mouseRel = g_newMousePos - g_beginMousePos;
		if(abs(mouseRel.x) > 2 || abs(mouseRel.y) > 2){
			if(abs(mouseRel.x) > abs(mouseRel.y)){
				g_viewZRotDragMode = true;
			}
			else{
				beginZoom = g_scene->getZoom();
				g_leftButtonZoom = true;
			}
			g_zoomSpin = false;
		}
		else{
			return;
		}
	}
	if(g_leftButtonZoom){
		dragZoom(beginZoom);
	}
	else if(g_viewRotDragMode || g_viewZRotDragMode){
		//if(part!=0 && shiftKeyState() && !(controlKeyState())){
		//	part->rotation.refValue()*=Rotation(0,1,0, PI*.01*(float)dims.lX)*Rotation(1,0,0, PI*.01*(float)dims.lY);
		//	part->rotation=part->rotation;
		//}
		//else {
			g_scene->rotateScene();
		//}
			g_vcArtAnimBase->calcRotCtrlPnts();
			g_vcArtAnimBase->houseKeep();
			//if(MultiSelect::pointList.length >= g_transformCP->leverSelection){
				//g_transformCP->calcScaleOrientation();
				//g_transformCP->calcRotCtrlPnts();
			//}
			//g_transformCP->houseKeep();
	}
	else{ 
		MouseFlag = true;

		if(MultiSelect::selectRectMode){
			MultiSelect::bottomRight = g_pointer.translation;
			g_viewDragMousePos = g_pointer.translation;

			houseKeepVisualClue();
		}
		//else if(g_viewPosDragMode && !(controlKeyState())){
		else if(g_touchViewPosDragMode){
			g_touchViewPosDragMode = false;
			Vec3f pointer = mouse2dSync(g_newMousePos);
			//pointer.z = g_scene->translation.z;
			//g_scene->translation = pointer;

			//Vec3f v = g_touchBeginViewPos + pointer - g_touchBeginMousePos;
			Vec3f v = g_touchBeginViewPos + -g_scene->rotation * (pointer - g_touchBeginMousePos) * (1.0f / g_scene->getZoom());
			v.z = g_scene->translation.z;
			g_scene->translation = v;
		}
		else if(g_viewPosDragMode){


			if(g_pointer.translation.y < -1 || g_pointer.translation.y > 1){
				int a = 0;
			}
			if(g_pointer.translation.x < -1 || g_pointer.translation.x > 1){
				int a = 0;
			}

			g_scene->translation = g_beginViewPos + -g_scene->rotation * (g_pointer.translation - g_mouseFrom) * (1.0f / g_scene->getZoom());
						
			g_viewDragMousePos = g_pointer.translation;

			g_vcArtAnimBase->calcRotCtrlPnts();
			g_vcArtAnimBase->houseKeep();
			if(g_scene->mode & SELECT_OBJECT_MODE){
				g_transformObj->calcRotCtrlPnts();
				g_transformObj->houseKeep();
			}
			else if(MultiSelect::pointList.length >= g_transformCP->leverSelection){
				g_transformCP->calcRotCtrlPnts();
			}
			g_transformCP->houseKeep();
		}
		else if(g_scene->mode & PAINT_MODE && g_scene->mode & BRUSH_MODE && SenseTugGlobals::actTriPart){
			smls = part->smls;
			Art* actTriPart = SenseTugGlobals::actTriPart;
			float a = g_mouseSense * .25 / g_scene->scale.x;
			float b = g_mouseSense * .25;

			POINT ptCursor;
			GetCursorPos(&ptCursor);
			ScreenToClient( hWndMain, &ptCursor);
			g_mousePos.x = ptCursor.x;
			g_mousePos.y = ptCursor.y;

			//g_mousePos.y += dims.lY;
			//g_mousePos.x += dims.lX;

			//g_pointer.translation.y -= dims.lY * g_mouseSense * .25;
			//g_pointer.translation.x += dims.lX * g_mouseSense * .25;


//						Vec3f v = Vec3f(dims.lX*b, -dims.lY*b, 0);
			//Vec3f v = g_pointer.translation;
			//MultiSelect::bottomRight = v;
//						Patch::bottomRight+=Vec3f(dims.lX*b, -dims.lY*b, 0);

			//g_viewDragMousePos+=Vec3f(dims.lX * b, -dims.lY * b, 0);
			//Vec2i vec=get2dMousePosFrom3d(g_viewDragMousePos);
			//WINDOWINFO wi;
			//GetWindowInfo(hWndMain, &wi);
			//SetCursorPos(vec.x + wi.rcClient.left, vec.y + wi.rcClient.top); 
			houseKeepVisualClue();

			if(dynamic_cast<ColorEffect*>(actTriPart->workEffect)){
				smls->colourWithBrush(part);
			}
			else if(TextureEffect* textureEffect = dynamic_cast<TextureEffect*>(actTriPart->workEffect)){
				if(Window::animMode()){
					MultiSelect::clickedIndex = 0;
				}
				SenseTugGlobals::node = 0;
				g_scene->senseTug(Matrix(),Matrix(), 0);

				//if(g_brush->smudge){
				//	textureEffect->getColorWithEyeDropRadius();
				//}
						
				textureEffect->writePixelsToTexture(g_palette->color);
			}
		}
		else if(!MultiSelect::selectRectMode){
			//if(VisualAid* visualAid = dynamic_cast<VisualAid*>(g_selNode.getRaw())){
			if(VisualAid* visualAid = dynamic_cast<VisualAid*>(MultiSelect::withinSightsNode.p)){
				if(g_transformCP->selectObject){
					if(Build* bld = dynamic_cast<Build*>(visualAid)){
						if(NurbsPatch* patch = bld->getPatch()){
							patch->drag(g_pointer.translation);
						}
					}
				}
				//else if((!(controlKeyState())) || dynamic_cast<NurbsPatch*>(visualAid) || dynamic_cast<PartCageClue*>(visualAid)){
				else if((!(controlKeyState())) || g_transformCP->dragRotPointMode){
					visualAid->drag(g_pointer.translation);
				}
			}
		}
	}
}