/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include <windows.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "ms-gui.h"
#include "scene_tree.h"
#include "robot_aid.h"

extern bool g_dontContinueButtonAction;

extern const char* msg_can_not_unfix_first_or_last_control_point;

template<const int n> void upDownButtonPaint(const Button* button, const int command, const int baseCommand, const Vec2i size, const Vec2i dstPos, const Vec2i srcPos, const Vec2i iconSize, Window& win);

template bool SFBoolButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);

FieldButton* SFBool::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	if(properties.push() || properties.link()){
		int strx = properties.name.length * 8;
		xPos = colum - strx + 10;
		if(xPos < 6 || properties.link()){
			xPos = 6;
		}
		return new SFBoolButton(Vec2i(strx + 16, 24), Vec2i(xPos, yPos+1), Vec2i(7*24, 24), properties.tip, 0, win, node, *this, properties, Vec2i(xPos+4, yPos), Vec2i(16,16));
	}
	else{
		return new SFBoolButton(Vec2i(12, 12), Vec2i(xPos, yPos+8), Vec2i(403, 0), properties.tip, 0, win, node, *this, properties, Vec2i(xPos+4, yPos), Vec2i(16,16));
	}
}
FieldButton* MFBool::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	return new MFBoolButton(Vec2i(12, 12), Vec2i(xPos, yPos+8), Vec2i(403, 0), properties.tip, 0, win, node, *this, properties, Vec2i(xPos+4, yPos), Vec2i(16,16));
}
FieldButton* MFColor::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*1+4;
	return new MFColorButton(Vec2i(18*3, 24), Vec2i(xPos, yPos), Vec2i(18*3, 0), properties.tip, 0, win, node, *this, properties, Vec2i(xPos+18*3+4, yPos), Vec2i(win.size.x-valXPos, 16));
}
FieldButton* MFFloat::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos + 18 * 1 + 4;
	return new MFFloatButton(Vec2i(18, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos, 16));
}
FieldButton* MFImage::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	throw this;
}
FieldButton* MFInt32::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*1+4;
	return new MFInt32Button(Vec2i(18, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos, 16));
}
FieldButton* SFNode::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	return 0;
}
FieldButton* MFNode::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	return 0;
}
FieldButton* MFRotation::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*3+4;
	return new MFRotationButton(Vec2i(18*3, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos, 16));
}
FieldButton* MFString::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+4;
	return new MFStringButton(Vec2i(), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos, 16));
}
FieldButton* MFTime::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*1+4;
	return new MFDoubleButton(Vec2i(18, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos, 16));
}
FieldButton* MFVec2f::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*2+4;
	return new MFVec2fButton(Vec2i(18*2, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos, 16));
}
FieldButton* MFVec3f::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*3+4;
	return new MFVec3fButton(Vec2i(18*3, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos, 16));
}

//void M_Route::setUpRoute( int& yPos){
//	yPos+=length.Length*pixSymbolCY;
//}
FieldButton* SFFloat::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*1+4;
	return new SFFloatButton(Vec2i(18, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos,16));
}
FieldButton* SFInt32::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*1+4;
	return new SFInt32Button(Vec2i(18, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos,16));
}
FieldButton* SFImage::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	throw;
}
FieldButton* SFTime::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*1+4;
	return new SFDoubleButton(Vec2i(18, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos,16));
}
FieldButton* SFString::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+4;
	return new SFStringButton(Vec2i(), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos,16));
}
FieldButton* SFColor::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*3+4;
	return new SFColorButton(Vec2i(18*3, 24), Vec2i(xPos, yPos), Vec2i(18*3, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos,16));
}
FieldButton* SFVec2f::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*2+4;
	return new SFVec2fButton(Vec2i(18*2, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos,16));
}
FieldButton* SFVec3f::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*3+4;
	return new SFVec3fButton(Vec2i(18*3, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos,16));
}
FieldButton* SFRotation::setUpControl(Node& node, Field& field, const Properties& properties, const int index, int xPos, int yPos, Window& win){
	const int valXPos=xPos+18*3+4;
	return  new SFRotationButton(Vec2i(18*3, 24), Vec2i(xPos, yPos), Vec2i(0, 0), properties.tip, 0, win, node, *this, properties, Vec2i(valXPos, yPos), Vec2i(win.size.x-valXPos,16));
}
bool g_update=false;
bool SFImageButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	throw this;
}
bool SFStringButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	if(bField!=textCellStr){
		bField=textCellStr;
		return true;
	}
	return false;
}

bool MFImageButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	throw;
}
template <class T> bool FB<T, Single>::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
//bool updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	int i=0;
	T value; textCellStr.toValue(value);
	if(notCloseEnough(bField,value)){
		bField=value;
		return true;
	}
	return false;
}
template <class T> bool FB<T, Multiple>::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	if(!bField.length)return false;
	T sval; textCellStr.toValue(sval);
	if(notCloseEnough(bField.refValue()[index],sval)){
		M_<T> m = bField;
		m[index] = sval;
		bField = m;
		return true;
	}
	return false;
}

//template bool SFInt32Button::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
template <> bool SFInt32Button::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	int value; textCellStr.toValue(value);
	if(value>=properties.rangeFrom&&value<=properties.rangeTo){
		if(bField.getValue()!=value){
			bField=value;
			return true;
		}
	}
	return false;
}
bool MFInt32Button::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	if(!bField.length)return false;
	int sval; textCellStr.toValue(sval);
	if(sval>=properties.rangeFrom&&sval<=properties.rangeTo){
		if(bField.refValue()[index]!=sval){
			bField.refValue()[index]=sval;
			bField=bField.refValue();
			return true;
		}
	}
	return false;
}
bool MFStringButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties){
	if(!bField.length)return false;
	if(bField.refValue()[index]!=textCellStr){
		bField.refValue()[index]=textCellStr;
		bField=bField.refValue();
		return true;
	}
	return false;
}
template bool SFBoolButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool SFDoubleButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool SFFloatButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool SFVec3fButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool SFColorButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool SFRotationButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool SFVec2fButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool SFVec3fButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);

template bool MFBoolButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool MFDoubleButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool MFFloatButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool MFVec3fButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool MFColorButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool MFRotationButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool MFVec2fButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);
template bool MFVec3fButton::updateValFromTextCell(String textCellStr, const int &index, const Properties& properties);


bool outOfSyncWithRobot(){//if the robot was syncronised with the 3d rendering we would not need this
	if(InputEvent::robotActive){
		if(Button::repeatCounter>=InputEvent::syncNumOfActions){
			dLog("Out 2Of Sync!!!");
			return true;
		}
	}
	Button::repeatCounter++;
	return false;
}

void SFBoolButton::action(){
	if(Viewpoint* view = dynamic_cast<Viewpoint*>(&this->node)){
		if(&view->live == &this->bField){
		//if(sprop.name=="live"){
			if((view->track && g_scene->mode & ANIMATION_MODE)){
				if(Anim* anim = Window::animMode()){
					if(anim->play == false && Anim::autoCapture){
						anim->captureCameraLive(view);
						view->live = true;
						g_dontContinueButtonAction = false;
						removeTypeFromList < Button>(g_callBackList);
						removeTypeFromList < DownButton>(g_callBackList);
						g_stw->commonPostInsertNodeStuff();
						return;
					}
				}
			}
		}
	}
	bField = !bField.refValue();
	if(g_dontContinueButtonAction == false){
		updateBuild3dWindow();
	}
	g_dontContinueButtonAction = false;

}

void quantizeIncColor(float& v){
	v+=colorIncrement;
	v = floorf(v/colorIncrement+colorIncrement*.5) * colorIncrement;
	if(v > 1){
		v = 1;
	}
}
float quantize(float v, float res){
	return floorf(v / res + .5) * res;
}
void quantizeDecColor(float& v){
	v -= colorIncrement;
	v = floorf(v / colorIncrement + colorIncrement * .5) * colorIncrement;
	if(v < 0){
		v = 0;
	}
}

void quantizeInc(float& v){
	float res = getIncDecRes();
	v+=res;
	v = floorf(v/res+res*.5) * res;
}

void quantizeDec(float& v){
	float res = getIncDecRes();
	v-=res;
	v = floorf(v/res+res*.5) * res;
}

void quantizeInc(double& v){
	float res = getIncDecRes();
	v+=res;
	v = floor(v/res+res*.5) * res;
}
void quantizeDec(double& v){
	float res = getIncDecRes();
	v-=res;
	v = floor(v/res+res*.5) * res;
}
/*

inline void getXAngle(Rotation& r, Vec3f& c1, Vec3f& c2, Rotation& rc){
	float xAngle = 0;
	if(fabs(c2.y)>.001 || fabs(c2.z)>.001){
		xAngle = -atan2( c2.y, c2.z);
	}
	else{
		int a = 5;
	}
	rc = Rotation(1, 0, 0, xAngle);
}
inline void getZAngle2(Rotation& r, Vec3f& c1, Vec3f& c2, Rotation& rc){
	float distance1=sqrt(c1.x*c1.z + c1.y*c1.y);
	float distance2=sqrt(c2.x*c2.z + c2.y*c2.y);
	float zAngle=0;
	if(distance1 > distance2){
		zAngle = atan2( c1.y, c1.x);
	}
	else{
		int a = 5;
	}
	rc = Rotation(0, 0, 1, zAngle);
}
inline void getYAngle2(Rotation& r, Vec3f& c1, Vec3f& c2, Rotation& rb, Rotation& rc){
	float yAngle = 0;
	if(fabs(c1.x)>.001 || fabs(c1.z)>.001){
		yAngle = atan2( c2.x, c2.z);
	}
	else{
		int a = 5;
	}
	rb = Rotation(0, 1, 0, -yAngle);
	c1 = rb * c1;
	c2 = rb * c2;
	rb = Rotation(0, 1, 0, yAngle);
	getZAngle2(r, c1, c2, rc);
}
inline void getYAngle(Rotation& r, Vec3f& c1, Vec3f& c2, Rotation& rb, Rotation& rc){
	float yAngle = 0;
	if(fabs(c1.x)>.001 || fabs(c1.z)>.001){
		yAngle = -atan2( c1.z, c1.x);
	}
	else{
		int a = 5;
	}
	rb = Rotation(0, 1, 0, -yAngle);
	c1 = rb * c1;
	c2 = rb * c2;
	rb = Rotation(0, 1, 0, yAngle);
}
inline void getZAngle(Rotation& r, Vec3f& c1, Vec3f& c2, Rotation& ra, Rotation& rb, Rotation& rc){
	float distance1=sqrt(c1.x*c1.x + c1.y*c1.y);
	float distance2=sqrt(c2.x*c2.x + c2.y*c2.y);
	if(distance1 > distance2){
		float zAngle = atan2( c1.y, c1.x);
		ra = Rotation(0, 0, 1, -zAngle);
		c1 = ra * c1;
		c2 = ra * c2;
		ra = Rotation(0, 0, 1, zAngle);
		getYAngle(r, c1, c2, rb, rc);
	}
	else{
		float zAngle = atan2( c2.y, c2.x) + PI;
		ra = Rotation(0, 0, 1, -zAngle);
		c1 = ra * c1;
		c2 = ra * c2;
		ra = Rotation(0, 0, 1, zAngle);
		getYAngle2(r, c1, c2, rb, rc);
	}
}
void quantiseRotation2(Rotation& r){
	Rotation ra;
	Rotation rb;
	Rotation rc;
	Vec3f c1 = r * Vec3f(1,0,0);
	Vec3f c2 = r * Vec3f(0,0,1);
	float zAngle = 0;
	float yAngle = 0;
	float xAngle = 0;

	getZAngle(r, c1, c2, ra, rb, rc);

	r = r * -ra * -rb * -rc;
}
*/
void commonXYgetPlane(Vec3f& b, Vec3f& c, Rotation& ra, Rotation& rb, Rotation& rc){
	rb = Rotation(0, 1, 0, atan2(b.z, b.x));
	b = rb * b;
	c = rb * c;
	rb.angle = -rb.angle;

	rc = Rotation(1, 0, 0, atan2(c.y, c.z));
	b = rc * b;
	c = rc * c;
	rc.angle = -rc.angle;
}
inline void getRotRelToXPlane(Vec3f& b, Vec3f& c, Rotation& ra, Rotation& rb, Rotation& rc){
	ra = Rotation(0, 0, 1, atan2(b.x, b.y)-PI*.5);
	b = ra * b;
	c = ra * c;
	ra.angle = -ra.angle;

	commonXYgetPlane(b, c, ra, rb, rc);
}
inline void getRotRelToYPlane(Vec3f b, Vec3f c, Rotation& ra, Rotation& rb, Rotation& rc){
	ra = Rotation(0, 1, 0, atan2(b.z, b.x));
	b = ra * b;
	c = ra * c;
	ra.angle = -ra.angle;

	rb = Rotation(0, 0, 1, -atan2(b.y, b.x));
	b = rb * b;
	c = rb * c;
	rb.angle = -rb.angle;

	rc = Rotation(1, 0, 0, atan2(c.y, c.z));
	b = rc * b;
	c = rc * c;
	rc.angle = -rc.angle;

}
inline void getRotRelToZPlane(Vec3f b, Vec3f c, Rotation& ra, Rotation& rb, Rotation& rc){
	ra = Rotation(0, 0, 1, -atan2(b.y, b.x));
	b = ra * b;
	c = ra * c;
	ra.angle = -ra.angle;

	commonXYgetPlane(b, c, ra, rb, rc);
}

void getRotationsForTriangleRelToYPlane(Vec3f a, Vec3f b, Vec3f c, Rotation& ra, Rotation& rb, Rotation& rc){
	//translate tri so a == 0 0 0
	b = b - a;
	c = c - a;
	//a = Vec3f() not needed but here as a comment for concept;

	//find clossest plane
	float mostX = fabs(b.x); 
	float mostY = fabs(b.y); 
	float mostZ = fabs(b.z); 
	if(mostX < fabs(c.x)) mostX = fabs(c .x);
	if(mostY < fabs(c.y)) mostY = fabs(c .y);
	if(mostZ < fabs(c.z)) mostZ = fabs(c .z);

	if(mostX < mostY && mostX < mostZ){
		getRotRelToXPlane(b, c, ra, rb, rc);
	}
	else if(mostY < mostZ){
		getRotRelToYPlane(b, c, ra, rb, rc);
	}
	else{
		getRotRelToZPlane(b, c, ra, rb, rc);
	}
}
void quantiseRotation(Rotation& r){
	Rotation ra;
	Rotation rb;
	Rotation rc;
	Vec3f b = r * Vec3f(1,0,0);
	Vec3f c = r * Vec3f(0,0,1);

	getRotationsForTriangleRelToYPlane(Vec3f(), b, c, ra, rb, rc);

	ra.angle = floor(ra.angle/(PI*.5)+PI*.25) * PI*.5;
	rb.angle = floor(rb.angle/(PI*.5)+PI*.25) * PI*.5;
	rc.angle = floor(rc.angle/(PI*.5)+PI*.25) * PI*.5;

	r = rc * rb * ra;
//	r = r * -ra * -rb * -rc;
}
Rotation Matrix::getRotation(){
	Rotation ra;
	Rotation rb;
	Rotation rc;
	Vec3f b = *this * Vec3f(1,0,0);
	Vec3f c = *this * Vec3f(0,0,1);

	getRotationsForTriangleRelToYPlane(Vec3f(), b, c, ra, rb, rc);

	return rc * rb * ra;
}
void SFColorButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	Color3f& v=bField.refValue();
	switch(command){
		case 0:
			quantizeDecColor(v.r);
			break;
		case 1:
			quantizeIncColor(v.r);
			break;
		case 2:
			quantizeDecColor(v.g);
			break;
		case 3:
			quantizeIncColor(v.g);
			break;
		case 4:
			quantizeDecColor(v.b);
			break;
		case 5:
			quantizeIncColor(v.b);
			break;
	}
	bField=v;
	updateBuild3dWindow();
}
void SFDoubleButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	double v=bField.refValue();

	switch(command){
		case 0:
			quantizeDec(v);
			break;
		case 1:
			quantizeInc(v);
			break;
	}
	bField=v;
	updateBuild3dWindow();
}
void SFFloatButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	float v=bField.refValue();
	switch(command){
		case 0:
			quantizeDec(v);
			if(v<sprop.rangeFrom){
				v=sprop.rangeFrom;
			}
			break;
		case 1:
			quantizeInc(v);
			if(v>sprop.rangeTo){
				v=sprop.rangeTo;
			}
			break;
	}
	bField=v;
	updateBuild3dWindow();
}
void SFInt32Button::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	int v=bField;
	switch(command){
		case 0:
			v--;
			break;
		case 1:
			v++;
			break;
	}
	if(v>=sprop.rangeFrom&&v<=sprop.rangeTo){
		bField=v;
		updateBuild3dWindow();
	}
}
void SFStringButton::action(){
	throw;
}
void SFVec2fButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	Vec2f v=bField.refValue();
	switch(command){
		case 0:
			quantizeDec(v.x);
			break;
		case 1:
			quantizeInc(v.x);
			break;
		case 2:
			quantizeDec(v.y);
			break;
		case 3:
			quantizeInc(v.y);
			break;
	}
	bField=v;
	updateBuild3dWindow();
}
void SFVec3fButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	Vec3f v = bField.refValue();
	switch(command){
		case 0:
			quantizeDec(v.x);
			break;
		case 1:
			quantizeInc(v.x);
			break;
		case 2:
			quantizeDec(v.y);
			break;
		case 3:
			quantizeInc(v.y);
			break;
		case 4:
			quantizeDec(v.z);
			break;
		case 5:
			quantizeInc(v.z);
			break;
	}
	bField=v;
	updateBuild3dWindow();
}
void sAction(Rotation& v, int command){
	float res = floatIncPIFrac;
	if(controlKeyState()){
		quantiseRotation(v);
		res = PI*.5;
	}
	switch(command){
		case 0:
			v*=Rotation(1,0,0,res);
			break;
		case 1:
			v*=Rotation(1,0,0,-res);
			break;
		case 2:
			v*=Rotation(0,1,0,res);
			break;
		case 3:
			v*=Rotation(0,1,0,-res);
			break;
		case 4:
			v*=Rotation(0,0,1,res);
			break;
		case 5:
			v*=Rotation(0,0,1,-res);
			break;
	}
}
void SFRotationButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	Rotation& v=bField.refValue();
	sAction(v, command);
	bField=v;
	updateBuild3dWindow();
}
void canNotUnfixMsg(){
	msg(msg_can_not_unfix_first_or_last_control_point);
}
bool Bezier::common_onActionForFix(Stem& stem, bool state){
	int idx=index[0];
	if(idx>0 && idx<fix.length-1){
		int fixSegBeginIndex=0;
		int groupIndex = getGroupIndices(fixSegBeginIndex);
		if(state==false){
			stem.bands.refValue().insertElement(groupIndex+1, 16);
			snapBorderTangentsForFirstCP(index[0]);
			return true;
		}
		else{
			stem.bands.refValue().deleteElement(groupIndex);
			return true;
		}
	}
	canNotUnfixMsg();
	return false;
}
bool BezierLatheBuild::common_onActionForFix(bool state){
	int idx=index[0];
	if(idx>0 && idx<fix.length-1){
		int fixSegBeginIndex=0;
		int groupIndex = getGroupIndices(fixSegBeginIndex);
		if(state==false){
			bands.refValue().insertElement(groupIndex+1, 16);
			snapBorderTangentsForFirstCP(index[0]);
			return true;
		}
		else{
			bands.refValue().deleteElement(groupIndex);
			return true;
		}
	}
	canNotUnfixMsg();
	return false;
}
bool Bezier::common_onActionForFix(bool state){
	int idx=index[0];
	if(idx>0 && idx<fix.length-1){
		int fixSegBeginIndex=0;
		int groupIndex = getGroupIndices(fixSegBeginIndex);
		if(state==false){
			snapBorderTangentsForFirstCP(index[0]);
			return true;
		}
		else{
			return true;
		}
	}
	canNotUnfixMsg();
	return false;
}
bool Bezier3f::common_onActionForFix(bool state){
	int idx=index[0];
	if(idx>0 && idx<fix.length-1){
		int fixSegBeginIndex=0;
		int groupIndex = getGroupIndices(fixSegBeginIndex);
		if(state==false){
			snapBorderTangentsForFirstCP(index[0]);
			return true;
		}
		else{
			return true;
		}
	}
	canNotUnfixMsg();
	return false;
}
bool BezierLathe::onActionForFix(bool state){
	Stem* stem=getStemOwner();
	if(stem && stem->sync==true){
		return common_onActionForFix(*stem, state);
	}
	else{
		return common_onActionForFix(state);
	}
}
bool BezierLatheBuild::onActionForFix(bool state){
	return common_onActionForFix(state);
}
bool BezierScale::onActionForFix(bool state){
	return common_onActionForFix(state);
}
bool BezierCircumScale::onActionForFix(bool state){
	return common_onActionForFix(state);
}
bool BezierShift::onActionForFix(bool state){
	return common_onActionForFix(state);
}
bool BezierTwist::onActionForFix(bool state){
	return common_onActionForFix(state);
}
bool BezierTape::onActionForFix(bool state){
	Stem* stem=getStemOwner();
	if(stem && stem->sync==true){
		return common_onActionForFix(*stem, state);
	}
	else{
		return common_onActionForFix(state);
	}
}
bool BezierBend::onActionForFix(bool state){
	return common_onActionForFix(state);
}
void MFBoolButton::action(){
	if(bField.length==0){
		return;
	}
	const int idx= node.getFieldGroupIndex(sprop.groupIndex);
	if(sprop.name=="fix"){
		if(Bezier* bezier=dynamic_cast<Bezier*>(&node)){
			if(bezier->onActionForFix(bField[idx])==false){
				return;
			}
		}
		else if(Bezier3f* bezier=dynamic_cast<Bezier3f*>(&node)){
			if(bezier->onActionForFix(bField[idx])==false){
				return;
			}
		}
	}
	bool& e=bField.refValue()[idx];
	e=!e;
	bField=bField.refValue();
	updateBuild3dWindow();
}
void MFRotationButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	if(bField.length==0){
		return;
	}
	int idx = node.getFieldGroupIndex(sprop.groupIndex)-sprop.startAtIndex;
	if(idx < 0){
		return;
	}
	Rotation& e=bField.refValue()[idx];

	sAction(e, command);

	bField=bField.refValue();
	updateBuild3dWindow();
}
void MFInt32Button::action(){
	if(bField.length==0){
		return;
	}
	if(outOfSyncWithRobot()){
		return;
	}
	int idx = node.getFieldGroupIndex(sprop.groupIndex)-sprop.startAtIndex;
	if(idx < 0){
		return;
	}
	M_Int t = bField;
	int& e=t[idx];

	switch(command){
		case 0:
			e-=1;
			if(e<sprop.rangeFrom){
				e=sprop.rangeFrom;
			}
			break;
		case 1:
			e+=1;
			if(e>sprop.rangeTo){
				e=sprop.rangeTo;
			}
			break;
	}
	bField=t;
	updateBuild3dWindow();
}
void MFDoubleButton::action(){
	if(bField.length==0){
		return;
	}
	if(outOfSyncWithRobot()){
		return;
	}
	int idx = node.getFieldGroupIndex(sprop.groupIndex)-sprop.startAtIndex;
	if(idx < 0){
		return;
	}
	double& e=bField.refValue()[idx];

	switch(command){
		case 0:
			quantizeDec(e);
			if(e<sprop.rangeFrom){
				e=sprop.rangeFrom;
			}
			break;
		case 1:
			quantizeInc(e);
			if(e>sprop.rangeTo){
				e=sprop.rangeTo;
			}
			break;
	}
	bField=bField.refValue();
	updateBuild3dWindow();
}
void MFFloatButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	if(bField.length==0){
		return;
	}
	int idx = node.getFieldGroupIndex(sprop.groupIndex)-sprop.startAtIndex;
	if(idx < 0){
		return;
	}
	float& e=bField.refValue()[idx];

/*
	Stem* stem=dynamic_cast<Stem*>(g_selNode.getRaw());
	if(stem && !(altKeyState())){
		if(stem->yDistance.length>1){
			int idx2 = stem->index[0];
			if(idx2 == stem->yDistance.length-1){
				idx2--;
			}
			else{
				idx2++;
			}
			float& e2=stem->yDistance.refValue()[idx2];
			float storeE = e;
			switch(command){
				case 0:
					quantizeDec(e);
					if(e<sprop.rangeFrom){
						e=sprop.rangeFrom;
					}
					break;
				case 1:
					quantizeInc(e);
					if(e>sprop.rangeTo){
						e=sprop.rangeTo;
					}
					break;
			}
			float diff = storeE - e;
			e2+=diff;
		}
	}
	*/
	switch(command){
		case 0:
			quantizeDec(e);
			if(e<sprop.rangeFrom){
				e=sprop.rangeFrom;
			}
			break;
		case 1:
			quantizeInc(e);
			if(e>sprop.rangeTo){
				e=sprop.rangeTo;
			}
			break;
	}
	bField=bField.refValue();
	updateBuild3dWindow();
}
void MFStringButton::action(){
}
void MFVec2fButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	if(bField.length==0){
		return;
	}
	int idx = node.getFieldGroupIndex(sprop.groupIndex)-sprop.startAtIndex;
	if(idx < 0){
		return;
	}
	M_Vec2f t = bField;
	Vec2f& e=t[idx];

	switch(command){
		case 0:
			quantizeDec(e.x);
			break;
		case 1:
			quantizeInc(e.x);
			break;
		case 2:
			quantizeDec(e.y);
			break;
		case 3:
			quantizeInc(e.y);
			break;
	}
	bField=t;
	updateBuild3dWindow();
}
void MFVec3fButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	float res = getIncDecRes();

	if(bField.length==0){
		return;
	}
	int idx = node.getFieldGroupIndex(sprop.groupIndex)-sprop.startAtIndex;
	if(idx < 0){
		return;
	}
	M_Vec3f t = bField;
	Vec3f& e = t[idx];

	if(sprop.axis()){
		switch(command){
			case 0:
				e=Rotation(-1, 0, 0, res)*e;
				break;
			case 1:
				e=Rotation(1, 0, 0, res)*e;
				break;
			case 2:
				e=Rotation(0, -1, 0, res)*e;
				break;
			case 3:
				e=Rotation(0, 1, 0, res)*e;
				break;
			case 4:
				e=Rotation(0, 0, -1, res)*e;
				break;
			case 5:
				e=Rotation(0, 0, 1, res)*e;
				break;
		}
		e=e.normalize();
	}
	else {
		switch(command){
			case 0:
				quantizeDec(e.x);
				break;
			case 1:
				quantizeInc(e.x);
				break;
			case 2:
				quantizeDec(e.y);
				break;
			case 3:
				quantizeInc(e.y);
				break;
			case 4:
				quantizeDec(e.z);
				break;
			case 5:
				quantizeInc(e.z);
				break;
		}
	}
	bField=t;
	updateBuild3dWindow();
}
void MFColorButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	if(bField.length==0){
		return;
	}
	int idx = node.getFieldGroupIndex(sprop.groupIndex)-sprop.startAtIndex;
	if(idx < 0){
		return;
	}
	Color3f& e=bField.refValue()[idx];

	switch(command){
		case 0:
			quantizeDecColor(e.r);
			break;
		case 1:
			quantizeIncColor(e.r);
			break;
		case 2:
			quantizeDecColor(e.g);
			break;
		case 3:
			quantizeIncColor(e.g);
			break;
		case 4:
			quantizeDecColor(e.b);
			break;
		case 5:
			quantizeIncColor(e.b);
			break;
	}
	bField=bField.refValue();
	updateBuild3dWindow();
}
void IndexButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	if(Art* part = dynamic_cast<Art*>(&node)){
		if(part->smls->workAnimCP && (Window::animMode() || g_colorOwnership)){
			M_ControlPoint& controlPoint = part->controlPoint;
			if(controlPoint.length){
				switch(command){
					case 0:
						MultiSelect::selectedPoint--;
						break;
					case 1:
						MultiSelect::selectedPoint++;
						break;
				}
				if(MultiSelect::selectedPoint <= 0){
					MultiSelect::selectedPoint = 0;
				}
				else if(MultiSelect::selectedPoint >= controlPoint.length){
					MultiSelect::selectedPoint = controlPoint.length-1;
				}
			}
			houseKeepVisualClue();
			win.paint();
			return;
		}
	}


	int index = node.getFieldGroupIndex(sprop.groupIndex);

	if(bField.size() && index >= (bField.size() + sprop.startAtIndex)){//safety code but why should we need it?
		index=bField.size() - 1 + sprop.startAtIndex;
		node.setFieldGroupIndex(sprop.groupIndex, index);
	}

	switch(command){
		case 0:
			index--;
			break;
		case 1:
			index++;
			break;
	}
	if(index<=0){
		index=0;
	}
	else if(index>=bField.size()){
		index=bField.size()-1;
	}
	node.setFieldGroupIndex(sprop.groupIndex,index);

	if(ALathe* lathe=dynamic_cast<ALathe*>(&node)){
		Stem* stem=getStemOwner();
		if(stem!=nullPtr&&stem->sync==true){
			int stemIndex=index-1;
			if(stemIndex<0){
				stemIndex=0;
			}
			stem->index[0]=stemIndex;
		}
	}
	if(MarkPart* markPart = dynamic_cast<MarkPart*>(g_selNode.getRaw())){
		markPart->checkToPaintSceneTree();
	}
	houseKeepVisualClue();
	win.paint();
}
void RingBuild::checkToPaintSceneTree(){
	int i = index[1];
	if(i < part.length){
		part[i].p->smls->ctrf=true;
		g_stw->paint();
	}
}

PaletteCP* paletteCPAllreadyOpen();

void TransferOwnership::checkToPaintSceneTree(){
	int i = index[0];
	if(i < dest.length){
		dest[i].p->smls->ctrf = true;
		g_stw->paint();
		if(g_colorOwnership){
			if(PaletteCP* paletteCP = paletteCPAllreadyOpen()){
				paletteCP->paint();
			}
		}
	}
}

void Bezier::common_removeControlPoint(int& newLen){
	if(point.length==0){
		return;
	}
	int idx=index[0];

	if(idx==0){
		point.deleteElement(idx);
		fix.deleteElement(idx);
		fix[idx]=true;
	}
	else if(idx==point.length-1){
		point.deleteElement(idx);
		fix.deleteElement(idx);
		idx--;
		index[0]=idx;
		fix[idx]=true;
	}
	else {//unfix segment
		point.deleteElement(idx);
		fix.deleteElement(idx);
	}

	if(point.length==1){
		point.length=0;
		fix.length=0;
		nopfs.length=0;
		line.length=0;
	}
	newLen=point.length;
}
void BezierLatheBuild::common_removeControlPoint(int& newLen){
	if(point.length==0){
		return;
	}
	int idx=index[0];

	if(idx==0){
		Bezier::point.deleteElement(idx);
		Bezier::fix.deleteElement(idx);
		Bezier::fix[idx]=true;
	}
	else if(idx==point.length-1){
		Bezier::point.deleteElement(idx);
		Bezier::fix.deleteElement(idx);
		idx--;
		index[0]=idx;
		Bezier::fix[idx]=true;
	}
	else {//unfix segment
		Bezier::point.deleteElement(idx);
		Bezier::fix.deleteElement(idx);
	}

	if(point.length==1){
		Bezier::point.length=0;
		Bezier::fix.length=0;
		Bezier::nopfs.length=0;
		line.length=0;
	}
	newLen=point.length;
}
void Bezier::common_removeControlPoint(Stem& stem, int& newLen){
	if(point.length==0){
		return;
	}
	int idx=index[0];

	int fixSegBeginIndex=0;
	int groupIndex = getGroupIndices(fixSegBeginIndex);

	if(idx==0){
		if(fix[idx+1]){
			stem.bands.refValue().deleteElement(groupIndex);
		}
		point.deleteElement(idx);
		fix.deleteElement(idx);
		fix[idx]=true;
	}
	else if(idx==point.length-1){
		if(fix[idx-1]){
			stem.bands.refValue().deleteElement(groupIndex);
		}
		point.deleteElement(idx);
		fix.deleteElement(idx);
		idx--;
		index[0]=idx;
		fix[idx]=true;
	}
	else if(fix[idx] == false){
		point.deleteElement(idx);
		fix.deleteElement(idx);
	}
	else {//unfix segment
		point.deleteElement(fixSegBeginIndex);
		fix.deleteElement(fixSegBeginIndex);
		stem.bands.refValue().deleteElement(groupIndex);
	}

	if(point.length==1){
		point.length=0;
		fix.length=0;
		stem.bands.refValue().length=0;
		nopfs.length=0;
		line.length=0;
	}
	newLen=point.length;

	if(stem.sync && newLen == 0){
		stem.height.refValue().length=0;
		stem.yDistance.refValue().length=0;
		stem.bands.refValue().length=0;
	}
}
void BezierLathe::removeControlPoint(int& newLen){
	Stem* stem=getStemOwner();
	if(stem && stem->sync==true){
		common_removeControlPoint(*stem, newLen);
	}
	else{
		common_removeControlPoint(newLen);
	}

}
void BezierLatheBuild::removeControlPoint(int& newLen){
	return common_removeControlPoint(newLen);
}
void BezierScale::removeControlPoint(int& newLen){
	return common_removeControlPoint(newLen);
}
void BezierCircumScale::removeControlPoint(int& newLen){
	return common_removeControlPoint(newLen);
}
void BezierShift::removeControlPoint(int& newLen){
	return common_removeControlPoint(newLen);
}
void BezierTwist::removeControlPoint(int& newLen){
	return common_removeControlPoint(newLen);
}
void BezierTape::removeControlPoint(int& newLen){
	return common_removeControlPoint(newLen);
}
void BezierBend::removeControlPoint(int& newLen){
	if(point.length==0){
		return;
	}
	int idx=index[0];

	int fixSegBeginIndex=0;
	int groupIndex = getGroupIndices(fixSegBeginIndex);

	if(idx==0){
		Bezier3f::point.deleteElement(idx);
		Bezier3f::fix.deleteElement(idx);
		Bezier3f::fix[idx]=true;
	}
	else if(idx==point.length-1){
		Bezier3f::point.deleteElement(idx);
		Bezier3f::fix.deleteElement(idx);
		idx--;
		index[0]=idx;
		Bezier3f::fix[idx]=true;
	}
	else if(fix[idx] == false){
		Bezier3f::point.deleteElement(idx);
		Bezier3f::fix.deleteElement(idx);
	}
	else {//unfix segment
		Bezier3f::point.deleteElement(fixSegBeginIndex);
		Bezier3f::fix.deleteElement(fixSegBeginIndex);
	}

	if(point.length==1){
		Bezier3f::point.length=0;
		Bezier3f::fix.length=0;
		Bezier3f::nopfs.length=0;
		line.length=0;
	}
	newLen=Bezier3f::point.length;
}
void LengthButton::action(){
	if(outOfSyncWithRobot()){
		return;
	}
	if(sprop.disableLen()){
		return;
	}

	const M_Properties& properties=node.getProperties();
	const Properties& sprop=node.getProp(fieldIndex);

	Stem* debug=0;

	Stem* stem=dynamic_cast<Stem*>(&node);
	if(stem){
		if(stem->yDistance.length==0 && stem->bands.length>0){
			return;
		}
	}
	else if(ALathe* lathe=dynamic_cast<ALathe*>(&node)){
		Stem* stem=getStemOwner();
		debug=stem;
		if(stem!=nullPtr){
			if(stem->bands.length==0){
				stem->sync.refValue()=true;
			}
		}
	}

	int mod=0;
	int newLen=0;
	int fieldLen=node.getField(fieldIndex)->size();
	
	Bezier* bezier=dynamic_cast<Bezier*>(&node);
	Bezier3f* bezier3f=dynamic_cast<Bezier3f*>(&node);
	RingBuild* ringBuild=dynamic_cast<RingBuild*>(&node);
	Oscillator* osc=dynamic_cast<Oscillator*>(&node);
	Chords* chords=dynamic_cast<Chords*>(&node);
	Harmonic* harmonic=dynamic_cast<Harmonic*>(&node);
	NurbsLathe* nurbsLathe=dynamic_cast<NurbsLathe*>(&node);
	NurbsTrack* nurbsTrack=dynamic_cast<NurbsTrack*>(&node);
	BendyLathe* bendyLathe=dynamic_cast<BendyLathe*>(&node);
	FieldGroupExtra* fieldGroupExtra=dynamic_cast<FieldGroupExtra*>(&node);

	if(stem && sprop.name=="height" && !(altKeyState())){
		if(command == 1){
			stem->insertBands(newLen);
		}
		else{
			stem->uninsertBands(newLen);
		}
	}
	else if(bezier && sprop.name=="point"){
		if(command == 1){
			bezier->addControlPoint(newLen);
		}
		else{
			bezier->removeControlPoint(newLen);
		}
		node.processFields();
	}
	else if(bezier3f && sprop.name=="point"){
		if(command == 1){
			bezier3f->addControlPoint(newLen);
		}
		else{
			bezier3f->removeControlPoint(newLen);
		}
		node.processFields();
	}
	else if(osc && sprop.name=="note"){
		if(command == 1){
			osc->addControlPoint(newLen);
		}
		else{
			osc->removeControlPoint(newLen);
		}
		node.processFields();
		houseKeepVisualClue();
	}
	else if(chords && sprop.name=="spectrum"){
		if(command == 1){
			chords->addControlPoint(newLen);
		}
		else{
			chords->removeControlPoint(newLen);
		}
		node.processFields();
		houseKeepVisualClue();
	}
	else if(harmonic && sprop.name=="point"){
		if(command == 1){
			harmonic->addControlPoint(newLen);
		}
		else{
			harmonic->removeControlPoint(newLen);
		}
		node.processFields();
		houseKeepVisualClue();
	}
	else if(nurbsTrack && sprop.name=="keyValue"){
		if(command == 1){
			nurbsTrack->insertControlPoint(nurbsTrack->index[1], .5);
		}
		else{
			nurbsTrack->removeControlPoint(nurbsTrack->index[1]);
		}
	}
	else if(bendyLathe){
		if(sprop.name=="radius"){
			if(command == 1){
				bendyLathe->addRadius(newLen);
			}
			else{
				bendyLathe->removeRadius(newLen);
			}
		}
		else if(sprop.name=="point"){
			if(command == 1){
				bendyLathe->addControlPoint(newLen);
			}
			else{
				bendyLathe->removeControlPoint(newLen);
			}
		}
		node.processFields();
		houseKeepVisualClue();
	}
	else if(ringBuild && sprop.name=="scaleX"){
		if(command == 1){
			ringBuild->addControlPoint(newLen);
		}
		else{
			ringBuild->removeControlPoint(newLen);
		}
		g_stw->paint();
	}
	else if(fieldGroupExtra && fieldGroupExtra->processGroup(sprop.groupIndex, command, newLen)){
	}
	else{
		if(fieldLen==0||command==0 && fieldLen==sprop.initElLen){
			switch(command){
				case 0:
					newLen=fieldLen-sprop.initElLen;
					break;
				case 1:
					newLen=fieldLen+sprop.initElLen;
					break;
			}
		}
		else {
			switch(command){
				case 0:
					newLen=fieldLen-sprop.incElLen;
					break;
				case 1:
					newLen=fieldLen+sprop.incElLen;
					break;
			}
		}
		if(newLen<0){
			newLen=0;
		}

		int index=newLen-1;
		if(index<0){
			index=0;
		}
		node.setFieldGroupIndex(sprop.groupIndex,index);

		int i=fieldIndex;
		do{
			int startAtIndex=properties[i].startAtIndex;
			Field* field=node.getField(i);
			int fieldLen=field->size();
			if(fieldLen==0&&newLen>0&&properties[i].initElVal!=""){
				field->addValFromStr(properties[i].initElVal);
			}
			else if(field->size()>0&&command>0&&properties[i].sameAsLast()){
				field->addValSameAsLast();
			}
			else if(command>0&&properties[i].incElVal!=""){
				field->addValFromStr(properties[i].incElVal);
			}
			else{
				if(newLen-startAtIndex>=0){
					field->resize(newLen-startAtIndex);
					field->set(field);
				}
				else{
					field->resize(0);
					field->set(field);
				}
				if(properties[i].name=="key"&&field->getType()==MFFLOAT&&newLen>1){
					M_Float& key=dynamic_cast<MFFloat*>(field)->refValue();
					for(int j=0; j<newLen; j++){
						key[j]=(float)1/(float)(newLen-1)*(float)j;
					}
				}
			}
			i++;
		}while(i<node.getFLen() && properties[i-1].groupIndex==properties[i].groupIndex);
	}

	if(ALathe* lathe=dynamic_cast<ALathe*>(&node)){
		Stem* stem=getStemOwner();
		if(stem && stem->sync==true){
			int stemIndex=node.getFieldGroupIndex(fieldIndex)-1;
			if(stemIndex<0)stemIndex=0;
			stem->index[0]=stemIndex;

			M_Int& bands=stem->bands;
			lathe->getBandAndYDist(bands, Stem::bandDist, stem->yDistance);
			if(stem->yDistance.length==0){
				stem->bands.refValue().length=0;
			}
			else{
				bands.length=stem->yDistance.length;
				int lastVal=5;
				for(int i=0; i<bands.length; i++){
					if(bands[i]==0){
						bands[i]=lastVal;
					}
					lastVal=bands[i];
				}
			}
		}
		stem->convertYDistanceToHeight();
	}
	else if(Stem* stem=dynamic_cast<Stem*>(&node)){
		if(stem->yDistance.length>0){
			stem->sync.refValue()=false;
		}
	}
	node.processFields();
	updateBuild3dWindow();
}
template<class T, const int n> void paintMFButton(const Button* button, const Properties& sprop, const int command, const int baseCommand, const Vec2i size, const Vec2i dstPos, const Vec2i srcPos, const Vec2i iconSize, Window& win){
	upDownButtonPaint<n>(button, command, baseCommand, size, dstPos, srcPos, iconSize, win);
}
template<class T, const int n> void paintSFButton(const Button* button, Node& node, const FD<T, Single>& field, const Properties& sprop, const int command, const int baseCommand, const Vec2i size, const Vec2i dstPos, const Vec2i srcPos, const Vec2i iconSize, Window& win){
	upDownButtonPaint<n>(button, command, baseCommand, size, dstPos, srcPos, iconSize, win);
}
/*
Vec2i SFDoubleButton::getValPos(){	return Vec2i(124+18*1+4, dstPos.y+16);}
Vec2i SFFloatButton::getValPos(){	return Vec2i(124+18*1+4, dstPos.y+16);}
Vec2i SFInt32Button::getValPos(){		return Vec2i(124+18*1+4, dstPos.y+16);}
Vec2i SFRotationButton::getValPos(){return Vec2i(124+18*3+4, dstPos.y+16);}
Vec2i SFStringButton::getValPos(){	return Vec2i(124+18*1+4, dstPos.y+16);}
Vec2i SFVec2fButton::getValPos(){	return Vec2i(124+18*2+4, dstPos.y+16);}
Vec2i SFVec3fButton::getValPos(){	return Vec2i(124+18*3+4, dstPos.y+16);}


Vec2i MFColorButton::getValPos(){	return Vec2i(124+18*3+4, dstPos.y+16);}
Vec2i MFDoubleButton::getValPos(){	return Vec2i(124+18*1+4, dstPos.y+16);}
Vec2i MFFloatButton::getValPos(){		return Vec2i(124+18*1+4, dstPos.y+16);}
Vec2i MFRotationButton::getValPos(){	return Vec2i(124+18*3+4, dstPos.y+16);}
Vec2i MFStringButton::getValPos(){	return Vec2i(124+18*1+4, dstPos.y+16);}
Vec2i MFVec2fButton::getValPos(){		return Vec2i(124+18*2+4, dstPos.y+16);}
Vec2i MFVec3fButton::getValPos(){		return Vec2i(124+18*3+4, dstPos.y+16);}

*/



void SFBoolButton::paint(){
	int* dstPix = win.pixBuf.getArray();

	if(sprop.push() || sprop.link()){

		const int subDx=bSize.x;
		const int subDy=bSize.y;
		const int yLn=dstPos.y;

		
		int offset=0;
		int xLn=dstPos.x;

		int strx = sprop.name.length * 8 - 8;

		if(this==getHoverButton()){
			pasteRectangle2( activeColor, dstPos.x, yLn, win.size.x, win.size.y, subDx, subDy, dstPix);
			drawHorizontalLine2( 0x808080, yLn+subDy-1, xLn+1, xLn+subDx, win.size.x, win.size.y, dstPix);
			drawVerticalLine2( 0x808080, xLn+subDx, yLn+1, yLn+subDy, win.size.x, win.size.y, dstPix);
		}
		if(this==DownButton::getDownButton()){
			pasteRectangle2( activeColor, dstPos.x, yLn, win.size.x, win.size.y, subDx, subDy, dstPix);
			offset++;
			drawHorizontalLine2( 0x808080, yLn, xLn, xLn+subDx, win.size.x, win.size.y, dstPix);
			drawVerticalLine2( 0x808080, xLn, yLn+1, yLn+subDy-2, win.size.x, win.size.y, dstPix);
		}
		int dx = dstPos.x + ((bSize.x - iconSize.x) >> 1) + offset + 4;
		int dy = dstPos.y + ((bSize.y - iconSize.y) >> 1) + offset;

		if(sprop.push()){
			if(srcPos == Vec2i(168,24)){
				win.blitHorizontalColorSweep(Vec2i(dx + 8, dy + 5), Vec2i(179, 29), Vec2i(180, 29), Vec2i(strx, 15));
				pasteSprite2(dx, dy, 172, 24, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, 8, 24, dstPix);
				pasteSprite2(dx + 8 + strx, dy, 179, 24, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, 9, 24, dstPix);
				drawText2(str=sprop.name, 0x000000, Vec2i(dx + 8 - 4, dy + 15), win.size.x, win.size.y, dstPix);
				//drawHorizontalLine2(0x0000ff,	dy + 19, dx + 8 - 4, dx + 12 + strx, win.size.x, win.size.y, dstPix);
			}
			else{
				pasteSprite2(dstPos.x+((bSize.x-iconSize.x)>>1)+offset, dstPos.y+((bSize.y-iconSize.y)>>1)+offset, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, iconSize.x, iconSize.y, win.pixBuf.getArray());
			}
		}
		else{//link
			drawText2(str=sprop.name, 0x0000ff, Vec2i(dx + 8 - 4, dy + 15), win.size.x, win.size.y, dstPix);
			drawHorizontalLine2(0x0000ff,	dy + 19, dx + 8 - 4, dx + 12 + strx, win.size.x, win.size.y, dstPix);
		}
	}
	else{
		if(bField.refValue()){
			//pasteSprite2( dstPos.x, dstPos.y, srcPos.x+bSize.x, srcPos.y,											 win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, iconSize.x, iconSize.y, dstPix);
			pasteSprite2(dstPos.x+((bSize.x-iconSize.x)>>1), dstPos.y+((bSize.y-iconSize.y)>>1), srcPos.x+iconSize.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, iconSize.x, iconSize.y, dstPix);

		}
		else{
			//pasteSprite2( dstPos.x, dstPos.y, srcPos.x, srcPos.y,													 win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, iconSize.x, iconSize.y, dstPix);
			pasteSprite2(dstPos.x+((bSize.x-iconSize.x)>>1), dstPos.y+((bSize.y-iconSize.y)>>1), srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, iconSize.x, iconSize.y, dstPix);
		}
	}
}
void SFColorButton::paint(){
	paintSFButton<Color3f, 3>(this, node, bField, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);
}
void SFDoubleButton::paint(){
	paintSFButton<double, 1>(this, node, bField, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);
}
void SFFloatButton::paint(){
	paintSFButton<float, 1>(this, node, bField, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);
}
void SFInt32Button::paint(){
	paintSFButton<int, 1>(this, node, bField, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);
}
void SFRotationButton::paint(){
	paintSFButton<Rotation, 3>(this, node, bField, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);
}
void SFStringButton::paint(){
//	drawText2(str=field, PROPERTY_COLOR, xPos-12, dstPos.y+8, win.size.x, win.size.y, win.pixBuf.getArray());
}
void SFVec2fButton::paint(){
	paintSFButton<Vec2f, 2>(this, node, bField, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);
}
void SFVec3fButton::paint(){
	paintSFButton<Vec3f, 3>(this, node, bField, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);
}
void MFBoolButton::paint(){
	if(bField.length){
		int index=node.getFieldGroupIndex(sprop.groupIndex);
		int startAtIndex=sprop.startAtIndex;
		int i=index-startAtIndex;
		if(i>=0 ){
			if(bField.refValue()[i]){
				pasteSprite2( dstPos.x, dstPos.y, srcPos.x+bSize.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray());
			}
			else{
				pasteSprite2( dstPos.x, dstPos.y, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray());
			}
		}
	}
}

void MFColorButton::paint(){	paintMFButton<Color3f, 3>	(this, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);}
void MFDoubleButton::paint(){	paintMFButton<double, 1>		(this, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);}
void MFFloatButton::paint(){	paintMFButton<float, 1>		(this, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);}
void MFInt32Button::paint(){	paintMFButton<int, 1>		(this, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);}
void MFRotationButton::paint(){	paintMFButton<Rotation, 3>	(this, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);}

void MFStringButton::paint(){
/*
	if(field.length){
		int index=node.getFieldGroupIndex(sprop.groupIndex);
		int startAtIndex=sprop.startAtIndex;
		int i=index-startAtIndex;
		if(i>=0 ){
			if(field.refValue()[i]){
				pasteSprite2( dstPos.x, dstPos.y, srcPos.x+bSize.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray(), g_tb->iconPixels);
			}
			else{
				pasteSprite2( dstPos.x, dstPos.y, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray(), g_tb->iconPixels);
			}
		}
	}
	*/
}
void MFVec2fButton::paint(){paintMFButton<Vec2f, 2>(this, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);}
void MFVec3fButton::paint(){paintMFButton<Vec3f, 3>(this, sprop, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);}
void TwoUpDownButton::paint(){upDownButtonPaint<1>(this, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);}
void IndexButton::paint(){
	if(Art* part = dynamic_cast<Art*>(&node)){
		if(part->smls->workAnimCP && (Window::animMode() || g_colorOwnership)){
			if(part->controlPoint.length){
				str="index";
				upDownButtonPaint<1>(this, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);
			}
			return;
		}
	}

	if(bField.size()==0){
		return;
	}

	str="index";
	upDownButtonPaint<1>(this, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);
}
void IndexButton::toString(String& str)const{
	if(Art* part = dynamic_cast<Art*>(&node)){
		if(part->smls->workAnimCP && (Window::animMode() || g_colorOwnership)){
			M_ControlPoint& controlPoint = part->controlPoint;
			if(controlPoint.length){
				if(MultiSelect::selectedPoint >= controlPoint.length){
					MultiSelect::selectedPoint = 0;
				}
				str = MultiSelect::selectedPoint;
				//const ControlPoint& cp = MultiSelect::controlPoint[MultiSelect::selectedPoint];
				//NurbsPatch* patch = cp.
			}
			return;
		}
	}

	if(bField.size()==0){
		return;
	}
	if(bField.size()){
		str = node.getFieldGroupIndex(sprop.groupIndex);
	}
}
void LengthButton::paint(){
	str = "length";
	upDownButtonPaint<1>(this, command, baseCommand, bSize, dstPos, srcPos, iconSize, win);
}
void LengthButton::toString(String& str)const{
	if(Art* part = dynamic_cast<Art*>(&node)){
		if(part->smls->workAnimCP && (Window::animMode() || g_colorOwnership)){
			str = part->controlPoint.length;
			//M_ControlPoint& controlPoint = part->controlPoint;
			//if(controlPoint.length){
			//	if(MultiSelect::selectedPoint >= controlPoint.length){
			//		MultiSelect::selectedPoint = 0;
			//	}
				//const ControlPoint& cp = MultiSelect::controlPoint[MultiSelect::selectedPoint];
				//NurbsPatch* patch = cp.
			//}
			return;
		}
	}
	str=node.getField(fieldIndex)->size();
}
void SFBoolButton::workOutSubIndex(const Vec2i ptCursor){}
void SFColorButton::workOutSubIndex(const Vec2i ptCursor){sixWorkOutSubIndex(*this, ptCursor);}
void SFDoubleButton::workOutSubIndex(const Vec2i ptCursor){twoWorkOutSubIndex(*this, ptCursor);}
void SFFloatButton::workOutSubIndex(const Vec2i ptCursor){twoWorkOutSubIndex(*this, ptCursor);}
void SFInt32Button::workOutSubIndex(const Vec2i ptCursor){twoWorkOutSubIndex(*this, ptCursor);}
void SFRotationButton::workOutSubIndex(const Vec2i ptCursor){sixWorkOutSubIndex(*this, ptCursor);}
void SFStringButton::workOutSubIndex(const Vec2i ptCursor){}
void SFVec2fButton::workOutSubIndex(const Vec2i ptCursor){fourWorkOutSubIndex(*this, ptCursor);}
void SFVec3fButton::workOutSubIndex(const Vec2i ptCursor){sixWorkOutSubIndex(*this, ptCursor);}

void MFBoolButton::workOutSubIndex(const Vec2i ptCursor){}
void MFColorButton::workOutSubIndex(const Vec2i ptCursor){sixWorkOutSubIndex(*this, ptCursor);}
void MFDoubleButton::workOutSubIndex(const Vec2i ptCursor){twoWorkOutSubIndex(*this, ptCursor);}
void MFFloatButton::workOutSubIndex(const Vec2i ptCursor){twoWorkOutSubIndex(*this, ptCursor);}
void MFInt32Button::workOutSubIndex(const Vec2i ptCursor){twoWorkOutSubIndex(*this, ptCursor);}
void MFRotationButton::workOutSubIndex(const Vec2i ptCursor){sixWorkOutSubIndex(*this, ptCursor);}
void MFStringButton::workOutSubIndex(const Vec2i ptCursor){}
void MFVec2fButton::workOutSubIndex(const Vec2i ptCursor){fourWorkOutSubIndex(*this, ptCursor);}
void MFVec3fButton::workOutSubIndex(const Vec2i ptCursor){sixWorkOutSubIndex(*this, ptCursor);}

void ArtCoordPointButton::toString(String& str)const{
	str="";
	if(part.smls->workAnimCP && (Window::animMode() || g_colorOwnership)){
		M_ControlPoint& controlPoint = part.controlPoint;
		if(controlPoint.length){
			if(MultiSelect::selectedPoint >= controlPoint.length){
				MultiSelect::selectedPoint = 0;
			}
			const ControlPoint& cp = part.controlPoint[MultiSelect::selectedPoint];
			str += cp.p->point[cp.i];
		}
		return;
	}
	MFVec3fButton::toString(str);
}
