#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"

#include <iostream>
#include <cstdlib>
#include "RtMidi.h"

RtMidiIn* g_midiin0 = 0;
RtMidiIn* g_midiin1 = 0;

extern M_<SceneTreeListItem> g_sceneTreeList;

Ptr<MidiFormant> g_midiFormant = 0;

const char* noteLookUp[12] = {
	"C",
	"C#",
	"D",
	"D#",
	"E",
	"F",
	"F#",
	"G",
	"G#",
	"A",
	"A#",
	"B",
};

void MidiFormant::genChordString(String& str){
	str = "";

	for(int k = 1; k < pitch.length; k++){
		int pitch = this->pitch[k];
		int octave = pitch / 12;
		int remainder = pitch - octave * 12;
		str += noteLookUp[remainder];
		str += octave - 2;
		str += ',';
	}

}

double MidiFormant::getFrequency(int formantidx){
	double cent = 100 * (pitch[formantidx] - 69); //69 = A3
	cent += g_midiFormant->cent[formantidx] - 64;

	double frequency = 220.0 * pow(pow(2.0,1.0/1200.0), cent);
	return frequency;
}

void Formant::updatePatch(){

	NurbsPatch* patch = dynamic_cast<NurbsPatch*>(this->patch.p);
	if(patch == 0){
		return;
	}
	int uDim = patch->dim.x;
	const M_Vec3f& dpoint = patch->dpoint.refValue();
	Seamless* smls = patch->smls;

	Vec3f pt;
	int uFrom = 0;

	for(int idx = 0; idx < freq.length; idx++){
		double f = freq[idx] / 10000.0;
		double b = bw[idx] / 10000.0;

		
		int uTo = uCount[idx] * 2 - 1;
		if(uFrom == 0 && uTo > 1){
			uTo--;
		}



		uTo += uFrom;

		if(uTo > uDim){
			uTo = uDim;
		}

		for(int u = uFrom; u < uTo; ){
			int pntIdx = uDim + u;
			pt.x = dpoint[u].x;
			pt.z = -(f - b / 2.0 * 1.4);
			if(pt.z > 0){
				pt.z = -.0001;
			}
			pt.y = 0;
			patch->setPointAndCp(pntIdx, pt);

			pt.z = -(f - b / 2.0 * .6);
			pt.y = 0;
			pntIdx += uDim * 2;
			patch->setPointAndCp(pntIdx, pt);


			pt.z = -f;
			if(u != 0 && u != uDim -1){
				pt.y = amp[idx];
			}
			pntIdx += uDim * 2;
			patch->setPointAndCp(pntIdx, pt);


			pt.z = -(f + b / 2.0 * .6);
			pt.y = 0;
			pntIdx += uDim * 2;
			patch->setPointAndCp(pntIdx, pt);

			pt.z = -(f + b / 2.0 * 1.4);
			pt.y = 0;
			pntIdx += uDim * 2;
			patch->setPointAndCp(pntIdx, pt);

			if(u == 0 || u == uDim - 2){
				u++;
			}
			else{
				u+=2;
			}
		}

		uFrom = uTo;
		if(uFrom > 1){
			uFrom++;
		}

		if(uFrom >= uDim){
			uFrom = uDim - 1;
		}
	}

	smls->calcAllPolyMidPoints();

	smls->doBuild();
	smls->ctrf = true;
	houseKeepVisualClue();
	g_stw->cp->paint();



	if(Anim* anim = Window::getAnimBar()){
		if(anim->soundTrack.length){
			if(Composition* composition = dynamic_cast<Composition*>(anim->soundTrack[0].p)){
				composition->updateAndPlay(anim, patch);
			}
		}
	}

}

void LowPass::updatePatch(){

	NurbsPatch* patch = dynamic_cast<NurbsPatch*>(this->patch.p);
	if(patch == 0){
		return;
	}
	int uDim = patch->dim.x;
	const M_Vec3f& dpoint = patch->dpoint.refValue();
	Seamless* smls = patch->smls;

	Vec3f pt;
	int uFrom = 0;

	for(int idx = 0; idx < freq.length; idx++){
		double f = freq[idx] / 10000.0;
		double b = bw[idx] / 10000.0;
		

		
		int uTo = uCount[idx] * 2 - 1;
		if(uFrom == 0 && uTo > 1){
			uTo--;
		}



		uTo += uFrom;

		if(uTo > uDim){
			uTo = uDim;
		}

		for(int u = uFrom; u < uTo; ){
			int pntIdx = u;
			pt.x = dpoint[u].x;
			pt.y = 0;
			if(u != 0 && u != uDim -1){
				pt.y = pass[idx];
			}
			pt.z = 0;
			patch->setPointAndCp(pntIdx, pt);


			pntIdx += uDim;
			pt.x = dpoint[u].x;
			pt.z = -(f - b / 2.0 * 1.4);
			if(pt.z >= 0){
				pt.z = -.0001;
			}
			patch->setPointAndCp(pntIdx, pt);

			pt.z = -(f - b / 2.0 * .6);
			pntIdx += uDim * 2;
			patch->setPointAndCp(pntIdx, pt);


			if(u != 0 && u != uDim -1){
				pt.y = amp[idx];
			}
			pt.z = -f;
			pntIdx += uDim * 2;
			patch->setPointAndCp(pntIdx, pt);


			pt.z = -(f + b / 2.0 * .6);
			if(u != 0 && u != uDim -1){
				pt.y = leak[idx];
			}
			pntIdx += uDim * 2;
			patch->setPointAndCp(pntIdx, pt);

			pt.z = -(f + b / 2.0 * 1.4);
			pntIdx += uDim * 2;
			patch->setPointAndCp(pntIdx, pt);

			pntIdx += uDim;
			pt.z = dpoint[pntIdx].z;
			patch->setPointAndCp(pntIdx, pt);

			if(u == 0 || u == uDim - 2){
				u++;
			}
			else{
				u+=2;
			}
		}

		uFrom = uTo;
		if(uFrom > 1){
			uFrom++;
		}

		if(uFrom >= uDim){
			uFrom = uDim - 1;
		}
	}

	smls->calcAllPolyMidPoints();

	smls->doBuild();
	smls->ctrf = true;
	houseKeepVisualClue();
	g_stw->cp->paint();



	if(Anim* anim = Window::getAnimBar()){
		if(anim->soundTrack.length){
			if(Composition* composition = dynamic_cast<Composition*>(anim->soundTrack[0].p)){
				composition->updateAndPlay(anim, patch);
			}
		}
	}

}

void MultiFormant::updatePatch(){
}

void MidiFormant::updateFilter(Filter* filter){
	M_Node& m_formant = filter->formant;

	for(int k = 0; k < peak.length && k < m_formant.length; k++){
		if(Formant* formant = dynamic_cast<Formant*>(m_formant[k].p)){
			formant->freq.refValue() = getFrequency(k);
			formant->amp.refValue() = peak[k] * .005;
			formant->bw.refValue() = 600;
			formant->uCount.refValue() = uCount.refValue();
		}
	}
}

void updatePatchesForMidi(){
	if(Formant* formant = dynamic_cast<Formant*>(g_selNode.getRaw())){
		if(formant->update){
			formant->update = false;
			formant->updatePatch();
		}
		else if(formant->justPlay){
			formant->justPlay = false;
			if(Anim* anim = Window::getAnimBar()){
				anim->play = true;
			}
		}
		return;
	}
	if(g_midiFormant == 0){
		return;
	}


	bool globalUpdate = false;

	if(Filter* filter = dynamic_cast<Filter*>(g_midiFormant->patch.p)){
		g_midiFormant->updateFilter(filter);
	}

	NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_midiFormant->patch.p);

	if(patch == 0){
		return;
	}
	const M_Int& pitch = g_midiFormant->pitch;
	M_Bool& m_update = g_midiFormant->update;

	bool update = false;

	for(int f = 0; f < pitch.length; f++){
		if(m_update[f]){
			update = true;
			break;
		}
	}

	if(update == false){
		return;
	}

	Seamless* smls = patch->smls;

	int uDim = patch->dim.x;
	int vDim = patch->dim.y;


	float beforeBase = 0;

	if(g_midiFormant->update.length != g_midiFormant->pitch.length){
		g_midiFormant->update.length = g_midiFormant->pitch.length;
	}


	for(int f = 0; f < pitch.length; f++){
		float afterBase = g_midiFormant->trough[f] * .005;
		//if(update[f]){
		{
			m_update[f] = false;
			globalUpdate = true;

			double frequency = g_midiFormant->getFrequency(f);
			double bw = g_midiFormant->bw[f] * 5;

			double previousFrequency = 0;
			double previousFootFrequency = 0;
			double previousBorder = 0;

			double nextFrequency = 0;
			double nextFootFrequency = 0;
			double nextBorder = 0;

			if(f > 0){
				previousFrequency = g_midiFormant->getFrequency(f - 1);
				previousFootFrequency = previousFrequency + g_midiFormant->bw[f - 1] * 5;
				previousBorder = ((frequency -  bw) + previousFootFrequency) * .5;
			}
			if(f < pitch.length - 1){
				nextFrequency = g_midiFormant->getFrequency(f + 1);
				nextFootFrequency = nextFrequency - g_midiFormant->bw[f + 1] * 5;
				nextBorder = ((frequency +  bw) + nextFootFrequency) * .5;
			}

			float peak = g_midiFormant->peak[f] * .005;
			
			const int vLine = g_midiFormant->vLine[f];
			int before = g_midiFormant->before[f];
			int after = g_midiFormant->after[f];

			int uFrom = 0;
			if(g_midiFormant->uOffset){
				uFrom = (g_midiFormant->uOffset - 1) * 2 + 1;
			}

			int uTo = uDim;

			if(g_midiFormant->uCount > 0){
				uTo = (g_midiFormant->uOffset + g_midiFormant->uCount - 1) * 2 + 1;
			}

			for(int u = uFrom; u < uTo; u++){
				/*
				int pntIdx = 0;
				if(vLine > 0){
					pntIdx = ((vLine - 1) * 2 + 1) * uDim;
				}
				pntIdx +=  u;
				*/

				int pntIdx = ((vLine - 1) * 2 + 1) * uDim + u;
				if(vLine == 0){
					pntIdx +=  uDim;
				}


				Vec3f pt = patch->dpoint[pntIdx];
				pt.y = 0;
				if(u > 0 && u < uDim -1){
					pt.y = peak;
				}
				pt.z = frequency * -.0001;
				patch->setPointAndCp(pntIdx, pt);


				for(int b = before; b > 0; b--){
					int pntIdxB = pntIdx - uDim * b * 2;
					if(vLine == 1){
						pntIdxB += uDim;
					}
					Vec3f ptB = patch->dpoint[pntIdxB];

					ptB.y = 0;
					if(u > 0 && u < uDim -1){
						ptB.y = beforeBase;
					}
					ptB.z = pt.z + b * bw * .0001;
					if(f > 0){
						if((frequency - b * bw) < previousBorder){
							ptB.z = previousBorder * -.0001 - (before + 1 - b) * .00001;
						}
					}

					patch->setPointAndCp(pntIdxB, ptB);
				}
				for(int a = 1; a <= after; a++){
					int pntIdxA = pntIdx + uDim * a * 2;
					if(vLine == 0){
						pntIdxA -= uDim;
					}
					Vec3f ptA = patch->dpoint[pntIdxA];
					ptA.y = 0;
					if(u > 0 && u < uDim -1){
						ptA.y = afterBase;
					}
					ptA.z = pt.z - a * bw * .0001;
					if(f < pitch.length - 1){
						if((frequency + a * bw) > nextBorder){
							ptA.z = nextBorder * -.0001 + (after + 1 - a) * .00001;
						}
					}

					patch->setPointAndCp(pntIdxA, ptA);
				}
			}
		}
		beforeBase = afterBase;
	}
	if(globalUpdate){
		smls->calcAllPolyMidPoints();

		smls->doBuild();
		smls->ctrf = true;
		houseKeepVisualClue();
		g_stw->cp->paint();

//		if(smls->animCP == false){
			if(Anim* anim = Window::getAnimBar()){
				if(anim->soundTrack.length){
					if(Composition* composition = dynamic_cast<Composition*>(anim->soundTrack[0].p)){
						composition->updateAndPlay(anim, patch);
					}
				}
			}
		//}
	}
}


int MidiFormant::getFormantFromPitchRange(int key){
	int k = 0;
	for(; k < pitch.length; k++){
		if(key <= range[k]){
			break;
		}
	}
	if(k >= pitch.length){
		k = pitch.length - 1;
	}
	return k;
}

int g_polyNote = 0;

void formantMycallback0( double deltatime, std::vector< unsigned char > *message, void *){
	Formant* formant = dynamic_cast<Formant*>(g_selNode.getRaw());
	if(formant == 0){
		return;
	}
	unsigned int nBytes = message->size();

	if(nBytes == 3){
		unsigned char byte0 = message->at(0);
		if(byte0 == 0x90){
			int key = message->at(1) + 12 * 3;
			double cent = 100 * (key - 69); //69 = A3
			double frequency = 220.0 * pow(pow(2.0,1.0/1200.0), cent);

			int idx = formant->index[0];
			float freq =  formant->freq.refValue()[idx];
			float diff = fabs(freq - frequency);

			if(diff > .1){
				formant->freq.refValue()[idx] = frequency;
				formant->update = true;
			}
			else{
				formant->justPlay = true;
			}
		}
	}
}

void mycallback0( double deltatime, std::vector< unsigned char > *message, void *){

	NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_midiFormant->patch.p);

	if(patch == 0){
		return;
	}
	int formantIdx = g_midiFormant->index[0];

	Seamless* smls = patch->smls;

	int uDim = patch->dim.x;
	int vDim = patch->dim.y;

	//57 = A
	unsigned int nBytes = message->size();

	if(nBytes == 3){
		unsigned char byte0 = message->at(0);
		if(byte0 == 0x90){
			int key = message->at(1) + 12 * 3;
			//formantIdx = g_midiFormant->getFormantFromPitchRange(key);
			formantIdx = ++g_polyNote;
			g_midiFormant->index[0] = formantIdx;
			g_midiFormant->pitch.refValue()[formantIdx] = key;
			if(g_polyNote == 3){
				bubbleSort(g_midiFormant->pitch.refValue());
				g_polyNote = 0;
				g_midiFormant->update[formantIdx] = true;
			}
			g_midiFormant->genChordString(g_midiFormant->chord.refValue());
		}
		else if(byte0 == 0xb0){
			g_midiFormant->cent.refValue()[formantIdx] = message->at(2);
			g_midiFormant->update[formantIdx] = true;
		}
	}
}

void mycallback1( double deltatime, std::vector< unsigned char > *message, void *){
	unsigned int nBytes = message->size();
	if(nBytes == 3 && message->at(0) == 0xbf){
		int c = message->at(1);

		if(c - 12 >= 0 && c - 12 < 9){
			c = c - 12;
			int formantIdx = c >> 1;
			if(formantIdx < g_midiFormant->pitch.length){
				g_midiFormant->index[0] = formantIdx;

				if(c & 1){
					g_midiFormant->trough.refValue()[formantIdx] = message->at(2);
					if(formantIdx + 1 < g_midiFormant->pitch.length){
						g_midiFormant->update[formantIdx + 1] = true;
					}
				}
				else{
					g_midiFormant->peak.refValue()[formantIdx] = message->at(2);
				}
				g_midiFormant->update[formantIdx] = true;
			}
		}
		else if(c - 21 >= 0 && c - 21 < 9){
			int formantIdx = c - 21;
			if(formantIdx < g_midiFormant->pitch.length){
				g_midiFormant->index[0] = formantIdx;
				g_stw->cp->paint();
			}
		}
		else if(c - 30 >= 0 && c - 30 < 8){
			int formantIdx = c - 30;
			if(formantIdx < g_midiFormant->pitch.length){
				g_midiFormant->index[0] = formantIdx;
				g_midiFormant->bw.refValue()[formantIdx] = message->at(2);
				g_midiFormant->update[formantIdx] = true;
			}
		}
	}
}

void MidiFormant::processFields(){
	update.length = pitch.length;

	M_Int& range = this->range.refValue();
	if(range.length != pitch.length){
		range.length = pitch.length;
		int f = 0;
		for(; f < range.length - 1; f++){
			range[f] = pitch[f + 1] - 2;
		}
		range[f] = 127;
	}
}



void setUpMidi(){
static Formant* g_formant = 0;

	if(Formant* formant = dynamic_cast<Formant*>(g_selNode.getRaw())){
		formant->update = true;
		if(g_formant == 0){
			g_formant = formant;

			if(g_midiin0 = new RtMidiIn()){
				g_midiin0->openPort(0);

				g_midiin0->setCallback(*formantMycallback0);

				// Don't ignore sysex, timing, or active sensing messages.
				g_midiin0->ignoreTypes( false, false, false );
			}

			if(g_midiin1 = new RtMidiIn()){
				g_midiin1->openPort(1);

				g_midiin1->setCallback(*mycallback1);

				// Don't ignore sysex, timing, or active sensing messages.
				g_midiin1->ignoreTypes( false, false, false );

			}
		}
		return;
	}

	if(g_midiFormant){
		for(int f = g_midiFormant->update.length; --f>=0;){
			g_midiFormant->update[f] = true;
		}
		return;
	}

	for(int k = 0; k < g_sceneTreeList.length; k++){
		if(g_midiFormant = dynamic_cast<MidiFormant*>(g_sceneTreeList[k].node.p)){
			if(g_midiFormant->patch.p){
				break;
			}
		}
	}
	if(g_midiFormant == 0){
		return;
	}
	g_midiFormant->processFields();

	for(int f = g_midiFormant->update.length; --f>=0;){
		g_midiFormant->update[f] = true;
	}


	if(g_midiin0 = new RtMidiIn()){
		g_midiin0->openPort(0);

		g_midiin0->setCallback(*mycallback0);

		// Don't ignore sysex, timing, or active sensing messages.
		g_midiin0->ignoreTypes( false, false, false );
	}

	if(g_midiin1 = new RtMidiIn()){
		g_midiin1->openPort(1);

		g_midiin1->setCallback(*mycallback1);

		// Don't ignore sysex, timing, or active sensing messages.
		g_midiin1->ignoreTypes( false, false, false );

	}

	g_midiFormant->genChordString(g_midiFormant->chord.refValue());
}

void MidiFormant::Select::operator=(const bool& par){
	g_midiFormant = dynamic_cast<MidiFormant*>(g_selNode.getRaw());
	g_midiFormant->processFields();
}
