/* Seamless3d

Copyright (C) 2001-2013 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#define STRICT

#include <windows.h>
#include <direct.h>
//#include <process.h>
//#include <shlobj.h>

#include "seamless3d.h"
#include "resource.h"
#include "gui.h"
#include "scene_tree.h"
#include "techuelife_triangle_editor.h"

#include"robot_aid.h"

//#define ROBOT_TIME_DIVIDE 0.00075

extern bool tugFlag;
extern bool tugFlag2;
extern bool dragPartFlag;
extern bool orientationTugFlag;
extern bool	g_mirrorModeVertexFound;
extern bool g_dontBackup;


extern const char* langExtension;
extern const char* msg_not_a_movie;
extern const char* str_interupted;
extern const char* menu_file;
extern const char* menu_save_as;
extern const char* menu_short_copy;

extern const char* str_space;
extern const char* str_l_alt;
extern const char* str_l_ctrl;
extern const char* str_l_shift;
extern const char* str_r_alt;
extern const char* str_r_ctrl;
extern const char* str_r_shift;
extern const char* str_arrow_down;
extern const char* str_left_click;
extern const char* str_double_click;
extern const char* str_right_click;
extern const char* str_left_drag;
extern const char* str_right_drag;
extern const char* msg_screen_res_must_be;
extern const char* msg_3d_win_res_must_be;
extern const char* str_done;
extern const char* str_complete;
extern const char* str_waiting;
extern const char* str_renameLtoR;

//UINT_PTR InputEvent::timeID;
float InputEvent::fraction_changed = 0;
double InputEvent::beginRobotTime = 0;
float InputEvent::period = .5;
const float InputEvent::downPeriod = .25;
const float InputEvent::upPeriod = .125;
const float InputEvent::prePeriod = .25;
const float InputEvent::postPeriod = .25;


bool InputEvent::robotActive = false;
M_<InputEvent*> InputEvent::list;
int InputEvent::sequencerIdx = 0;
int InputEvent::sceneTreeScrolPos = 0;
Vec2i InputEvent::startPos;
Vec2i InputEvent::pos;


Vec3f InputEvent::startPos3d;

bool InputEvent::spaceDown = false;
bool InputEvent::l_altDown = false;
bool InputEvent::l_ctrlDown = false;
bool InputEvent::l_shiftDown = false;

bool InputEvent::r_altDown = false;
bool InputEvent::r_ctrlDown = false;
bool InputEvent::r_shiftDown = false;

bool InputEvent::leftClick = false;
bool InputEvent::doubleClick = false;
bool InputEvent::rightClick = false;
bool InputEvent::leftDrag = false;
bool InputEvent::rightDrag = false;
bool InputEvent::arrowDown = false;

Rotation DragViewRot::from;
Vec3f DragViewPos::from;
Vec3f DemoMove::from;
float MouseZoom::from = 0;

int InputEvent::renderCounter = 0;
int InputEvent::syncNumOfActions = 1;
int InputEvent::prevRenderCounter = 0;
extern Ptr<AnimClue> g_vcArtAnimBase;
extern Ptr<PartClue> g_vcArtTriBase;
bool smlsIsSaved();
void updateIconPointers();
void addNodesForNewFile();

Vec2i get2dMousePosFrom3d(Vec3f par);
Vec2i get2dMousePosFrom2d(Vec2f par);

extern Group* g_vcSelectRectGroup;
extern ShapePtr g_vcSelectRectShape;
bool getCollapsedLineNoGlobalChange(int nonCollapsedLine, int& collapsedLine);
extern Ptr<PasteOptions> g_pasteOptions;
extern Ptr<DeleteOptions> g_deleteOptions;
extern Ptr<TimelineFilter> g_timelineFilter;

bool checkRobotActive(){
	return InputEvent::robotActive;
}
M_<char> InputEvent::dynamChar;
void setPasteDefault();
void loadGridBrushEtcNodes();

Ptr<Demo> g_robotDemo = new Demo;
Ptr<ControlPanel> g_rcp;
//#define TOP_TOOLBAR_ROW 700
//#define NEW_SMLS_FILE_BUTTON 20
//#define SCENE_TREE_BUTTON 140

void InputEvent::outputStatus(){
	String& str=g_robotDemo->status.refValue();
	str="";
	if(spaceDown){
		str+=str_space;
	}
	if(l_altDown){
		str+=str_l_alt;
	}
	if(l_ctrlDown){
		str+=str_l_ctrl;
	}
	if(l_shiftDown){
		str+=str_l_shift;
	}
	if(r_altDown){
		str+=str_r_alt;
	}
	if(r_ctrlDown){
		str+=str_r_ctrl;
	}
	if(r_shiftDown){
		str+=str_r_shift;
	}
	if(arrowDown){
		str+=str_arrow_down;
	}
	if(leftClick){
		str+=str_left_click;
	}
	if(doubleClick){
		str+=str_double_click;
	}
	if(rightClick){
		str+=str_right_click;
	}
	if(leftDrag){
		str+=str_left_drag;
	}
	if(rightDrag){
		str+=str_right_drag;
	}
	g_rcp->paint();
}

void resetDragActuallyHappned(){
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		tugFlag2=tugFlag = false;
		g_dragActuallyHappned = false;
		if(dynamic_cast<AnimClue*>(g_visualClue) == 0){
			part->smls->ctrf = true;
		}
	}
	g_dragActuallyHappned = false;
	mouse3dMode = false;
	tugFlag=false;
	orientationTugFlag=false;
	dragPartFlag=false;
	showCursor(TRUE);
	InputEvent::outputStatus();
}

void InputEvent::_killTimer(){
	robotActive = false;
	resetTemporaryModes();
	//KillTimer(hWndMain, TIMEID);
	

	keybd_event(VK_LMENU, 0, KEYEVENTF_KEYUP, 0);
	keybd_event(VK_RMENU, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
	keybd_event(VK_LCONTROL, 0, KEYEVENTF_KEYUP, 0);
	keybd_event(VK_RCONTROL, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
	keybd_event(VK_LSHIFT, 0, KEYEVENTF_KEYUP, 0);
	keybd_event(VK_RSHIFT, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
	//SetKeyboardState(lpKeyState)

	mouse_event(
		MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTUP,
		0,
		0,
		0,
		0
	);
	mouse_event(
		MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_RIGHTUP,
		0,
		0,
		0,
		0
	);

	spaceDown = false;
	l_altDown = false;
	l_ctrlDown = false;
	l_shiftDown = false;
	r_altDown = false;
	r_ctrlDown = false;
	r_shiftDown = false;

	if(mouse3dMode){
		g_mirrorModeVertexFound = false;
		UnAcquire();
		resetDragActuallyHappned();
	}

	setSmlsSaved();
}

bool InputEvent::resumeSequence(){
	if(robotActive){
		if(dynamic_cast<Section*>(list[sequencerIdx])){
			return true;
		}
		robotActive = false;
		resetTemporaryModes();
		//KillTimer(hWndMain, TIMEID);
		_killTimer();
		g_robotDemo->status = str_interupted;
		g_rcp->paint();

	}
	return false;
}

Vec2f interpMove(float time, float period, Vec2i startPos, Vec2i endPos){
	Vec2f startPosf=Vec2f(startPos.x, startPos.y);
	Vec2f endPosf=Vec2f(endPos.x, endPos.y);

	float interp=1.0/period*(time);
	interp=1.0f-(cosf(interp*PI)*.5+.5);

	return (startPosf+(endPosf-startPosf)*interp);
}
float InputEvent::move(Vec2i endPos){
	Vec2i diff=Vec2i(endPos.x-startPos.x, endPos.y-startPos.y);
	float distance=sqrtf(diff.x*diff.x + diff.y*diff.y);
	float period=fabs(distance)*.005 + .5;
	if(period>1.5){
		period=1.5;
	}

/*
	if(period<50){
		period*=.01;
	}
	else if(period<100){
		period*=.0075;
	}
	else if(period<200){
		period*=.005;
	}
	else if(period<300){
		period*=.003;
	}
	else if(period<400){
		period*=.002;
	}
	else if(period<500){
		period*=.0015;
	}
	else{
		period*=.001;
	}
	*/
	Vec2f newPos;
	if(fraction_changed>=period){
		newPos.x=endPos.x;
		newPos.y=endPos.y;
	}
	else{
		newPos=interpMove(fraction_changed, period, startPos, endPos);
	}
//	dLog(String ("startPos=") + startPos.x + String(" ") + startPos.y + String("period=")+period);
	{
		pos.x=newPos.x;
		pos.y=newPos.y;

		Vec2i sysPos=getSysMouseCoord();
		INPUT input={
			INPUT_MOUSE,
			{
				sysPos.x,
				sysPos.y,
				0,
				MOUSEEVENTF_MOVE | MOUSEEVENTF_ABSOLUTE,
				0,
				0
			}
		};
		SendInput(1, &input, sizeof(INPUT));
		if(fraction_changed >= period+.25){
			POINT point;
			GetCursorPos(&point);
			if(abs((int)pos.x-(int)point.x)<=1 && abs((int)pos.y-(int)point.y)<=1){
				return period;
			}
		}
	}
	return -1;
}
float InputEvent::localMove(Vec2i endPos){
	ClientToScreen(hWndMain, reinterpret_cast<POINT*>(&endPos));
	return move(endPos);
}
Vec2f interpMove(float time, float period, Vec2f startPos, Vec2f endPos){
	float interp=1.0/period*(time);
	interp=1.0f-(cosf(interp*PI)*.5+.5);

	return startPos+(endPos-startPos)*interp;
}
float InputEvent::moveTo3dPoint(Vec2f endPos){
	Vec2f diff = Vec2f(endPos.x - startPos3d.x, endPos.y - startPos3d.y);
	float distance = sqrtf(diff.x * diff.x + diff.y * diff.y);
	float period = fabs(distance) * 400 * .005 + .5;
	if(period > 2){
		period = 2;
	}

	/*
	if(period<50){
		period*=.01;
	}
	else if(period<100){
		period*=.0075;
	}
	else if(period<200){
		period*=.005;
	}
	else if(period<300){
		period*=.003;
	}
	else if(period<400){
		period*=.002;
	}
	else if(period<500){
		period*=.0015;
	}
	else{
		period*=.001;
	}
	*/
	Vec2f newPos;
	if(fraction_changed >= period){
		newPos = endPos;
	}
	else{
		newPos = interpMove(fraction_changed, period, Vec2f(startPos3d.x, startPos3d.y), endPos);
	}
	{
		g_pointer.translation = Vec3f(newPos.x, newPos.y, g_pointer.translation.z);
//		dLog(g_pointer.translation.toString());
		if(fraction_changed >= period + .25){
			if(g_pointer.translation.x == endPos.x && g_pointer.translation.y == endPos.y){
				//dLog("g_pointer.translation==endPos");
				return period;
			}
		}
	}
	return -1;
}

Vec2i InputEvent::getSysMouseCoord(){
	HWND hWndDeskTop=GetDesktopWindow();
	RECT deskTopRect;
	GetWindowRect(
		hWndDeskTop,      // handle to window
		&deskTopRect   // address of structure for client coordinates
	);
	float width=deskTopRect.right - deskTopRect.left;
	float height=deskTopRect.bottom - deskTopRect.top;

	return Vec2i(
		static_cast<unsigned>(static_cast<float>(pos.x)/width * 65535.0f),
		static_cast<unsigned>(static_cast<float>(pos.y)/height * 65535.0f)
	);
	
}
void InputEvent::rightClickNode(int nodeLine){
	list+=new MoveToNodeAsChild(nodeLine);
	list+=new RightClick();
}
void InputEvent::copyNode(int nodeLine){
	rightClickNode(nodeLine);
	list+=new Move_MenuItem(menu_short_copy);
	list+=new Click;
}
void InputEvent::selNode(int nodeLine){
	list+=new MoveToNodeAsChild(nodeLine);
	list+=new Click();
}
void InputEvent::selNodeByName(const char* name){
	list+=new MoveToNodeNameAsChild(name);
	list+=new Click();
}
void InputEvent::newNodeAsChild(int nodeID, int nodeLine){
	list+=new MoveToNewNode(nodeID);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsChild(nodeLine);
	list+=new LeftButtonUp();
}
void InputEvent::newNodeAsReplacement(int nodeID, int nodeLine){
	list+=new MoveToNewNode(nodeID);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsReplacement(nodeLine);
	list+=new LeftButtonUp();
}
void InputEvent::newNodeAsSiblingBelow(int nodeID, int nodeLine){
	list+=new MoveToNewNode(nodeID);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsSiblingBelow(nodeLine);
	list+=new LeftButtonUp();
}
void InputEvent::useNodeAsSource(int nodeLineFrom, int nodeLineTo){
	list+=new MoveToNode(nodeLineFrom);
	list+=new VirtKeyDown(VK_LMENU);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsSource(nodeLineTo);
	list+=new LeftButtonUp();
	list+=new VirtKeyUp(VK_LMENU);
}
void InputEvent::useNodeAsSiblingBelow(int nodeLineFrom, int nodeLineTo){
	list+=new MoveToNode(nodeLineFrom);
	list+=new VirtKeyDown(VK_LMENU);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsSiblingBelow(nodeLineTo);
	list+=new LeftButtonUp();
	list+=new VirtKeyUp(VK_LMENU);
}
void InputEvent::useNodeAsSiblingAbove(int nodeLineFrom, int nodeLineTo){
	list+=new MoveToNode(nodeLineFrom);
	list+=new VirtKeyDown(VK_LMENU);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsSiblingAbove(nodeLineTo);
	list+=new LeftButtonUp();
	list+=new VirtKeyUp(VK_LMENU);
}
void InputEvent::useNodeAsReplacement(int nodeLineFrom, int nodeLineTo){
	list+=new MoveToNode(nodeLineFrom);
	list+=new VirtKeyDown(VK_LMENU);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsReplacement(nodeLineTo);
	list+=new LeftButtonUp();
	list+=new VirtKeyUp(VK_LMENU);
}
void InputEvent::useNodeAsDest(int nodeLineFrom, int nodeLineTo){
	list+=new MoveToNode(nodeLineFrom);
	list+=new VirtKeyDown(VK_LMENU);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsDest(nodeLineTo);
	list+=new LeftButtonUp();
	list+=new VirtKeyUp(VK_LMENU);
}
void InputEvent::copyNodeAsChild(int nodeLineFrom, int nodeLineTo){
	list+=new MoveToNode(nodeLineFrom);
	list+=new VirtKeyDown(VK_LCONTROL);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsChild(nodeLineTo);
	list+=new LeftButtonUp();
	list+=new VirtKeyUp(VK_LCONTROL);
}
void InputEvent::copyNodeAsSiblingAbove(int nodeLineFrom, int nodeLineTo){
	list+=new MoveToNode(nodeLineFrom);
	list+=new VirtKeyDown(VK_LCONTROL);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsSiblingAbove(nodeLineTo);
	list+=new LeftButtonUp();
	list+=new VirtKeyUp(VK_LCONTROL);
}
void InputEvent::copyNodeAsReplacement(int nodeLineFrom, int nodeLineTo){
	list+=new MoveToNode(nodeLineFrom);
	list+=new VirtKeyDown(VK_LCONTROL);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsReplacement(nodeLineTo);
	list+=new LeftButtonUp();
	list+=new VirtKeyUp(VK_LCONTROL);
}
void InputEvent::copyNodeAsSiblingBelow(int nodeLineFrom, int nodeLineTo){
	list+=new MoveToNode(nodeLineFrom);
	list+=new VirtKeyDown(VK_LCONTROL);
	list+=new LeftButtonDown();
	list+=new MoveToNodeAsSiblingBelow(nodeLineTo);
	list+=new LeftButtonUp();
	list+=new VirtKeyUp(VK_LCONTROL);
}
void InputEvent::setFieldValue(const char* field, const char* value){
	list+=new MoveToField(field);
	list+=new Click;
	list+=new VirtKeyDown(VK_LSHIFT);
	list+=new TypeVirtKey(VK_END);
	list+=new VirtKeyUp(VK_LSHIFT);
	list+=new TypeString(value);
}
void InputEvent::setFieldValueNoShift(const char* field, const char* value){
	list+=new MoveToField(field);
	list+=new Click;
	list+=new TypeString(value);
}
void InputEvent::clickButton(void* id, int numOfTimes, const int buttonOffset){
	list+=new MoveToButton(id, buttonOffset);
	while(--numOfTimes>=0){
		list+=new Click;
	}
}
void InputEvent::autoRepeatCtrl(void* id, int numOfTimes, const int buttonOffset){
	list+=new MoveToButton(id, buttonOffset);
	list+=new VirtKeyDown(VK_LCONTROL);
	list+=new ClickAndAutoRepeat(numOfTimes);
	list+=new VirtKeyUp(VK_LCONTROL);
}
void InputEvent::autoRepeat(void* id, int numOfTimes, const int buttonOffset){
	list+=new MoveToButton(id, buttonOffset);
	list+=new ClickAndAutoRepeat(numOfTimes);
}
void InputEvent::incField(const char* field, int numOfTimes, const int buttonOffset, const int layer){
	list+=new MoveToFieldInc(field, buttonOffset, layer);
	list+=new ClickAndAutoRepeat(numOfTimes);
}
void InputEvent::decField(const char* field, int numOfTimes, const int buttonOffset){
	list+=new MoveToFieldDec(field, buttonOffset);
	list+=new ClickAndAutoRepeat(numOfTimes);
}
void InputEvent::incFieldCtrl(const char* field, int numOfTimes, const int buttonOffset){
	list+=new MoveToFieldInc(field, buttonOffset);
	list+=new VirtKeyDown(VK_LCONTROL);
	list+=new ClickAndAutoRepeat(numOfTimes);
	list+=new VirtKeyUp(VK_LCONTROL);
}
void InputEvent::decFieldCtrl(const char* field, int numOfTimes, const int buttonOffset){
	list+=new MoveToFieldDec(field, buttonOffset);
	list+=new VirtKeyDown(VK_LCONTROL);
	list+=new ClickAndAutoRepeat(numOfTimes);
	list+=new VirtKeyUp(VK_LCONTROL);
}
void InputEvent::collapse(int nodeLine){
	list+=new MoveToNodeAsReplacement(nodeLine);
	list+=new Click;
}
void InputEvent::textureTile(Vec2i point){
	list+=new MoveToTextureTile(point);
	list+=new Click;
}
void InputEvent::hide(int nodeLine){
	list+=new MoveToNode(nodeLine);
	list+=new DoubleClick;
}
void InputEvent::deleteMinus(const char* field){
	list+=new MoveToField(field);
	list+=new Click;
	list+=new TypeVirtKey(VK_DELETE);
}
void InputEvent::addMinus(const char* field){
	list+=new MoveToField(field);
	list+=new Click;
	list+=new TypeString("-");
}
void InputEvent::invertRotZ(const char* field){
	list+=new MoveToField(field);
	list+=new Click;
	for(int i=4; --i>=0;){
		list+=new TypeVirtKey(VK_RIGHT);
	}
	list+=new TypeString("-");
}
void InputEvent::newCenterJoint(int nodeLine, int numOf){
	selNode(nodeLine);
	clickButton(decPivotX, 1, 0);
	clickButton(decPivotX, 1, 1);
	incField("numOf", numOf);
}
void InputEvent::incGroupCount(const char* field){
	list+=new IncFieldGroup(field);
}
void InputEvent::decGroupCount(const char* field){
	list+=new DecFieldGroup(field);
}
void InputEvent::decIndex(const char* field, int numOf){
	list+=new MoveToDecGroupIndex(field);
	while(--numOf>=0){
		list+=new Click;
	}
}
void InputEvent::incIndex(const char* field, int numOf){
	list+=new MoveToIncGroupIndex(field);
	while(--numOf>=0){
		list+=new Click;
	}
}
void InputEvent::renameNode(int nodeLine, const char* newName){
	selNode(nodeLine);
	list+=new TypeVirtKey(VK_HOME);
	list+=new VirtKeyDown(VK_LSHIFT);
	list+=new TypeVirtKey(VK_END);
	list+=new VirtKeyUp(VK_LSHIFT);
	list+=new TypeString(newName);
}
void InputEvent::typeToNode(int nodeLine, const char* text, int charIndex){
	list+=new MoveToNodeLable(nodeLine, charIndex);
	list+=new Click();
	list+=new TypeString(text);
}
void InputEvent::nameNode(int nodeLine, const char* newName){
	selNode(nodeLine);
	list+=new TypeString(newName);
}

void InputEvent::renameLtoR(int nodeLine){
	selNode(nodeLine);
	list+=new TypeString(str_renameLtoR);

}
void InputEvent::joinParts(int copy, int sel){//copy ends up owning the close vertices
	copyNode(copy);
	selNode(sel);
	clickButton(transferCloseVerticesJoin);
}
void InputEvent::joinVertices(const int* coordIndex, int numOf){
	for(int i = 0; i < numOf; ){
		list+=new SelectVertex(coordIndex[i++]);
		clickButton(remember);
		list+=new SelectVertex(coordIndex[i++]);
		clickButton(joinVertexKeepBothColors);
	}
}
void InputEvent::joinVertices2(const int* data, int numOf){
	for(int i = 0; i < numOf; ){
		list+=new SelectVertex(data[i+0], data[i+1]);
		i+=2;
		clickButton(remember);
		list+=new SelectVertex(data[i+0], data[i+1]);
		i+=2;
		clickButton(joinVertexKeepBothColors);
	}
}
void InputEvent::projectVertices(const int* data, int numOf, int line){
	for(int i = 0; i < numOf; ){
		list+= new SelectVertex(data[i++], line);
		clickButton(projectSingleVertexOntoPoly);
	}
};

void InputEvent::tugVertices(const GenPolyData* tugData, int numOf){
	for(int i = 0; i < numOf; i++){
		list+=new TugVertex(tugData[i].index, tugData[i].coord, tugData[i].selLine);
	}
}
void InputEvent::tugVertices(const TugData* tugData, int numOf, int selLine){
	for(int i = 0; i < numOf; i++){
		list+=new TugVertex(tugData[i].index, tugData[i].coord, selLine);
	}
}
void InputEvent::tugRings(const TugData* tugData, int numOf){
	for(int i = 0; i < numOf; i++){
		list+=new TugRing(tugData[i].index, tugData[i].coord);
	}
}
void InputEvent::selVertices(const int* tugData, int numOf){
	for(int i = 0; i < numOf; i++){
		list+=new SelectVertex(tugData[i*2+1], tugData[i*2]);
	}
}
void InputEvent::deleteTriangles(const int* tugData, int numOf, int line){
	for(int i = 0; i < numOf; i++){
		if(tugData[i]>=0){
			list+=new SelectTriCornerAnd2d(tugData[i], line);
		}
		clickButton(deleteTriangle);
	}
}
void InputEvent::patchTriangles(const int line, const int* patchData, int numOf){
	for(int i = 0; i < numOf; ){
		list+=new SelectVertex(patchData[i++], line);
		clickButton(remember);
		list+=new SelectTriangle(patchData[i++], line);
		clickButton(quickTriangle);
	}
}
void InputEvent::patchTriangles2(const int* patchData, int numOf){
	for(int i = 0; i < numOf; ){
		list+=new SelectVertex(patchData[i], patchData[i+1]);
		i+=2;
		clickButton(remember);
		list+=new SelectTriangle(patchData[i], patchData[i+1]);
		i+=2;
		clickButton(quickTriangle);
	}
}
void InputEvent::splitMove(const GenPolyData* tugData, int numOf){
	for(int i = 0; i < numOf; i++){
		list+=new SelectTriangle(tugData[i].index, tugData[i].selLine);
		clickButton(splitTriangleImmediate);
		list+=new TugVertex2(tugData[i].index, tugData[i].coord, tugData[i].selLine);
	}
}
void InputEvent::swapTriangles(const SwapTriData* tugData, int numOf){
	for(int i = 0; i < numOf; i++){
		list+=new SelectTriangle(tugData[i].selLine, tugData[i].index);
		clickButton(swapDivide);
	}
}
void InputEvent::paintTriangles(const int* triIndex, int numOf, const int line){
	for(int i = 0; i < numOf; i++){
		list+=new SelectTriangle(triIndex[i], line);
		list+=new TypeString("k");
		//clickButton(pasteColorToTriangle);
	}
}
/*
String createUniqueFileName(String filename){
	int i=0;
	FILE* file=NULL;
	chdir(g_seamlessDir.c_str());
	String str;
	while(1){
		str=filename;
		if(i>0){
			str+=i;
		}
		i++;
		str+=".smls";
		file=fopen(str.c_str(), "rb");
		if(file!=NULL){
			fclose(file);
		}
		else break;
	}
	return str;
}
*/
void setupSequence(){
	
	MultiSelect::reset();
	setSmlsSaved();
	resetTemporaryModes();
	
	g_robotDemo->operation.refValue()=0;
	g_robotDemo->status="";
	g_seamlessDir=g_demoDir;
	g_currentFilename = g_demoDir;
	g_currentFilename += g_robotDemo->fileName;
	g_currentFilename += ".smls";
	_chdir(g_seamlessDir.c_str());

	g_rcp->size=Vec2i(480,200);
	g_rcp->pos=Vec2i(0,g_tb2->pos.y-137-25);
	g_rcp->pixBuf.length=g_rcp->size.x*g_rcp->size.y;
	g_rcp->setupNodeWindow(g_robotDemo);
	Window::activeWindow=g_rcp.p;
	Window::closeAllControlPanelsExceptDemo();
	Window::closeWindow(g_stw.p);

	InputEvent::emptyList();
	InputEvent::sequencerIdx = 0;
	g_robotDemo->function();
	g_scene->children.refValue().length = 0;
	addNodesForNewFile();
	g_stw->genNodeList();
	reset3dScene();
	resetLooseEnds();
	g_colorOwnership = false;
	g_transformCP->sameScreenPosition.refValue() = false;
	g_transformCP->selectObject.refValue() = false;
	g_transformCP->selectRing.refValue() = false;
	g_transformCP->selectRings.refValue() = false;
	g_scene->backgroundColor = Color3f(0,0,0);
	g_scene->scene3DFontColor = Color3f(1,1,0);
	//g_settings->backfaceIgnore = true;
	g_settings->tugRadius = 0;
	g_vcArtTriBase->circle = false;
	g_settings->absoluteRot90 = false;
	InputEvent::syncNumOfActions = 1;
	//g_scene->mode = POSE_MODE | BRUSH_MODE | CREATE_MIRROR_OBJECT;
	//g_wireFrameMode = g_scene->mode & WIREFRAME_MODE;
	setHomeMode();
	toggleBackfaceSelect();
	g_tb2->paint();
	setPasteDefault();
	MessageWin::msg(msg_not_a_movie, MB_OK | MB_ICONEXCLAMATION);
}
void setupSequence2(){
	MultiSelect::reset();
	setSmlsSaved();
	resetTemporaryModes();
	g_seamlessDir=g_demoDir;
	g_currentFilename = g_demoDir;
	g_currentFilename += g_robotDemo->fileName;
	g_currentFilename += ".smls";
	_chdir(g_seamlessDir.c_str());

	HWND hWndDeskTop=GetDesktopWindow();
	RECT deskTopRect;
	GetWindowRect(
		hWndDeskTop,      // handle to window
		&deskTopRect   // address of structure for client coordinates
	);
	if(deskTopRect.right<1024 || deskTopRect.bottom<768){
		msG(msg_screen_res_must_be);
		return;
	}
	if(g_3d->size < Vec2i(1000, 720)){
		msG(msg_3d_win_res_must_be);
		return;
	}
	g_settings->showToolBar = true;
	g_perspective = false;

	g_settings->tugRadius = 0;
	g_settings->incDecRes = .01;
	g_settings->incDecResCtrl = .001;
	g_vcArtTriBase->circle = false;
	g_settings->absoluteRot90 = false;
	InputEvent::syncNumOfActions = 1;
	//updateIconPointers();

	RECT rect;
	GetClientRect(hWndMain, &rect);

	POINT point;
	GetCursorPos(&point);
	InputEvent::pos.x = point.x;
	InputEvent::pos.y = point.y;
	InputEvent::startPos = InputEvent::pos;
	//g_scene->mode = WIREFRAME_MODE;
	//setPoseMode();
	//setPasteDefault();

	//Window::closeAllControlPanelsExceptDemo();
	//Window::closeWindow(g_stw.p);

	InputEvent::robotActive = true;
	
#ifdef ROBOT_SYNC
	InputEvent::beginRobotTime = g_time;;
#else
	InputEvent::beginRobotTime = timeGetTime() * MILISECONDS_TO_SECONDS;
#endif
	//timeID = SetTimer(hWndMain,TIMEID,10,NULL);
	g_forceSaveAs = false;
	setSmlsSaved();
}
/*
void logMathProcessorState(){
	static String str;

	short unsigned int state;
_asm{
	fstcw state
}
	str="math ";
	unsigned int intState=state;
	str.catHex(intState);
	dLog(str);
}
*/

void outputScreenShotForDemo();

void InputEvent::runRobotSequence(){
	if(robotActive){
#ifdef ROBOT_SYNC
		fraction_changed = g_time - beginRobotTime;
#else
		fraction_changed = timeGetTime() * MILISECONDS_TO_SECONDS - beginRobotTime;
#endif
		POINT point;
		GetCursorPos(&point);
		if(abs((int)pos.x-(int)point.x)>3 || abs((int)pos.y-(int)point.y)>3){
			if(resumeSequence()==false){
				return;
			}
			pos.x=point.x;
			pos.y=point.y;
		}
		if(sequencerIdx<list.length){
			if(list[sequencerIdx]->doEvent()){
				syncNumOfActions=1;
				sequencerIdx++;
				beginRobotTime = g_time;
				startPos=pos;
			}
		}
		else{
			_killTimer();
		}
	}
}
bool MoveToF10::doEvent(){
	if(localMove(Vec2i(10, 10))>=0){
		return true;
	}
	return false;
}

bool ToggleField::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
			M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
			for(int k=buttonList.length; --k>=0; ){
				FieldButton* button = buttonList[k];
				if(button->sprop.name==fieldName){
					if(localMove(controlPanel->pos + button->dstPos + Vec2i(button->bSize.x>>1, button->bSize.y>>1))>=0){
						phase++;
						beginRobotTime = g_time;
						phase2=0;
					}
					return false;
				}
			}
		}
	}
	else{
		return doSingleClick(phase2);
	}
	return false;
}
bool ClickYes::doEvent(){
	static int phase2=0;
	ControlPanel* controlPanel = Window::getControlPanel(0);
	if(controlPanel == 0){
		return true;
	}

	if(controlPanel->node != g_questionMessage3Button){
		return true;
	}
	if(phase==0){
		M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
		for(int k=buttonList.length; --k>=0; ){
			FieldButton* button = buttonList[k];
			if(button->sprop.name=="yes"){
				if(localMove(controlPanel->pos + button->dstPos + Vec2i(button->bSize.x>>1, button->bSize.y>>1))>=0){
					phase++;
					beginRobotTime = g_time;
					phase2=0;
				}
				return false;
			}
		}
	}
	else{
		return doSingleClick(phase2);
	}
	return false;
}

bool InputEvent::doMoveToFieldValue(const char* fieldName, int& phase, const int layer){
	if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
		M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
		for(int k=buttonList.length; --k>=0; ){
			FieldButton* button = buttonList[k];
			if(button->sprop.name==fieldName){
				if(localMove(controlPanel->pos + button->valPos + Vec2i(3,15))>=0){
					return true;
				}
				return false;
			}
		}
	}
	else{
		throw;
	}
	_killTimer();
	return false;
}
bool MoveToTextureTile::doEvent(){
	if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
		TextureMapper* textureMapper=dynamic_cast<TextureMapper*>(controlPanel->node.p);
		Vec2i pos=textureMapper->pos + controlPanel->pos;
		pos+=Vec2i(tile.x * 16, (7 - tile.y) * 16) + Vec2i(12, 12);
		if(localMove(pos)>=0){
			return true;
		}
	}
	else throw;
	return false;
}
bool MoveToFieldInc::doEvent(){
	if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
		M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
		for(int k=buttonList.length; --k>=0; ){
			FieldButton* button = buttonList[k];
			if(button->sprop.name==fieldName){
				if(localMove(controlPanel->pos + button->dstPos + Vec2i(buttonOffset*18+10, 6))>=0){
					return true;
				}
				return false;
			}
		}
	}
	else throw;
	return false;
}
bool MoveToFieldDec::doEvent(){
	if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
		M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
		for(int k=buttonList.length; --k>=0; ){
			FieldButton* button = buttonList[k];
			if(button->sprop.name==fieldName){
				if(localMove(controlPanel->pos + button->dstPos + Vec2i(buttonOffset*18+10, 16))>=0){
					return true;
				}
				return false;
			}
		}
	}
	else throw;
	return false;
}
bool InputEvent::doDecIncField(const char* fieldName, const int xOffset, const int yOffset, const int layer){ //16 for dec, 6 for inc 
	if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
		M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
		for(int k=buttonList.length; --k>=0; ){
			FieldButton* button = buttonList[k];
			if(button->sprop.name==fieldName){
				if(localMove(controlPanel->pos + button->dstPos + Vec2i(xOffset, yOffset))>=0){
					return true;
				}
				break;
			}
		}
	}
	return false;
}
bool InputEvent::doClickAndAutoRepeat(int& phase, const int numOfActions){
	static int endOfDownTime=0;
	if(phase==0){
		phase++;
		syncNumOfActions=numOfActions;
		Button::repeatCounter=0;
		mouse_event(
			MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTDOWN,
			0,
			0,
			0,
			0
		);
		leftClick=true;
		outputStatus();
	}
	else if(phase==1){
		if(Button::repeatCounter>numOfActions){
			throw;
		}
		if(Button::repeatCounter==numOfActions && fraction_changed>=downPeriod){
			endOfDownTime=fraction_changed;
			phase++;
			mouse_event(
				MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTUP,
				0,
				0,
				0,
				0
			);
			leftClick=false;
			outputStatus();
		}
	}
	else if(phase==2){
		if(fraction_changed>=endOfDownTime+upPeriod){
			return true;
		}
	}
	return false;
}
bool InputEvent::doVirtKeyDown(const unsigned char vk){
	const float period=.125;

	if(fraction_changed<period){
		return false;
	}

	if(vk == VK_LCONTROL || vk == VK_LSHIFT || vk == VK_LMENU){
		keybd_event(vk, 0, 0, 0);
	}
	else{
		keybd_event(vk, 0x45, KEYEVENTF_EXTENDEDKEY , 0);
	}
	if(vk == VK_SPACE){
		spaceDown = true;
		outputStatus();
	}
	if(vk == VK_LMENU){
		l_altDown = true;
		outputStatus();
	}
	else if(vk == VK_LCONTROL){
		l_ctrlDown = true;
		outputStatus();
	}
	else if(vk == VK_LSHIFT){
		l_shiftDown = true;
		outputStatus();
	}
	if(vk == VK_RMENU){
		r_altDown = true;
		outputStatus();
	}
	else if(vk == VK_RCONTROL){
		r_ctrlDown = true;
		outputStatus();
	}
	else if(vk == VK_RSHIFT){
		r_shiftDown = true;
		outputStatus();
	}
	else if(vk == VK_DOWN){
		arrowDown=true;
		outputStatus();
	}
	return true;
}
bool InputEvent::doVirtKeyUp(const unsigned char vk){
	const float period=.5;
	if(fraction_changed<period){
		return false;
	}

	if(vk == VK_LCONTROL || vk == VK_LSHIFT || vk == VK_LMENU){
		keybd_event(vk, 0, KEYEVENTF_KEYUP, 0);
	}
	else{
		keybd_event(vk, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
	}
	if(vk == VK_SPACE){
		spaceDown = false;
		outputStatus();
	}
	if(vk == VK_LMENU){
		l_altDown = false;
		outputStatus();
	}
	else if(vk == VK_LCONTROL){
		l_ctrlDown = false;
		outputStatus();
	}
	else if(vk == VK_LSHIFT){
		l_shiftDown = false;
		outputStatus();
	}
	if(vk == VK_RMENU){
		r_altDown = false;
		outputStatus();
	}
	else if(vk == VK_RCONTROL){
		r_ctrlDown = false;
		outputStatus();
	}
	else if(vk == VK_RSHIFT){
		r_shiftDown = false;
		outputStatus();
	}
	else if(vk == VK_DOWN){
		arrowDown=false;
		outputStatus();
	}
	return true;
}

bool DecFieldX::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(doDecIncField(fieldName, 0*18+10, 16, layer)){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else{
		return doClickAndAutoRepeat(phase2, numOf);
	}
	return false;
}
bool DecFieldY::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(doDecIncField(fieldName, 1*18+10, 16, layer)){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else{
		return doClickAndAutoRepeat(phase2, numOf);
	}
	return false;
}
bool DecFieldZ::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(doDecIncField(fieldName, 2*18+10, 16, layer)){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else{
		return doClickAndAutoRepeat(phase2, numOf);
	}
	return false;
}

bool IncFieldX::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(doDecIncField(fieldName, 0*18+10, 6, layer)){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else{
		return doClickAndAutoRepeat(phase2, numOf);
	}
	return false;
}
bool IncFieldY::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(doDecIncField(fieldName, 1*18+10, 6, layer)){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else{
		return doClickAndAutoRepeat(phase2, numOf);
	}
	return false;
}
bool IncFieldZ::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(doDecIncField(fieldName, 2*18+10, 6, layer)){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else{
		return doClickAndAutoRepeat(phase2, numOf);
	}
	return false;
}
bool InputEvent::doFieldCtrl(int& phase, const char* fieldName, const int xOffset, const int yOffset, const int numOf, const int layer){
	static int phase2=0;
	if(phase==0){
		if(doVirtKeyDown(VK_LCONTROL)){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else if(phase==1){
		if(doDecIncField(fieldName, xOffset, yOffset, layer)){
			phase++;
			beginRobotTime = g_time;
		}
	}
	else if(phase==2){
		if(doClickAndAutoRepeat(phase2, numOf)){
			phase++;
			beginRobotTime = g_time;
		}
	}
	else{
		return doVirtKeyUp(VK_LCONTROL);
	}
	return false;
}
bool DecFieldCtrlX::doEvent(){
	return doFieldCtrl(phase, fieldName, 0*18+10, 16, numOf, layer);
}
bool DecFieldCtrlY::doEvent(){
	return doFieldCtrl(phase, fieldName, 1*18+10, 16, numOf, layer);
}
bool DecFieldCtrlZ::doEvent(){
	return doFieldCtrl(phase, fieldName, 2*18+10, 16, numOf, layer);
}
bool IncFieldCtrlX::doEvent(){
	return doFieldCtrl(phase, fieldName, 0*18+10, 6, numOf, layer);
}
bool IncFieldCtrlY::doEvent(){
	return doFieldCtrl(phase, fieldName, 1*18+10, 6, numOf, layer);
}
bool IncFieldCtrlZ::doEvent(){
	return doFieldCtrl(phase, fieldName, 2*18+10, 6, numOf, layer);
}


bool IncFieldGroup::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
			M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
			for(int k=buttonList.length; --k>=0; ){
				FieldButton* button = buttonList[k];
				if(button->sprop.name==fieldName){
					if(localMove(controlPanel->pos + Vec2i(260, button->dstPos.y-19))>=0){
						phase++;
						beginRobotTime = g_time;
						phase2=0;
					}
					return false;
				}
			}
		}
	}
	else{
		return doSingleClick(phase2);
	}
	return false;
}
bool DecFieldGroup::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
			M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
			for(int k=buttonList.length; --k>=0; ){
				FieldButton* button = buttonList[k];
				if(button->sprop.name==fieldName){
					if(localMove(controlPanel->pos + Vec2i(260, button->dstPos.y-10))>=0){
						phase++;
						beginRobotTime = g_time;
						phase2=0;
					}
					return false;
				}
			}
		}
	}
	else{
		return doSingleClick(phase2);
	}
	return false;
}
bool SetIndex::doEvent(){
	static int phase2=0;
	static int numOfActions=0;
	if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
		M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
		for(int k=buttonList.length; --k>=0; ){
			FieldButton* button = buttonList[k];
			const Properties& sprop = button->sprop;
			if(sprop.name==fieldName){
				int fieldIndex = controlPanel->node->getFieldGroupIndex(sprop.groupIndex);
				if(phase==0){
					if(target > fieldIndex){
						if(localMove(controlPanel->pos + Vec2i(110, button->dstPos.y-19))>=0){
							phase++;
							beginRobotTime = g_time;
							phase2=0;
							numOfActions = target - fieldIndex;
						}
					}
					else if(target < fieldIndex){
						if(localMove(controlPanel->pos + Vec2i(110, button->dstPos.y-10))>=0){
							phase++;
							beginRobotTime = g_time;
							phase2=0;
							numOfActions = fieldIndex - target;
						}
					}
					else return true;
					return false;
				}
				else{
					return doClickAndAutoRepeat(phase2, numOfActions);
				}
			}
		}
	}
	return false;
}

bool IncIndex::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
			M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
			for(int k=buttonList.length; --k>=0; ){
				FieldButton* button = buttonList[k];
				if(button->sprop.name==fieldName){
					if(localMove(controlPanel->pos + Vec2i(110, button->dstPos.y-19))>=0){
						phase++;
						beginRobotTime = g_time;
						phase2=0;
					}
					return false;
				}
			}
		}
	}
	else{
		return doSingleClick(phase2);
	}
	return false;
}
bool DecIndex::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
			M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
			for(int k=buttonList.length; --k>=0; ){
				FieldButton* button = buttonList[k];
				if(button->sprop.name==fieldName){
					if(localMove(controlPanel->pos + Vec2i(110, button->dstPos.y-10))>=0){
						phase++;
						beginRobotTime = g_time;
						phase2=0;
					}
					return false;
				}
			}
		}
	}
	else{
		return doSingleClick(phase2);
	}
	return false;
}
bool MoveToDecGroupIndex::doEvent(){
	if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
		M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
		for(int k=buttonList.length; --k>=0; ){
			FieldButton* button = buttonList[k];
			if(button->sprop.name==fieldName){
				if(localMove(controlPanel->pos + Vec2i(110, button->dstPos.y-10))>=0){
					return true;
				}
				return false;
			}
		}
	}
	else throw;
	_killTimer();
	return false;
}
bool MoveToIncGroupIndex::doEvent(){
	if(ControlPanel* controlPanel=Window::getControlPanel(layer)){
		M_<Ptr<FieldButton> >& buttonList = controlPanel->buttonList;
		for(int k=buttonList.length; --k>=0; ){
			FieldButton* button = buttonList[k];
			if(button->sprop.name==fieldName){
				if(localMove(controlPanel->pos + Vec2i(110, button->dstPos.y-19))>=0){
					return true;
				}
				return false;
			}
		}
	}
	else throw;
	_killTimer();
	return false;
}

bool DragAnimSlider::doEvent(){
	static bool released=false;
	AnimSlider* as=g_acp->animSlider;

	if(phase==0){
		if(localMove(g_acp->pos + as->dstPos + Vec2i(10, 4))>=0){
			phase++;
			beginRobotTime = g_time;
			startPos = pos;
		}
	}
	else if(phase==1){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		AnimTimeLine* tl=g_acp->timeLine;
		int m = tl->dstPos.x+tl->lineOffset + (value - tl->anim.displayFrom) * tl->animCP.animZoom * static_cast<float>(tl->lineDist);
		if(localMove(g_acp->pos + Vec2i(m, as->dstPos.y + 4))>=0){
			phase++;
			beginRobotTime = g_time;
			released = false;
		}
	}
	else if(phase==3){
		return doLeftButtonUp(released);
	}
	return false;
}

bool DragAnimSelect::doEvent(){
	static bool released=false;
	AnimTimeLine* tl = g_acp->timeLine;
	AnimSlider* as=g_acp->animSlider;

	if(phase==0){
		int m = tl->dstPos.x + tl->lineOffset + (firstValue - tl->anim.displayFrom) * tl->animCP.animZoom * static_cast<float>(tl->lineDist);
		if(localMove(g_acp->pos + Vec2i(m, as->dstPos.y + 4))>=0){
			phase++;
			beginRobotTime = g_time;
			startPos = pos;
		}
	}
	else if(phase==1){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		int m = tl->dstPos.x+tl->lineOffset + (secondValue - tl->anim.displayFrom) * tl->animCP.animZoom * static_cast<float>(tl->lineDist);
		if(localMove(g_acp->pos + Vec2i(m, as->dstPos.y + 4))>=0){
			phase++;
			beginRobotTime = g_time;
			released = false;
		}
	}
	else if(phase==3){
		return doLeftButtonUp(released);
	}
	return false;
}

bool MoveToExitWindow::doEvent(){
	Window* window=Window::getControlPanel(layer);
	if(localMove(window->pos + Vec2i(window->size.x-15, 15))>=0){
		return true;
	}
	return false;
}
bool MoveToExitSceneTree::doEvent(){
	if(localMove(g_stw->pos + Vec2i(g_stw->size.x-15, 15))>=0){
		return true;
	}
	return false;
}
bool MoveToTitleBar::doEvent(){
	if(localMove(Window::getControlPanel(layer)->pos + Vec2i(100, 15))>=0){
		return true;
	}
	return false;
}
bool MoveRelWindow::doEvent(){
	if(localMove(Window::getControlPanel(layer)->pos + to)>=0){
		return true;
	}
	return false;
}
bool MoveToNodeNameAsChild::doEvent(){
	int nodeLine=getLineByNodeName(name);
	int collapsedNodeLine;
	getCollapsedLineNoGlobalChange(nodeLine, collapsedNodeLine);
	if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeLine].indent*pixSymbolCX+pixSymbolCX+8,	SceneTree::beginYOffset + (collapsedNodeLine-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8))>=0){
		return true;
	}
	return false;
}
bool MoveToNodeLable::doEvent(){
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
	if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+charIdx*8,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8))>=0){
		return true;
	}
	return false;
}


bool InputEvent::doMoveToNodeAsChild(int& phase, const int nodeIdx){//phase must be initialised with 0 before calling this function for the first time
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);

	int nodeMax=(g_stw->size.y-g_stw->beginYOffset)/pixSymbolCY+1;
	int nodeCnt=nodeMax-1;
	int lastNode=g_scene->scroll+nodeCnt;
	if(phase==0){
		if(collapsedNodeIdx < lastNode){
			if(collapsedNodeIdx < g_scene->scroll){
				phase=3;
			}
			else{
				phase=5;
			}
		}
		else{
			phase=1;
		}
	}
	if(phase==1){
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+8,	g_stw->size.y-8))>=0){
			phase=2;
		}
	}
	else if(phase==2){
		if(collapsedNodeIdx < lastNode){
			phase=5;
			beginRobotTime = g_time;
			startPos=pos;
		}
	}
	else if(phase==3){
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+8,	8))>=0){
			phase=4;
		}
	}
	else if(phase==4){
		if(collapsedNodeIdx >= g_scene->scroll){
			phase=5;
			beginRobotTime = g_time;
			startPos=pos;
		}
	}
	else if(phase==5){
		getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+8,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8))>=0){
			return true;
		}
	}
	return false;
}
bool InputEvent::doMoveToNodeAsSource(int& phase, const int nodeIdx){//phase must be initialised with 0 before calling this function for the first time
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);

	int nodeMax=(g_stw->size.y-g_stw->beginYOffset)/pixSymbolCY+1;
	int nodeCnt=nodeMax-1;
	int lastNode=g_scene->scroll+nodeCnt;
	if(phase==0){
		if(collapsedNodeIdx < lastNode){
			if(collapsedNodeIdx < g_scene->scroll){
				phase=3;
			}
			else{
				phase=5;
			}
		}
		else{
			phase=1;
		}
	}
	if(phase==1){
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+8,	g_stw->size.y-8))>=0){
			phase=2;
		}
	}
	else if(phase==2){
		if(collapsedNodeIdx < lastNode){
			phase=5;
			beginRobotTime = g_time;
			startPos=pos;
		}
	}
	else if(phase==3){
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+8,	8))>=0){
			phase=4;
		}
	}
	else if(phase==4){
		if(collapsedNodeIdx >= g_scene->scroll){
			phase=5;
			beginRobotTime = g_time;
			startPos=pos;
		}
	}
	else if(phase==5){
		getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+12,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 4))>=0){
			return true;
		}
	}
	return false;
}
bool InputEvent::doMoveToNodeAsDest(int& phase, const int nodeIdx){//phase must be initialised with 0 before calling this function for the first time
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);

	int nodeMax=(g_stw->size.y-g_stw->beginYOffset)/pixSymbolCY+1;
	int nodeCnt=nodeMax-1;
	int lastNode=g_scene->scroll+nodeCnt;
	if(phase==0){
		if(collapsedNodeIdx < lastNode){
			if(collapsedNodeIdx < g_scene->scroll){
				phase=3;
			}
			else{
				phase=5;
			}
		}
		else{
			phase=1;
		}
	}
	if(phase==1){
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+8,	g_stw->size.y-8))>=0){
			phase=2;
		}
	}
	else if(phase==2){
		if(collapsedNodeIdx < lastNode){
			phase=5;
			beginRobotTime = g_time;
			startPos=pos;
		}
	}
	else if(phase==3){
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+8,	8))>=0){
			phase=4;
		}
	}
	else if(phase==4){
		if(collapsedNodeIdx >= g_scene->scroll){
			phase=5;
			beginRobotTime = g_time;
			startPos=pos;
		}
	}
	else if(phase==5){
		getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+12,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 12))>=0){
			return true;
		}
	}
	return false;
}
bool InputEvent::doMoveToNodeAsReplacement(int& phase, const int nodeIdx){//phase must be initialised with 0 before calling this function for the first time
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);

	int nodeMax=(g_stw->size.y-g_stw->beginYOffset)/pixSymbolCY+1;
	int nodeCnt=nodeMax-1;
	int lastNode=g_scene->scroll+nodeCnt;
	if(phase==0){
		if(collapsedNodeIdx < lastNode){
			if(collapsedNodeIdx < g_scene->scroll){
				phase=3;
			}
			else{
				phase=5;
			}
		}
		else{
			phase=1;
		}
	}
	if(phase==1){
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+8,	g_stw->size.y-8))>=0){
			phase=2;
		}
	}
	else if(phase==2){
		if(collapsedNodeIdx < lastNode){
			phase=5;
			beginRobotTime = g_time;
			startPos=pos;
		}
	}
	else if(phase==3){
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX+8,	8))>=0){
			phase=4;
		}
	}
	else if(phase==4){
		if(collapsedNodeIdx >= g_scene->scroll){
			phase=5;
			beginRobotTime = g_time;
			startPos=pos;
		}
	}
	else if(phase==5){
		getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX - 7,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8))>=0){
			return true;
		}
	}
	return false;
}
bool MoveToNodeAsChild::doEvent(){
	return doMoveToNodeAsChild(phase, nodeIdx);
}
bool MoveToNodeAsSource::doEvent(){
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
	if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX*2,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8 - 4))>=0){
		return true;
	}
	return false;
}
bool MoveToNodeAsDest::doEvent(){
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
	if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+pixSymbolCX*2,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 12))>=0){
		return true;
	}
	return false;

}
bool InputEvent::doMoveToNodeAsSiblingAbove(const int nodeIdx){
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
	if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+8,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8 - 4))>=0){
		return true;
	}
	return false;
}
bool InputEvent::doMoveToNodeAsSiblingBelow(const int nodeIdx){
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
	if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+8,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8 + 4))>=0){
		return true;
	}
	return false;
}
bool MoveToNodeAsSiblingAbove::doEvent(){
	return doMoveToNodeAsSiblingAbove(nodeIdx);
}
bool MoveToNodeAsReplacement::doEvent(){
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);

	int nodeMax=(g_stw->size.y-g_stw->beginYOffset)/pixSymbolCY+1;
	int nodeCnt=nodeMax-1;
	int lastNode=g_scene->scroll+nodeCnt;
	if(phase==0){
		if(collapsedNodeIdx < lastNode){
			phase=3;
		}
		else{
			phase=1;
		}
	}
	if(phase==1){
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX - 7,	g_stw->size.y-7))>=0){
			phase=2;
		}
	}
	else if(phase==2){
		if(collapsedNodeIdx < lastNode){
			phase=3;
			beginRobotTime = g_time;
			startPos=pos;
		}
	}
	else if(phase==3){
		getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX - 7,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8))>=0){
			return true;
		}
	}
	return false;
}
bool MoveToNodeAsSiblingBelow::doEvent(){
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
	if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX+8,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8 + 4))>=0){
		return true;
	}
	return false;
}

bool InputEvent::doMoveToNode(const int line){
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(line, collapsedNodeIdx);
	if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[line].indent*pixSymbolCX + 10,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8))>=0){
		return true;
	}
	return false;
}
bool MoveToNode::doEvent(){
	return doMoveToNode(nodeIdx);
}
bool InputEvent::doSingleClick(int& phase){
	if(phase==0){
		phase++;
		mouse_event(
			MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTDOWN,
			0,
			0,
			0,
			0
		);
		leftClick=true;
		outputStatus();

	}
	else if(phase==1){
		if(fraction_changed>=downPeriod){
			phase++;
			mouse_event(
				MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTUP,
				0,
				0,
				0,
				0
			);
			leftClick=false;
			outputStatus();
		}
	}
	else if(phase==2){
		if(fraction_changed>=downPeriod+upPeriod){
			return true;
		}
	}
	return false;
}

bool InputEvent::doDoubleClick(int& phase){
	static const float wholePeriod=.5;
	Vec2i sysPos=getSysMouseCoord();

	if(phase==0){
		phase=1;
		Vec2i sysPos=getSysMouseCoord();
		mouse_event(
			MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTDOWN,
			sysPos.x,
			sysPos.y,
			0,
			0
		);
		doubleClick = true;
		outputStatus();
	}
	else if(phase==1 && fraction_changed > wholePeriod*.25){
		phase=2;
		mouse_event(
			MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTUP,
			sysPos.x,
			sysPos.y,
			0,
			0
		);
	}
	else if(phase==2 && fraction_changed > wholePeriod*.5){
		phase=3;
		mouse_event(
			MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTDOWN,
			sysPos.x,
			sysPos.y,
			0,
			0
		);
	}
	else if(phase==3 && fraction_changed > wholePeriod*.75){
		phase=4;
		mouse_event(
			MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTUP,
			sysPos.x,
			sysPos.y,
			0,
			0
		);
		doubleClick=false;
		outputStatus();
	}
	else if(phase==4 && fraction_changed >= wholePeriod){
		return true;
	}
	return false;
}
bool Hide::doEvent(){
	static int phase2=0;
	if(phase==0){
		int collapsedNodeIdx;
		getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX + 10,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8))>=0){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else{
		return doDoubleClick(phase2);
	}
	return false;
}
bool SelectNode::doEvent(){
	static int phase2=0;
	if(phase==0){
		int collapsedNodeIdx;
		getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX + 10,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8))>=0){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else{
		return doSingleClick(phase2);
	}
	return false;
}

bool SelectNodeByName::doEvent(){
	static int phase2=0;
	if(phase==0){
		int collapsedNodeIdx;
		int nodeIdx=getLineByNodeName(name);
		getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent*pixSymbolCX + 10,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8))>=0){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else{
		return doSingleClick(phase2);
	}
	return false;
}

bool NameNode::doEvent(){
	static int phase2=0;

	if(phase==0){
		phase2=0;
		phase++;
	}
	if(phase==1){
		if(doMoveToNodeAsChild(phase2, nodeIdx)){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else if(phase==2){
		if(doSingleClick(phase2)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else{//(phase==3)
		return doTypeString(name, charIdx);
	}
	return false;
}
bool ReplacePartialCode::doEvent(){
	static int phase2=0;
	static bool released=false;
	int collapsedNodeIdx;
	getCollapsedLineNoGlobalChange(line, collapsedNodeIdx);
	Vec2i targetPos=g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[line].indent*pixSymbolCX + 16 + offset*8,	SceneTree::beginYOffset + (collapsedNodeIdx-g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8);

	if(phase==0){
		if(localMove(targetPos)>=0){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(localMove(targetPos+Vec2i(numOf*8, 0))>=0){
			phase2=0;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==3){
		if(doLeftButtonUp(released)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==4){
		return doTypeString(str, charIdx);
	}
	return false;
}

bool Collapse::doEvent(){
	static int phase2=0;
	if(phase==0){
		int collapsedNodeIdx;
		getCollapsedLineNoGlobalChange(nodeIdx, collapsedNodeIdx);
		if(localMove(g_stw->pos + Vec2i(tLnOffst + g_sceneTreeList[nodeIdx].indent * pixSymbolCX - 7,	SceneTree::beginYOffset + (collapsedNodeIdx - g_stw->getScrollPosCollapsed()) * pixSymbolCY + 8)) >= 0){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else{
		return doSingleClick(phase2);
	}
	return false;
}

extern Ptr<InsertComplete> g_insertComplete;

bool SelNurbsLine::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;
	if(phase == 0){
		if(Seamless* smls = getSeamless2()){
			M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
			for(int k = cpNodeList.length; --k>=0;){
				NurbsPatch* patchA = cpNodeList[k];
				if(idxA >= patchA->fromPoint){
					Vec3f vecA = g_scene->getViewMatrix() * patchA->dpoint.refValue()[idxA - patchA->fromPoint];
					while(k >= 0){
						NurbsPatch* patchB = cpNodeList[k];
						if(idxB >= patchB->fromPoint){
							Vec3f vecB = g_scene->getViewMatrix() * patchB->dpoint.refValue()[idxB - patchB->fromPoint];
							startPos3d = vecA * (1.0 - ratio) + vecB * ratio;
							break;
						}
						--k;
					}
					break;
				}
			}
		}
		Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos) >= 0){
			phase++;
			outputStatus();
			beginRobotTime = g_time;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation = startPos3d;
			SenseTugGlobals::initTug();
			//if(g_insertPatch.imousePhase == InsertPatch::NEUTRAL || g_insertPatch.imousePhase == InsertPatch::DIVERT_STANDBY){
			//	SenseTugGlobals::initTug();
			//}
			//else{
			//	g_insertPatch.imousePhase = InsertPatch::DIVERT_STANDBY;
			//	g_stw->cp->setupNodeWindow(g_insertComplete);
			//	Window::activeWindow = g_stw->cp.p;
			//	g_insertPatch.addInsertVisualClue();
			//}
		}
	}
	else if(fraction_changed>=.25){
		r_ctrlDown = false;
		UnAcquire();
		resetDragActuallyHappned();
		return true;
	}
	return false;
}


bool AddCtrlPnt::doEvent(){
	Vec2i mouse2dPos = get2dMousePosFrom3d(Vec3f(to.x, to.y, 0));

	if(phase == 0){
		if(localMove(mouse2dPos) >= 0){
			phase++;
			outputStatus();
			beginRobotTime = g_time;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation.x = to.x;
			g_pointer.translation.y = to.y;
			if(g_insertPatch.imousePhase == InsertPatch::NEUTRAL){
				SenseTugGlobals::initTug();
			}
			else{
				g_insertPatch.imousePhase = InsertPatch::DIVERT_STANDBY;
				g_stw->cp->setupNodeWindow(g_insertComplete);
				Window::activeWindow = g_stw->cp.p;
				g_insertPatch.addInsertVisualClue();
			}
		}
	}
	else if(fraction_changed>=.25){
		//r_ctrlDown = false;
		UnAcquire();
		resetDragActuallyHappned();
		return true;
	}
	return false;
}

bool AddDiversion::doEvent(){
	Vec2i mouse2dPos = get2dMousePosFrom3d(Vec3f(to.x, to.y, 0));

	if(phase == 0){
		if(localMove(mouse2dPos) >= 0){
			phase++;
			outputStatus();
			beginRobotTime = g_time;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation.x = to.x;
			g_pointer.translation.y = to.y;
			SenseTugGlobals::initTug();
		}
	}
	else if(fraction_changed>=.25){
		r_ctrlDown = false;
		UnAcquire();
		resetDragActuallyHappned();
		return true;
	}
	return false;
}

bool RectPoly::doEvent(){
	Vec2i mouse2dPos=get2dMousePosFrom3d(Vec3f(from.x, from.y, 0));
	Vec2i mouse2dToPos=get2dMousePosFrom3d(Vec3f(to.x, to.y, 0));

	if(phase == 0){
		if(localMove(mouse2dPos)>=0){
			phase++;
			beginRobotTime = g_time;


			g_pointer.translation=Vec3f(from.x, from.y, 0);
			//mouse3dMode=true;
			//SetAcquire();
		}
	}
	else if(phase == 1){
		if(doVirtKeyDown(VK_RSHIFT)){
			phase++;
			beginRobotTime = g_time;

			if(MultiSelect::triList.length){
				MultiSelect::reset();
			}
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			MultiSelect::topLeft = MultiSelect::bottomRight;

			Window::activeWindow=g_3d.p;

			startPos3d=MultiSelect::bottomRight=MultiSelect::topLeft=g_pointer.translation;

			MultiSelect::reset();
			MultiSelect::selectRectMode = true;

			MultiSelect::bottomRight=MultiSelect::topLeft=g_pointer.translation;
			g_vcSelectRectGroup->children.refValue().length=0;
			g_vcSelectRectGroup->children.refValue()+=g_vcSelectRectShape.p;
			g_visualClue = g_vcSelectRectGroup;
			houseKeepVisualClue();
		}
	}
	else if(phase == 2){
		float distance=moveTo3dPoint(to);

		Vec3f v = g_pointer.translation;
		MultiSelect::bottomRight=v;
//						Patch::bottomRight+=Vec3f(dims.lX*b, -dims.lY*b, 0);

		houseKeepVisualClue();
		
		
		//		String str;
//		g_pointer.translation.toString(str);
//		dLog(str);
		if(distance>0){
			phase++;
			beginRobotTime = g_time;
		}
	}
	else{
		if(doVirtKeyUp(VK_RSHIFT)){
			Vec2i vec=get2dMousePosFrom3d(g_pointer.translation);
			WINDOWINFO wi;
			GetWindowInfo(hWndMain, &wi);
			SetCursorPos(vec.x+wi.rcClient.left, vec.y+wi.rcClient.top); 

			UnAcquire();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}

bool ClickControlPoint::doEvent(){
	if(NurbsPatch* nurbsPatch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		startPos3d=nurbsPatch->dpoint[pointIndex];
		Vec2i mouse2dPos=get2dMousePosFrom3d(g_scene->getViewMatrix()*(nurbsPatch->matrix*startPos3d));

		if(localMove(mouse2dPos)>=0){
			g_pointer.translation=startPos3d;
			tugFlag=true;
			Window::activeWindow=g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			nurbsPatch->index[0]=pointIndex;

			g_stw->cp->textCell.cellActive=false;
			g_selNode->processFields();
			g_stw->cp->paint();
			fastFullBuildForAutoIn(false);
			return true;
		}
	}
	else if(NurbsLathe* nurbsLathe=dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
		Vec2f vec=nurbsLathe->dpoint[pointIndex];
		startPos3d=Vec3f(vec.x, vec.y, 0);
		Vec2i mouse2dPos=get2dMousePosFrom3d(g_scene->getViewMatrix()*(nurbsLathe->matrix*startPos3d));

		if(localMove(mouse2dPos) >= 0){
			tugFlag = true;
			Window::activeWindow = g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			nurbsLathe->index[0]=pointIndex;

			g_stw->cp->textCell.cellActive = false;
			g_selNode->processFields();
			g_stw->cp->paint();
			fastFullBuildForAutoIn(false);
			return true;
		}
	}
	else if(Bezier* bezier=dynamic_cast<Bezier*>(g_selNode.getRaw())){
		Vec2f vec=bezier->point[pointIndex];
		startPos3d=Vec3f(vec.x, vec.y, 0);
		Vec2i mouse2dPos=get2dMousePosFrom3d(g_scene->getViewMatrix()*(bezier->matrix*startPos3d));

		if(localMove(mouse2dPos)>=0){
			tugFlag=true;
			Window::activeWindow=g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			bezier->index[0] = pointIndex;

			g_stw->cp->textCell.cellActive = false;
			g_selNode->processFields();
			g_stw->cp->paint();
			fastFullBuildForAutoIn(false);
			return true;
		}
	}
	return false;
}
bool MoveToControlPoint::doEvent(){
	if(NurbsPatch* nurbsPatch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		startPos3d=nurbsPatch->dpoint[pointIndex];
		Vec2i mouse2dPos=get2dMousePosFrom3d(g_scene->getViewMatrix()*(nurbsPatch->matrix*startPos3d));

		if(localMove(mouse2dPos)>=0){
			return true;
		}
	}
	else if(NurbsLathe* nurbsLathe=dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
		Vec2f vec=nurbsLathe->dpoint[pointIndex];
		startPos3d=Vec3f(vec.x, vec.y, 0);
		Vec2i mouse2dPos=get2dMousePosFrom3d(g_scene->getViewMatrix()*(nurbsLathe->matrix*startPos3d));

		if(localMove(mouse2dPos)>=0){
			return true;
		}
	}
	else if(Bezier* bezier=dynamic_cast<Bezier*>(g_selNode.getRaw())){
		Vec2f vec=bezier->point[pointIndex];
		startPos3d=Vec3f(vec.x, vec.y, 0);
		Vec2i mouse2dPos=get2dMousePosFrom3d(g_scene->getViewMatrix()*(bezier->matrix*startPos3d));

		if(localMove(mouse2dPos)>=0){
			return true;
		}
	}
	return false;
}

extern Ptr<PartCageClue> g_partCageClue;

bool TugCP::doEvent(){
	if(prevRenderCounter == renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;
	if(phase == 0){
		if(Seamless* smls = getSeamless2()){
			M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
			for(int k = cpNodeList.length; --k>=0;){
				NurbsPatch* patch = cpNodeList[k];
				if(pointIndex >= patch->fromPoint){
					startPos3d = g_scene->getViewMatrix() * patch->dpoint.refValue()[pointIndex-patch->fromPoint];
					break;
				}
			}
		}
		Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			phase = 1;
			beginRobotTime = g_time;
			Window::activeWindow=g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation=startPos3d;
			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance = moveTo3dPoint(to);

		VisualAid* va = dynamic_cast<VisualAid*>(g_selNode.getRaw());
		if(g_transformCP->selectObject){
			if(Build* build = dynamic_cast<Build*>(g_selNode.getRaw())){
				if(NurbsPatch* patch = build->getPatch()){
					va = patch;
				}
			}
		}

		va->drag(g_pointer.translation);
		if(distance > 0){
			
			UnAcquire();

			if(Anim* anim = Window::animMode()){
				if(Part* part = dynamic_cast<Part*>(g_selNode.getRaw())){
					Seamless* smls = part->smls;
					if(smls->workAnimCP && g_partCageClue->cpTug){
						anim->captureMorph(smls);
					}
				}
			}

			resetDragActuallyHappned();

			return true;
		}
	}
	return false;
}


bool Scale3d::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;

	if(phase == 0){
		l_altDown = true;
		outputStatus();
		phase++;
	}
	if(phase <= 1){
		if(NurbsPatch* patch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
			Seamless* smls = patch->smls;
			M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
			for(int k = cpNodeList.length; --k>=0;){
				patch = cpNodeList[k];
				if(pointIndex >= patch->fromPoint){
					startPos3d = g_scene->getViewMatrix()*patch->dpoint.refValue()[pointIndex-patch->fromPoint];
					break;
				}
			}
		}
		Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			phase++;
			beginRobotTime = g_time;
			Window::activeWindow=g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation=startPos3d;
			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance=moveTo3dPoint(to);

		VisualAid* va=dynamic_cast<VisualAid*>(g_selNode.getRaw());
		va->drag(g_pointer.translation);
		if(distance>0){
			UnAcquire();
			l_altDown = false;
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}

bool TugVertex::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;
	if(phase == 0){
		if(NurbsPatch* nurbsPatch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
			startPos3d=g_scene->getViewMatrix()*nurbsPatch->dpoint.refValue()[coordIndex];
		}
		else{
			Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
			if(nodeLine>=0){
				part=dynamic_cast<Art*>(g_sceneTreeList[nodeLine].node.p);
			}
			Seamless* smls = part->smls;
			startPos3d = part->pivotPointMatrix * (part->coord.refValue()[coordIndex] - part->absolutePivotPoint);
		}
		Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			phase = 1;
			beginRobotTime = g_time;
			Window::activeWindow=g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation=startPos3d;
			SenseTugGlobals::initTug();
		}
	}
	else{
		Seamless* test = 0;
		float distance=moveTo3dPoint(to);

		VisualAid* va=dynamic_cast<VisualAid*>(g_selNode.getRaw());
		va->drag(g_pointer.translation);
		if(distance>0){
			if(Anim* anim = Window::animMode()){
				if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
					Seamless* smls = part->smls;
					test = smls;
					if(smls->animCP == false){
						anim->capturePolyMorph(smls);
						tugFlag2 = tugFlag = false;
						//anim->update = true;
					}
					else if(smls->workAnimCP){
						anim->captureMorph(smls);
					}
				}
			}
			UnAcquire();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}
bool TugVertex2::doEvent(){
	if(phase == 0){
		Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
		if(nodeLine>=0){
			part=dynamic_cast<Art*>(g_sceneTreeList[nodeLine].node.p);
		}
		Seamless* smls=part->smls;
		startPos3d=part->pivotPointMatrix * (part->coord.refValue()[part->triangle.refIndex(coordIndexIndex/3*3+(coordIndexIndex+1)%3)] - part->absolutePivotPoint);//the selcted part must be the same as refPart for triangle
		Vec2i mouse2dPos=get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			phase = 1;
			beginRobotTime = g_time;
			Window::activeWindow=g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation=startPos3d;
			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance=moveTo3dPoint(to);

		VisualAid* va=dynamic_cast<VisualAid*>(g_selNode.getRaw());
		va->drag(g_pointer.translation);
		if(distance>0){
			UnAcquire();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}
bool SelectVertex::doEvent(){
	if(clickTime == 0){
		Node* node = g_selNode.getRaw();
		if(nodeLine>=0){
			node=g_sceneTreeList[nodeLine].node.p;
		}
		if(Art* part = dynamic_cast<Art*>(node)){
			Seamless* smls=part->smls;
			startPos3d=part->pivotPointMatrix * (part->coord.refValue()[coordIndex] - part->absolutePivotPoint);
			Vec2i mouse2dPos=get2dMousePosFrom3d(startPos3d);
			if(localMove(mouse2dPos)>=0){
				clickTime=fraction_changed;
				Window::activeWindow=g_3d.p;
				SetAcquire();
				mouse3dMode = true;
				showCursor(FALSE);
				g_pointer.translation=startPos3d;
				SenseTugGlobals::initTug();
			}
		}
		else if(NurbsLathe* nurbsLathe = dynamic_cast<NurbsLathe*>(node)){
			Vec2f vec=nurbsLathe->dpoint[coordIndex];
			startPos3d=Vec3f(vec.x, vec.y, 0);
			Vec2i mouse2dPos=get2dMousePosFrom3d(g_scene->getViewMatrix()*(nurbsLathe->matrix*startPos3d));

			if(localMove(mouse2dPos)>=0){
				nurbsLathe->index[0]=coordIndex;
				clickTime=fraction_changed;
				Window::activeWindow=g_3d.p;
				SetAcquire();
				mouse3dMode = true;
				showCursor(FALSE);
				g_pointer.translation=startPos3d;
				SenseTugGlobals::initTug();
				tugFlag=true;
				nurbsLathe->houseKeep();
			}
		}
	}
	else{
		if(fraction_changed>=clickTime+.25){
			UnAcquire();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}
bool SelectTriCornerAnd2d::doEvent(){
	if(clickTime == 0){
		l_altDown = true;
		outputStatus();
		Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
		if(nodeLine>=0){
			part=dynamic_cast<Art*>(g_sceneTreeList[nodeLine].node.p);
		}
		Seamless* smls=part->smls;

		int idx=coordIdxIdx/3*3;
		int corner=coordIdxIdx%3;

		Art* partA = part->triangle.refPart(idx);
		Art* partB = part->triangle.refPart(idx+1);
		Art* partC = part->triangle.refPart(idx+2);

		Vec3f v[3]={
			partA->pivotPointMatrix * (partA->coord[part->triangle.refIndex(idx)] - partA->absolutePivotPoint),
			partB->pivotPointMatrix * (partB->coord[part->triangle.refIndex(idx+1)] - partB->absolutePivotPoint),
			partC->pivotPointMatrix * (partC->coord[part->triangle.refIndex(idx+2)] - partC->absolutePivotPoint)
		};

		startPos3d=Vec3f();
		startPos3d += v[corner]*.60;
		startPos3d += v[(corner+1)%3]*.30;
		startPos3d += v[(corner+2)%3]*.10;


		startPos3d.z=0;
		Vec2i mouse2dPos=get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			clickTime=fraction_changed;
			Window::activeWindow=g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation=startPos3d;
			SenseTugGlobals::initTug();
		}
	}
	else{
		if(fraction_changed >= clickTime + .25){
			UnAcquire();
			l_altDown = false;
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}

bool SelectTriangle::doEvent(){
	if(clickTime == 0){
		l_altDown = true;
		outputStatus();
		Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
		if(nodeLine>=0){
			part=dynamic_cast<Art*>(g_sceneTreeList[nodeLine].node.p);
		}
		Seamless* smls=part->smls;

		int idx=coordIdxIdx/3*3;
		int corner=coordIdxIdx%3;

		Art* partA = part->triangle.refPart(idx);
		Art* partB = part->triangle.refPart(idx+1);
		Art* partC = part->triangle.refPart(idx+2);

		Vec3f v[3]={
			partA->pivotPointMatrix * (partA->coord[part->triangle.refIndex(idx)] - partA->absolutePivotPoint),
			partB->pivotPointMatrix * (partB->coord[part->triangle.refIndex(idx+1)] - partB->absolutePivotPoint),
			partC->pivotPointMatrix * (partC->coord[part->triangle.refIndex(idx+2)] - partC->absolutePivotPoint)
		};

		startPos3d=Vec3f();
		/*
		for(int i=0; i<3; i++){
			if(i != corner){
				startPos3d += (v[i] - v[corner]) * .1;
			}
		}
		startPos3d += v[corner];
		*/

		startPos3d += v[corner]*.45;
		startPos3d += v[(corner+1)%3]*.45;
		startPos3d += v[(corner+2)%3]*.1;

		startPos3d.z=0;
		Vec2i mouse2dPos=get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			clickTime=fraction_changed;
			Window::activeWindow=g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation=startPos3d;
//			keybd_event(VK_LMENU, 0, 0, 0);
			SenseTugGlobals::initTug();
//			altDown=false;
//			keybd_event(VK_LMENU, 0, KEYEVENTF_KEYUP, 0);
//			outputStatus();
		}
	}
	else{
		if(fraction_changed>=clickTime+.25){
			UnAcquire();
			l_altDown = false;
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}


bool Click3d::doEvent(){
	if(clickTime == 0){

		startPos3d = Vec3f();

		startPos3d.z=0;
		Vec2i mouse2dPos=get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			clickTime=fraction_changed;
			Window::activeWindow=g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation=startPos3d;
//			keybd_event(VK_LMENU, 0, 0, 0);
			SenseTugGlobals::initTug();
//			altDown=false;
//			keybd_event(VK_LMENU, 0, KEYEVENTF_KEYUP, 0);
//			outputStatus();
		}
	}
	else{
		if(fraction_changed>=clickTime+.25){
			UnAcquire();
			l_altDown = false;
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}

bool DragViewRot::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter = renderCounter;
	static Vec2i mouseFrom;
	Vec3f mouseFrom3d = Vec3f(mouseFromX, mouseFromY, 0);
	if(phase==0){
		from=g_scene->rotation;
		phase++;
		mouseFrom = get2dMousePosFrom3d(mouseFrom3d);
	}
	if(phase==1){
		if(localMove(mouseFrom)>=0){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else{
		if(fraction_changed < 2){
			float interp = 1.0/2*(fraction_changed);
			interp=1.0f-(cosf(interp*PI)*.5+.5);

			g_scene->rotation = from.slerp(to, interp);

			pos = g_mousePos; //done only to make cursor change for robot, not needed when manually done
			WINDOWINFO wi;
			GetWindowInfo(hWndMain, &wi);
			pos += Vec2i(wi.rcClient.left, wi.rcClient.top);
			SetCursorPos(pos.x, pos.y); 

			houseKeepVisualClueViewChange();
		}
		else{
			g_scene->rotation = to;
			houseKeepVisualClueViewChange();
			return true;
		}
	}
	return false;
}

bool DragViewPos::doEvent(){
	if(prevRenderCounter == renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;
	static Vec2i mouseFrom;
	Vec3f mouseFrom3d = Vec3f(mouseFromX, mouseFromY, 0);
	if(phase == 0){
		from = g_scene->translation;
		phase++;
		mouseFrom = get2dMousePosFrom3d(mouseFrom3d);
	}
	if(phase==1){
		if(localMove(mouseFrom) >= 0){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else{
		if(fraction_changed < 2){
			float interp = 1.0 / 2 * fraction_changed;
			interp = 1.0f - (cosf(interp*PI) *.5 +.5);

			g_scene->translation = from + (to - from) * interp;
			houseKeepVisualClueViewChange();

			g_mousePos = get2dMousePosFrom3d(g_scene->scale * (g_scene->rotation * (g_scene->translation - from)) + mouseFrom3d);
			pos = g_mousePos;
			WINDOWINFO wi;
			GetWindowInfo(hWndMain, &wi);
			pos += Vec2i(wi.rcClient.left, wi.rcClient.top);
			SetCursorPos(pos.x, pos.y); 
		//localMove(g_mousePos);
		}
		else{
			g_scene->translation = to;
			houseKeepVisualClueViewChange();
			return true;
		}
	}
	return false;
}

bool DemoMove::doEvent(){
	if(prevRenderCounter == renderCounter){
		return false;
	}
	prevRenderCounter = renderCounter;
	static Vec2i mouseFrom;
	Vec3f mouseFrom3d = Vec3f(mouseFromX, mouseFromY, 0);
	if(phase==0){
		from=g_scene->translation;
		phase++;
		mouseFrom=get2dMousePosFrom3d(mouseFrom3d);
	}
	if(phase == 1){
		if(localMove(mouseFrom)>=0){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else{
		return true;
	}
	return false;
}
//void hoverWin3d();
bool MouseZoom::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;
	static Vec2i mouseFrom;
	Vec3f mouseFrom3d = Vec3f(mouseFromX, mouseFromY, 0);

	if(phase==0){
		from=g_scene->scale.x;
		phase++;
		mouseFrom = get2dMousePosFrom3d(mouseFrom3d);
	}
	if(phase==1){
		if(localMove(mouseFrom)>=0){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else{
		if(fraction_changed < 2){
			float interp = 1.0 / 2 * (fraction_changed);
			interp = 1.0f-(cosf(interp * PI) *.5 + .5);

			float zoom = from+(to-from)*interp;
			g_scene->setZoom(zoom);

			pos = g_mousePos; //done only to make cursor change for robot, not needed when manually done
			WINDOWINFO wi;
			GetWindowInfo(hWndMain, &wi);
			pos += Vec2i(wi.rcClient.left, wi.rcClient.top);
			SetCursorPos(pos.x, pos.y); 


			houseKeepVisualClueViewChange();
			//SetCursor(g_currentType2);

			//dLog("robot");
			//hoverWin3d();
		}
		else{
			g_scene->setZoom(to);
			houseKeepVisualClueViewChange();
			//SetCursor(g_currentType2);
			//hoverWin3d();

			return true;
		}
	}
	return false;
}

bool MovePointer3d::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;
	static Vec2i mouseFrom;
	Vec3f mouseFrom3d = Vec3f(mouseFromX, mouseFromY, 0);

	if(phase==0){
		phase++;
		mouseFrom=get2dMousePosFrom3d(mouseFrom3d);
	}
	if(phase==1){
		if(localMove(mouseFrom)>=0){
			return true;
		}
	}
	return false;
}


bool TugRing::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;
	if(phase == 0){
		RingBuild* ringBuild=dynamic_cast<RingBuild*>(g_selNode.getRaw());
		Art* part = ringBuild->part[0].p;
		Seamless* smls=part->smls;
		const int& staves = ringBuild->staves;

		M_Vec2f& scaleX = ringBuild->scaleX.refValue();
		M_Vec2f& scaleZ = ringBuild->scaleZ.refValue();

		int i= ringPointIndex/5;


		Vec3f sr;
		switch(ringPointIndex % 5){
			case 0:
				sr= Vec3f(0, 0, -scaleZ[i].x);
				break;
			case 1:
				sr= Vec3f(-scaleX[i].x, 0, 0);
				break;
			case 2:
				sr= Vec3f(0, 0, scaleZ[i].y);
				break;
			case 3:
				sr= Vec3f(scaleX[i].y, 0, 0);
				break;
		}
		Matrix matrix = rotationMatrix(ringBuild->orientation.refValue()[i]) * translationMatrix(ringBuild->location.refValue()[i]) * 
		scaleMatrix(ringBuild->scale.refValue()) * rotationMatrix(ringBuild->rotation.refValue()) * translationMatrix(ringBuild->translation.refValue()) * g_scene->getViewMatrix();

		startPos3d = matrix * sr;

		Vec2i mouse2dPos=get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			phase = 1;
			beginRobotTime = g_time;
			Window::activeWindow=g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			g_pointer.translation=startPos3d;
			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance = moveTo3dPoint(to);

		VisualAid* va=dynamic_cast<VisualAid*>(g_selNode.getRaw());
		va->drag(g_pointer.translation);
		if(distance>0){
			UnAcquire();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}

bool MoveToAnimRotPoint::doEvent(){
	startPos3d = g_vcArtAnimBase->rotCtrlPnt;
	Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

	if(localMove(mouse2dPos)>=0){
		tugFlag = true;
		Window::activeWindow = g_3d.p;
		SetAcquire();
		mouse3dMode = true;
		showCursor(FALSE);
		return true;
	}
	return false;
}

bool MoveToMultiSelectRotPoint::doEvent(){
	startPos3d = g_transformCP->rotCtrlPnt[leverIdx];
	Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

	if(localMove(mouse2dPos)>=0){
		g_transformCP->dragPivotPointMode = false;
		g_transformCP->dragRotPointMode = true;
		tugFlag = true;
		Window::activeWindow = g_3d.p;
		SetAcquire();
		mouse3dMode = true;
		showCursor(FALSE);
		return true;
	}
	return false;
}

bool Scale1d::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;

	if(phase == 0){
		startPos3d = g_transformCP->rotCtrlPnt[leverIdx];
		Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			Window::activeWindow = g_3d.p;
			SetAcquire();
			mouse3dMode=true;
			showCursor(FALSE);
			phase++;
			beginRobotTime = g_time;
			g_pointer.translation=startPos3d;
			l_altDown = true;
			outputStatus();
			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance=moveTo3dPoint(to);

		VisualAid* va=dynamic_cast<VisualAid*>(g_selNode.getRaw());
		va->drag(g_pointer.translation);
		if(distance>0){
			UnAcquire();
			l_altDown = false;
			outputStatus();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}

bool Scale2d::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;

	if(phase == 0){
		startPos3d = g_transformCP->rotCtrlPnt[leverIdx];
		Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			Window::activeWindow = g_3d.p;
			SetAcquire();
			mouse3dMode=true;
			showCursor(FALSE);
			phase++;
			beginRobotTime = g_time;
			g_pointer.translation=startPos3d;
			l_ctrlDown = true;
			l_altDown = true;
			outputStatus();
			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance=moveTo3dPoint(to);

		VisualAid* va=dynamic_cast<VisualAid*>(g_selNode.getRaw());
		va->drag(g_pointer.translation);
		if(distance>0){
			UnAcquire();
			l_ctrlDown = false;
			l_altDown = false;
			outputStatus();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}

bool UnRectVert::doEvent(){
	if(phase == 0){
		if(localMove(get2dMousePosFrom2d(from))>=0){
			phase++;
			beginRobotTime = g_time;
			startPos3d=MultiSelect::bottomRight=MultiSelect::topLeft=g_pointer.translation = Vec3f(from.x, from.y, 0);
			l_shiftDown = true;
			l_ctrlDown = true;

			mouse3dMode=true;
			showCursor(FALSE);

			Window::activeWindow=g_3d.p;
			SetAcquire();

			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance=moveTo3dPoint(to);

		Vec3f v = g_pointer.translation;
		MultiSelect::bottomRight = v;
		houseKeepVisualClue();
		
		if(distance>0){
			Vec2i vec=get2dMousePosFrom3d(g_pointer.translation);
			WINDOWINFO wi;
			GetWindowInfo(hWndMain, &wi);
			SetCursorPos(vec.x+wi.rcClient.left, vec.y+wi.rcClient.top); 

			UnAcquire();
			l_shiftDown = false;
			l_ctrlDown = false;
			outputStatus();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}

bool RectVert::doEvent(){
	if(phase == 0){
		if(localMove(get2dMousePosFrom2d(from))>=0){
			phase++;
			beginRobotTime = g_time;
			startPos3d=MultiSelect::bottomRight=MultiSelect::topLeft=g_pointer.translation = Vec3f(from.x, from.y, 0);
			l_shiftDown = true;

			mouse3dMode=true;
			showCursor(FALSE);

			Window::activeWindow=g_3d.p;
			SetAcquire();

			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance=moveTo3dPoint(to);

		Vec3f v = g_pointer.translation;
		MultiSelect::bottomRight = v;
		houseKeepVisualClue();
		
		if(distance>0){
			Vec2i vec=get2dMousePosFrom3d(g_pointer.translation);
			WINDOWINFO wi;
			GetWindowInfo(hWndMain, &wi);
			SetCursorPos(vec.x+wi.rcClient.left, vec.y+wi.rcClient.top); 

			UnAcquire();
			l_shiftDown = false;
			outputStatus();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}
/*
bool RectVert::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;


	Vec2i mouse2dPos=get2dMousePosFrom3d(Vec3f(from.x, from.y, 0));

	if(phase == 0){
		if(localMove(mouse2dPos)>=0){
			MultiSelect::reset();
			phase++;
			beginRobotTime = g_time;


			g_pointer.translation=Vec3f(from.x, from.y, 0);
			l_shiftDown = true;
			outputStatus();
			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance=moveTo3dPoint(to);

		Vec3f v = g_pointer.translation;
		MultiSelect::bottomRight = v;
		houseKeepVisualClue();
		
		if(distance>0){
			UnAcquire();
			l_shiftDown = false;
			outputStatus();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}
*/
bool Neutral::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;

	if(phase == 0){
		startPos3d = g_transformCP->rotCtrlPnt[leverIdx];
		Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			Window::activeWindow = g_3d.p;
			SetAcquire();
			mouse3dMode=true;
			showCursor(FALSE);
			phase++;
			beginRobotTime = g_time;
			g_pointer.translation=startPos3d;
			l_ctrlDown = true;
			outputStatus();
			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance=moveTo3dPoint(to);

		VisualAid* va=dynamic_cast<VisualAid*>(g_selNode.getRaw());
		va->drag(g_pointer.translation);
		if(distance>0){
			//if(dynamic_cast<NeutralSnap*>(this) == 0){
			//	l_ctrlDown = false;
			//}
			UnAcquire();
			l_ctrlDown = false;
			outputStatus();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}


bool Rotate::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter = renderCounter;

	if(phase == 0){
		startPos3d = g_transformCP->rotCtrlPnt[leverIdx];
		Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			Window::activeWindow = g_3d.p;
			SetAcquire();
			mouse3dMode = true;
			showCursor(FALSE);
			phase++;
			beginRobotTime = g_time;
			g_pointer.translation = startPos3d;
			outputStatus();
			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance=moveTo3dPoint(to);

		VisualAid* va = dynamic_cast<VisualAid*>(g_selNode.getRaw());
		if(g_transformCP->selectObject){
			if(Build* build = dynamic_cast<Build*>(g_selNode.getRaw())){
				if(NurbsPatch* patch = build->getPatch()){
					va = patch;
				}
			}
		}

		va->drag(g_pointer.translation);
		if(distance>0){
			UnAcquire();
			outputStatus();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}

bool Stretch::doEvent(){
	if(prevRenderCounter==renderCounter){
		return false;
	}
	prevRenderCounter=renderCounter;

	if(phase == 0){
		startPos3d = g_transformCP->rotCtrlPnt[leverIdx];
		Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

		if(localMove(mouse2dPos)>=0){
			Window::activeWindow = g_3d.p;
			SetAcquire();
			mouse3dMode=true;
			showCursor(FALSE);
			phase++;
			beginRobotTime = g_time;
			g_pointer.translation=startPos3d;
			l_ctrlDown = true;
			l_shiftDown = true;
			outputStatus();
			SenseTugGlobals::initTug();
		}
	}
	else{
		float distance=moveTo3dPoint(to);

		VisualAid* va=dynamic_cast<VisualAid*>(g_selNode.getRaw());
		va->drag(g_pointer.translation);
		if(distance>0){
			UnAcquire();
			l_ctrlDown = false;
			l_shiftDown = false;
			outputStatus();
			resetDragActuallyHappned();
			return true;
		}
	}
	return false;
}

bool MoveToMultiSelectPivotPoint::doEvent(){
	startPos3d = g_scene->getViewMatrix() * (g_transformCP->pivotPoint + g_transformCP->translation);
	Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

	if(localMove(mouse2dPos) >= 0){
		SenseTugGlobals::initTug();
		//g_transformCP->dragPivotPointMode = true;
		//g_transformCP->dragRotPointMode = false;
		//tugFlag = true;
		//Window::activeWindow = g_3d.p;
		//SetAcquire();
		mouse3dMode = true;
		//showCursor(FALSE);
		return true;
	}
	return false;
}

bool MoveToPivotPoint::doEvent(){
	Art* part = dynamic_cast<Art*>(g_sceneTreeList[nodeIdx].node.p);
	if(part->ik){
		startPos3d  = g_scene->getViewMatrix() * part->ikPos;
	}
	else{
		startPos3d  = part->pivotPointMatrix * Vec3f(0,0,0);
	}
	Vec2i mouse2dPos = get2dMousePosFrom3d(startPos3d);

	if(localMove(mouse2dPos)>=0){
		tugFlag = true;
		SetAcquire();
		mouse3dMode = true;
		showCursor(FALSE);

		if(g_scene->mode & XRAY_EDIT_MODE){
			SetAcquire();
			g_pointer.translation = startPos3d;
			SenseTugGlobals::initTug();
		}
		else if(Window::animMode()){
			SetAcquire();
			g_pointer.translation = startPos3d;
			SenseTugGlobals::initTug();
			g_vcArtAnimBase->dragTranPointMode = true;
		}
		else{
			g_vcArtAnimBase->dragTranPointMode = true;
			g_vcArtAnimBase->remTranValue = part->translation;
		}

		return true;
	}
	return false;
}


bool Move3d::doEvent(){
	float distance = moveTo3dPoint(to);

	VisualAid* va=dynamic_cast<VisualAid*>(g_selNode.getRaw());

	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		patch->drag(g_pointer.translation);
		//nurbsPatch->dpoint.refValue()[nurbsPatch->index[0]].x=g_pointer.translation.z;
		//nurbsPatch->dpoint.refValue()[nurbsPatch->index[0]].z=g_pointer.translation.x;
		//nurbsPatch->dpoint.refValue()[nurbsPatch->index[0]].y=g_pointer.translation.y;
		//nurbsPatch->dpointsToPoints();
		if(distance > 0 || prevRenderCounter != renderCounter){
			prevRenderCounter = renderCounter;
		}
	}
	else if(NurbsLathe* nurbsLathe=dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
		nurbsLathe->drag(nurbsLathe->matrix * g_scene->getViewMatrix() * g_pointer.translation);

//		nurbsLathe->dpoint.refValue()[nurbsLathe->index[0]]=Vec2f(g_pointer.translation.x, g_pointer.translation.y);
//		nurbsLathe->dpointsToPoints();
		if(distance > 0 || prevRenderCounter != renderCounter){
			g_selNode->processFields();
			va->houseKeep();
			updateBuild3dWindow();
			prevRenderCounter=renderCounter;
		}
	}
	else if(Bezier* bezier=dynamic_cast<Bezier*>(g_selNode.getRaw())){
		bezier->matchBorderTangents(Vec2f(g_pointer.translation.x, g_pointer.translation.y));
		if(distance>0 || prevRenderCounter!=renderCounter){
			g_selNode->processFields();
			va->houseKeep();
			updateBuild3dWindow();
			prevRenderCounter=renderCounter;
		}
	}
	else{
		va->drag(g_pointer.translation);
	}
	if(distance>0){
		if(AnimClue* animClue = dynamic_cast<AnimClue*>(g_visualClue)){
			Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
			if(g_dragActuallyHappned){
				if(animClue->dragTranPointMode){
					if(Anim::autoCapture){
						if(part->ik){
							animClue->anim->captureSinglePartPose(part, &part->ikPos);
						}
						else{
							animClue->anim->captureSinglePartPose(part, &part->translation);
						}
					}
					g_vcArtAnimBase->dragTranPointMode = false;
					part->houseKeep();
				}
				else{
					if(Anim::autoCapture){
						if(part->ik){
							animClue->anim->captureSinglePartPose(part, &part->ikRot);
						}
						else{
							animClue->anim->captureSinglePartPose(part, &part->rotation);
						}
					}
				}
				Window::paintAllControlPanels();
			}
		}

		UnAcquire();
		resetDragActuallyHappned();
		//dLog(String("UnAcquire")+g_pointer.translation.toString());
		return true;
	}
	return false;
};

Button* getButtonForRobot(int id, Window*& win);

bool MoveToButton::doEvent(){
	Window* win  = 0;
	if(Button* button = getButtonForRobot(id, win)){
		POINT screenPoint={win->pos.x, win->pos.y};
		ClientToScreen(hWndMain, &screenPoint);
		Vec2i clickPos=button->getSimClickPos(subButtonIdx);
				
		Vec2i endPos = Vec2i(screenPoint.x + button->dstPos.x + clickPos.x, + screenPoint.y + button->dstPos.y + clickPos.y);

		if(move(endPos)>=0){
			return true;
		}
		return false;
	}
	_killTimer();
	return false;
}

bool ClickButton::doClickButton(int& phase, const int& numOfActions){
	static int phase2 = 0;
	Window* win  = 0;
	if(phase == 0){
		if(Button* button = getButtonForRobot(id, win)){
			POINT screenPoint = {win->pos.x, win->pos.y};
			ClientToScreen(hWndMain, &screenPoint);
			Vec2i clickPos=button->getSimClickPos(subButtonIdx);
					
			Vec2i endPos = Vec2i(screenPoint.x + button->dstPos.x + clickPos.x, + screenPoint.y + button->dstPos.y + clickPos.y);

			if(move(endPos) >= 0){
				Button::repeatCounter=0;//seems like we need this for creating a movie
				//dLog("Button::repeatCounter=0;");
				phase=1;
				phase2=0;
				beginRobotTime = g_time;
			}
		}
	}
	else if(phase==1){
		return doClickAndAutoRepeat(phase2, numOfActions);
	}
	return false;
}
bool ClickButton::doEvent(){
	return doClickButton(phase, numOfActions);
};
bool ButtonCtrl::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(doVirtKeyDown(VK_LCONTROL)){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else if(phase==1){
		if(doClickButton(phase2, numOfActions)){
			phase++;
			beginRobotTime = g_time;
		}
	}
	else{
		return doVirtKeyUp(VK_LCONTROL);
	}
	return false;
}
bool InputEvent::doLeftButtonDown(){
	if(fraction_changed<period){
		return false;
	}
	Vec2i sysPos=getSysMouseCoord();
	mouse_event(
		MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTDOWN,
		sysPos.x,
		sysPos.y,
		0,
		0
	);
	leftDrag=true;
	outputStatus();
	return true;
};
bool LeftButtonDown::doEvent(){
	return doLeftButtonDown();
};
bool InputEvent::doLeftButtonUp(bool& released){
	if(fraction_changed<prePeriod){
		return false;
	}
	if(released==false){
		Vec2i sysPos=getSysMouseCoord();
		mouse_event(
			MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTUP,
			sysPos.x,
			sysPos.y,
			0,
			0
		);
		released=true;
		leftDrag=false;
		outputStatus();
	}
	if(fraction_changed<prePeriod+postPeriod){
		return false;
	}
	released=false;
	return true;
};
bool LeftButtonUp::doEvent(){
	return doLeftButtonUp(released);
}
bool InputEvent::doMoveToNewNode(const int id){
	M_<M_<int> >& newNodeBarList=g_stw->newNodeBar.newNodeBarList;
	for(int bar=newNodeBarList.length; --bar>=0;){
		for(int i=newNodeBarList[bar].length; --i>=0;){
			if(newNodeBarList[bar][i]==id){
				POINT point={
					g_stw->pos.x+g_stw->newNodeBar.dstPos.x + bar * pixSymbolCX + pixSymbolCX/2,
					g_stw->pos.y+g_stw->newNodeBar.dstPos.y + i * pixSymbolCY + pixSymbolCY/2 + DragNode::yOffset
				};
				if(localMove(Vec2i(point.x,point.y))==-1){
					return false;
				}
				else{
					return true;
				}
			}
		}
	}
	_killTimer();
	return false;
}
bool MoveToNewNode::doEvent(){
	return doMoveToNewNode(id);
}
bool DragNewNodeBar::doEvent(){
	static bool released=false;

	if(phase==0){
		if(localMove(Vec2i(g_stw->pos.x+g_stw->newNodeBar.dstPos.x+10, g_stw->pos.y+g_stw->newNodeBar.dstPos.y+6))>=0){
			beginRobotTime = g_time;
			startPos=pos;
			released=false;
			phase++;
		}
	}
	else if(phase==1){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(localMove(Vec2i(g_stw->pos.x+position.x, g_stw->pos.y+position.y))>=0){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==3){
		if(doLeftButtonUp(released)){
			return true;
		}
	}
	return false;
}
bool NewNodeAsSiblingAbove::doEvent(){
	static int phase2=0;
	static bool released=false;
	if(phase==0){
		if(doMoveToNewNode(id)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(doMoveToNodeAsSiblingAbove(nodeLine)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==3){
		if(doLeftButtonUp(released)){
			return true;
		}
	}
	return false;
}
bool NewNodeAsSiblingBelow::doEvent(){
	static int phase2=0;
	static bool released=false;
	if(phase==0){
		if(doMoveToNewNode(id)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(doMoveToNodeAsSiblingBelow(nodeLine)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==3){
		if(doLeftButtonUp(released)){
			return true;
		}
	}
	return false;
}
bool NewNodeAsChild::doEvent(){
	static int phase2=0;
	static bool released=false;
	if(phase==0){
		if(doMoveToNewNode(id)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(doMoveToNodeAsChild(phase2, nodeLine)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==3){
		if(doLeftButtonUp(released)){
			return true;
		}
	}
	return false;
}
bool MoveNodeAsChild::doEvent(){
	static int phase2=0;
	static bool released=false;
	if(phase==0){
		if(doMoveToNode(fromLine)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(doMoveToNodeAsChild(phase2, toLine)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==3){
		if(doLeftButtonUp(released)){
			return true;
		}
	}
	return false;
}
bool CopyNodeAsChild::doEvent(){
	static bool released=false;
	static int phase2=0;
	if(phase==0){
		if(doMoveToNode(nodeLineFrom)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doVirtKeyDown(VK_LCONTROL)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
			phase2=0;
		}
	}
	else if(phase==3){
		if(doMoveToNodeAsChild(phase2, nodeLineTo)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==4){
		if(doLeftButtonUp(released)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==5){
		return doVirtKeyUp(VK_LCONTROL);
	}
	return false;
}
bool CopyNodeAsSiblingAbove::doEvent(){
	static bool released=false;
	static int phase2=0;
	if(phase==0){
		if(doMoveToNode(nodeLineFrom)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doVirtKeyDown(VK_LCONTROL)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
			phase2=0;
		}
	}
	else if(phase==3){
		if(doMoveToNodeAsSiblingAbove(nodeLineTo)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==4){
		if(doLeftButtonUp(released)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==5){
		return doVirtKeyUp(VK_LCONTROL);
	}
	return false;
}
bool CopyNodeAsSiblingBelow::doEvent(){
	static bool released=false;
	static int phase2=0;
	if(phase==0){
		if(doMoveToNode(nodeLineFrom)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doVirtKeyDown(VK_LCONTROL)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
			phase2=0;
		}
	}
	else if(phase==3){
		if(doMoveToNodeAsSiblingBelow(nodeLineTo)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==4){
		if(doLeftButtonUp(released)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==5){
		return doVirtKeyUp(VK_LCONTROL);
	}
	return false;
}
bool UseNodeAsSource::doEvent(){
	static bool released=false;
	static int phase2=0;
	if(phase==0){
		if(doMoveToNode(nodeLineFrom)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doVirtKeyDown(VK_LMENU)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
			phase2=0;
		}
	}
	else if(phase==3){
		if(doMoveToNodeAsSource(phase2, nodeLineTo)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==4){
		if(doLeftButtonUp(released)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==5){
		return doVirtKeyUp(VK_LMENU);
	}
	return false;
}
bool UseNodeAsDest::doEvent(){
	static bool released=false;
	static int phase2=0;
	if(phase==0){
		if(doMoveToNode(nodeLineFrom)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doVirtKeyDown(VK_LMENU)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
			phase2=0;
		}
	}
	else if(phase==3){
		if(doMoveToNodeAsDest(phase2, nodeLineTo)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==4){
		if(doLeftButtonUp(released)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==5){
		return doVirtKeyUp(VK_LMENU);
	}
	return false;
}
bool MoveNodeAsDest::doEvent(){
	static bool released=false;
	static int phase2=0;
	if(phase==0){
		if(doMoveToNode(nodeLineFrom)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
			phase2=0;
		}
	}
	else if(phase==2){
		if(doMoveToNodeAsDest(phase2, nodeLineTo)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==3){
		return doLeftButtonUp(released);
	}
	return false;
}
bool UseNodeAsReplacement::doEvent(){
	static bool released=false;
	static int phase2=0;
	if(phase==0){
		if(doMoveToNode(nodeLineFrom)){
			phase2=0;
			released=false;
			beginRobotTime = g_time;
			startPos=pos;
			phase++;
		}
	}
	else if(phase==1){
		if(doVirtKeyDown(VK_LMENU)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==2){
		if(doLeftButtonDown()){
			beginRobotTime = g_time;
			phase++;
			phase2=0;
		}
	}
	else if(phase==3){
		if(doMoveToNodeAsReplacement(phase2, nodeLineTo)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==4){
		if(doLeftButtonUp(released)){
			beginRobotTime = g_time;
			phase++;
		}
	}
	else if(phase==5){
		return doVirtKeyUp(VK_LMENU);
	}
	return false;
}
bool MouseWheel::doEvent(){
	if(fraction_changed<(click+1)*clickPeriod){
		return false;
	}
	if(click>=clickSum){
		return true;
	}
	/*
	mouse_event(
		MOUSEEVENTF_WHEEL,
		0,
		0,
		120,
		0
	);
*/
	INPUT input={
		INPUT_MOUSE,
		{
			0,
			0,
			direction,
			MOUSEEVENTF_WHEEL | MOUSEEVENTF_MOVE,
			0,
			0
		}
	};
	SendInput(1, &input, sizeof(INPUT));

	click++;
	return false;
}
bool Move_MenuItem::doEvent(){
	MenuItem* menuItem = Window::getMenuItem(str);
	float movePeriod = localMove(menuItem->win.pos+menuItem->dstPos + (menuItem->bSize >> 1));
	if(movePeriod >= 0){
		if(fraction_changed >= movePeriod + pausePeriod){
			return true;
		}
	}
	return false;
}
/*
const int shiftedCharLen=13;
const char shiftedChar[shiftedCharLen][2]={
	{'~', '`'},
	{'!', '1'},
	{'@', '2'},
	{'#', '3'},
	{'$', '4'},
	{'%', '5'},
	{'^', '6'},
	{'&', '7'},
	{'*', '8'},
	{'(', '9'},
	{')', '0'},
	{'_', '_'},
	{'+', '='}
};
inline bool checkShift(unsigned char& ch){
	for(int i=shiftedCharLen; --i>=0;){
		if(shiftedChar[i][0]==ch){
			ch = shiftedChar[i][1];
			return true;
		}
	}
	return false;
}
*/
bool InputEvent::doTypeVirtKey(char vk){
	if(fraction_changed<period){
		return false;
	}
	keybd_event(vk, 0x45, KEYEVENTF_EXTENDEDKEY, 0);
	keybd_event(vk, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
	return true;
}
bool InputEvent::doTypeString(const String str, int& charIdx){
	//static const float charPeriod=.125;
	static const float charPeriod=.25;
#ifndef DEMO_TO_MOVIE
	if(fraction_changed<(charIdx+1)*charPeriod){
		return false;
	}
#endif
	if(charIdx>=str.length){
		return true;
	}
	unsigned char ch = str[charIdx++];
	short keyScan=VkKeyScan(ch);

	bool shiftState=false;
	if(keyScan&0x100){
		shiftState=true;
		keybd_event(VK_LSHIFT, 0, 0, 0);
	}
	ch = keyScan&0xff;

	keybd_event(ch, 0, 0, 0);
	keybd_event(ch, 0, KEYEVENTF_KEYUP, 0);
	if(shiftState){
		keybd_event(VK_LSHIFT, 0, KEYEVENTF_KEYUP, 0);
	}
	return false;
}
bool SetFieldValue::doEvent(){
	static int phase2=0;
	if(phase==0){
		phase2=0;
		phase++;
	}
	if(phase==1){
		if(doMoveToFieldValue(field, phase2, layer)){
			phase++;
			beginRobotTime = g_time;
			phase2=0;
		}
	}
	else if(phase==2){
		if(doSingleClick(phase2)){
			phase++;
			beginRobotTime = g_time;
		}
	}
	else if(phase==3){
		if(doVirtKeyDown(VK_LSHIFT)){
			phase++;
			beginRobotTime = g_time;
		}
	}
	else if(phase==4){
		if(doTypeVirtKey(VK_END)){
			phase++;
			beginRobotTime = g_time;
		}
	}
	else if(phase==5){
		if(doVirtKeyUp(VK_LSHIFT)){
			phase++;
			beginRobotTime = g_time;
		}
	}
	else if(phase==6){
		return doTypeString(value, charIdx);
	}
	return false;
}

void restoreRobotFrom3d(){
	POINT point;
	GetCursorPos(&point);
	InputEvent::pos=Vec2i(point.x, point.y);
}

void playDemo(){
	g_dontBackup = true;
	InputEvent::emptyList();
	g_robotDemo->function();
	for(int i = 0, k = -1; i<InputEvent::list.length-1; i++){
		if(Section* section=dynamic_cast<Section*>(InputEvent::list[i])){
			k++;
			if(k == g_robotDemo->operation){
				InputEvent::loadPrevFromCache(k);
				g_robotDemo->firstOperation = true;
				InputEvent::sequencerIdx= i;
				setupSequence2();
				break;
			}
		}
	}
}

void Demo::Play::operator=(bool const &){
//	removeTypeFromList<Button>(g_callBackList);
	removeTypeFromList<DownButton>(g_callBackList);

	g_forceSaveAs = true;
	robotDemoStart(playDemo);
}

void Demo::Operation::operator=(int const& par){
	resetTemporaryModes();

	int saveFlag = smlsIsSaved();
	if(IDCANCEL==saveFlag){
		return;
	}
	else if(IDYES==saveFlag){
		saveSmlsAs();
	}
	g_dontBackup=true;
	setSmlsSaved();
	if(InputEvent::robotActive == false){
		for(int i = 0, k = -1; i<InputEvent::list.length; i++){
			if(Section* section=dynamic_cast<Section*>(InputEvent::list[i])){
				k++;
				if(k==par){
					if(InputEvent::loadPrevFromCache(par)){
						raw=par;
						InputEvent::sequencerIdx= i;
						g_robotDemo->name=section->title;
						g_robotDemo->status="";
						g_rcp->paint();
						return;
					}
					else{
						break;
					}
				}
			}
		}
	}
	InputEvent::loadPrevFromCache(raw);//undo damge
	
}
void Demo::Stop::operator=(bool const&){
	if(InputEvent::robotActive){
		InputEvent::robotActive = false;
		resetTemporaryModes();
		//KillTimer(hWndMain, InputEvent::TIMEID);
		InputEvent::_killTimer();
		g_robotDemo->status=str_interupted;
		g_rcp->paint();
	}
}
bool Section::doEvent(){
	static int waiting=0;
	if(phase==0){
		waiting=3;
		if(g_robotDemo->firstOperation==false && g_robotDemo->express == false){
			g_robotDemo->backupToCache();
			_killTimer();
			g_robotDemo->status=str_done;
			g_rcp->paint();
			return true;
		}
		g_robotDemo->name=title;
		if(g_robotDemo->name==str_complete){
			period=0;
		}
		g_rcp->paint();
//		g_rcp->setupNodeWindow(g_robotDemo);
//		Window::activeWindow=g_rcp;
	}
	if(g_robotDemo->firstOperation==false){
		g_robotDemo->backupToCache();
		g_robotDemo->operation.refValue()++;//dont wait move onto next operation
//		dLog(g_robotDemo->operation.refValue());
		return true;
	}

	phase++;

	int w=2-fraction_changed+1;
	g_robotDemo->status.refValue()="waiting: ";
	g_robotDemo->status.refValue()+=w;
	if(w != waiting){
		waiting--;
		g_rcp->paint();
	}

	if(fraction_changed>=period){
		g_robotDemo->firstOperation=false;
		g_robotDemo->status="";
		g_rcp->paint();
		return true;
	}
	return false;
}

void Demo::backupToCache(){
	Viewpoint* viewpoint=new Viewpoint;
	g_scene->setViewpointFromScene(*viewpoint);
	g_scene->children.refValue()+=viewpoint;
	g_scene->children.refValue()+=g_pasteOptions.p;
	g_scene->children.refValue()+=g_deleteOptions.p;
	g_scene->children.refValue()+=g_timelineFilter.p;

	String str=g_demoCacheDir;
	str+=fileName;
	str+=operation;
	str+=langExtension;
	str+=".smls";
	writeSmlsFile(str);
	g_scene->children.refValue().length-=4;
}
bool InputEvent::loadPrevFromCache(int operation){
	g_transformCP->sameScreenPosition.refValue() = false;
	g_transformCP->selectRing.refValue() = false;
	g_transformCP->selectRings.refValue() = false;
	g_transformCP->selectObject.refValue() = false;
	MultiSelect::reset();
	if(operation==0){
		g_scene->children.refValue().length = 0;
		addNodesForNewFile();
		g_stw->genNodeList();

		g_tb->pos = Vec2i(0, 0);
		g_tb2->pos = Vec2i(0, g_3d->size.y - g_tb2->size.y);
		g_settings->incDecRes=.01;
		g_settings->incDecResCtrl=.001;

		g_stw->size=Vec2i(330,660); 
		g_stw->pos=Vec2i(g_3d->size.x-330,0);
		g_stw->newNodeBar.dstPos=Vec2i(g_stw->size.x-80, 30);
		g_stw->cp->size=Vec2i(350, g_stw->cp->size.y);
		g_stw->cp->pos=Vec2i(0,180);
		resetLooseEnds();
		return true;
	}
	String str=g_demoCacheDir;
	str+=g_robotDemo->fileName;
	str+=operation-1;
	str+=langExtension;
	str+=".smls";
	initCompile();
	if(Browser::createSceneFromFile(str) && g_scene->version > (2.172-.0001)){
		M_Node& children = g_scene->children.refValue();
		Viewpoint* viewpoint = dynamic_cast<Viewpoint*>(children[g_scene->children.refValue().length-4].p);
		g_scene->setSceneFromViewpoint(*viewpoint);
		g_pasteOptions = dynamic_cast<PasteOptions*>(children[g_scene->children.refValue().length-3].p);
		g_deleteOptions = dynamic_cast<DeleteOptions*>(children[g_scene->children.refValue().length-2].p);
		g_timelineFilter = dynamic_cast<TimelineFilter*>(children[g_scene->children.refValue().length-1].p);

		g_scene->children.refValue().length-=4;
		BackUp::commonUndoRedo();
		if(VisualAid* va=dynamic_cast<VisualAid*>(g_selNode.getRaw())){
			va->houseKeep();
		}
		g_wireFrameMode = g_scene->mode & WIREFRAME_MODE;
		Color3f backgroundColor = g_scene->backgroundColor;
		g_stw->paint();

		loadGridBrushEtcNodes();

		return true;
	}
	else{
		g_scene->children.refValue().length = 0;
		addNodesForNewFile();
		g_stw->genNodeList();
		resetLooseEnds();
	}
	return false;
}
void get2dMousePosFrom3d(){
	Vec3f mouse3d;
	if(InputEvent::robotActive){
		if(NurbsLathe* nurbsLathe=dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
//			mouse3d = nurbsLathe->matrix * g_scene->getViewMatrix() * g_pointer.translation;
			Vec2f point2f = nurbsLathe->point[nurbsLathe->index[0]];
			mouse3d = nurbsLathe->matrix * g_scene->getViewMatrix() * Vec3f(point2f.x, point2f.y, 0);
		}
		else if(Bezier* bezier=dynamic_cast<Bezier*>(g_selNode.getRaw())){
			mouse3d = g_scene->getViewMatrix() * bezier->matrix * g_pointer.translation;
		}
		else{
			mouse3d = g_pointer.translation;
		}
	}
	else{
		mouse3d = g_pointer.translation;
	}

	Vec2i vec=get2dMousePosFrom3d(mouse3d);

	WINDOWINFO wi;
	GetWindowInfo(hWndMain, &wi);
	SetCursorPos(vec.x+wi.rcClient.left, vec.y+wi.rcClient.top); 
}
void InputEvent::emptyList(){
	for(int i=list.length; --i>=0;){
		delete list[i];
	}
	list.length=0;
}
struct IntLookUp{
	String label;
	int v;
	IntLookUp():
		label(0)
	{}
	IntLookUp(const String label, const int v):
		label(label),
		v(v)
	{}
};
M_<IntLookUp> g_intLookUp;
int convertLabelToInt(const String label){
	for(int i=g_intLookUp.length; --i>=0;){
		if(g_intLookUp[i].label==label){
			return g_intLookUp[i].v;
		}
	}
	throw;
}

void InputEvent::getInstruction(const String& str, int& i){
	str.stb(i);
	switch(str[i++]){
		case 'i':{
			if(str.cmpCppWord("nt",i)){
				str.stb(i);
				String label=str.extractCppLabel(i);
				str.stb(i);
				if(str[i++]!='='){
					throw;
				}
				int j=getInt(str, i);
				g_intLookUp+=IntLookUp(label, j);

				break;
			}
		}
		case 'l':{
			if(str.cmpCppWord("ist",i)){
				if(str[i++]!='+'){
					throw;
				}
				if(str[i++]!='='){
					throw;
				}
				str.stb(i);
				if(str[i++]!='{'){
					throw;
				}
				fillList(str, i);
				break;
			}
		}
		default: throw;
	}
}
void InputEvent::decodeDemo(const String& str, int& i){
	g_intLookUp.length=0;
	str.stb(i);
	str.stw(i);
	str.stb(i);
	str.stw(i);
	str.stb(i);
	
	while(str[i]!='}'){
		getInstruction(str, i);
		str.stb(i);
		if(str[i++]!=';'){
			throw;
		}
		str.stb(i);
	}
}
void nameOrLineNumber(const int nodeLine, String& str){
	Node* node=g_sceneTreeList[nodeLine].node;
	if(node->name!=""){
		str+=node->name;
	}
	else{
		str+=nodeLine;
	}
}
extern int g_frame;
extern int g_clipCounter;
bool DemoSound::doEvent(){
	int a = PlaySound(
		filename,
		0,
		SND_ASYNC
	);
#ifdef DEMO_TO_MOVIE
	g_frame=0;
	g_clipCounter++;
	g_rcp->pos=Vec2i(0,800);

	String soundName = g_screenshotDir;
	soundName+= "\\clip";

	String cntrStr = g_clipCounter;
	while(cntrStr.length < 4){
		cntrStr = '0' + cntrStr;
	}
	soundName += cntrStr;
	soundName+= "\"";

	String command = "copy \"" + String(filename) + "\" \"" + soundName;
	system(command.c_str());
#endif
	return true;
}
bool RelMove::doEvent(){
	static Vec2i remPos;
	if(phase==0){
		remPos = pos;
		phase++;
	}
	if(move(remPos+to)>=0){
		return true;
	}
	return false;
};
bool Move_CreateNewFile_Save::doEvent(){
	HWND hWnd=GetActiveWindow();
	RECT rect;
	GetWindowRect(GetDlgItem(hWnd, 1), &rect);
	if(phase==0){
		phase++;
		beginRobotTime = timeGetTime() * MILISECONDS_TO_SECONDS;
		fraction_changed = 0;
	}
	else if(
		move(Vec2i(rect.left+(rect.right-rect.left)/2, rect.top+(rect.bottom-rect.top)/2))>=0){
		return true;
	}
	return false;
}

void Scene::extraToString(String& str){
	if(InputEvent::robotActive && g_transformCP->sameScreenPosition){
		str+="\nsameScreenPosition TRUE";
	}
}
