/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "seamless3d.h"
#include "gui.h"
#include "scene_tree.h"

Seamless* getSeamless(int& selLine){
	selLine = g_stw->selLine;
	while(selLine != 0){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[selLine].node.p)){
			return smls;
		}
		selLine=getParent(selLine);
	}
	return 0;
}

Seamless* getSeamless2(){
	if(Build* selBuild = dynamic_cast<Build*>(g_selNode.getRaw())){
		if(NurbsPatch* selPatch = selBuild->getPatch()){
			if(selPatch->smls){
				return selPatch->smls;
			}
		}
	}


	int selLine = g_stw->selLine;
	while(selLine != 0){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[selLine].node.p)){
			return smls;
		}
		selLine = getParent(selLine);
	}
	return 0;
}

Seamless* getSeamless3(int& selLine){
	while(selLine > 0 && selLine < g_sceneTreeList.length){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[selLine].node.p)){
			return smls;
		}
		selLine = getParent(selLine);
	}
	return 0;
}

GroupBuild* getGroup(NurbsPatch* patch){
	int line = getNodeLine(patch);
	if(line > 0){
		if(GroupBuild* group = dynamic_cast<GroupBuild*>(g_sceneTreeList[g_sceneTreeList[line].parent].node.p)){
			return group;
		}
	}
	return 0;
}

Art* getGroupPart(NurbsPatch* patch){
	if(GroupBuild* group = getGroup(patch)){
		return group->part.refValue();
	}
	return 0;
}

Hair* getHair(int selLine){
	while(selLine != 0){
		if(Hair* hair = dynamic_cast<Hair*>(g_sceneTreeList[selLine].node.p)){
			return hair;
		}
		selLine = getParent(selLine);
	}
	return 0;
}

Hair* getHair(){
	return getHair(g_stw->selLine);
}

int getSeamless2Line(){
	int selLine = g_stw->selLine;
	while(selLine != 0){
		if(dynamic_cast<Seamless*>(g_sceneTreeList[selLine].node.p)){
			return selLine;
		}
		selLine = getParent(selLine);
	}
	return -1;
}

Node* SelNode::operator ->()const{
	return g_sceneTreeList[g_stw->selLine].node;
}
Node* SelNode::getRaw()const{
	return g_sceneTreeList[g_stw->selLine].node;
}

Node* CopyNode::operator ->()const{
	return g_sceneTreeList[g_stw->copyLine].node;
}
Node* CopyNode::getRaw()const{
	return g_sceneTreeList[g_stw->copyLine].node;
}
void SceneTree::escape(){
	if(g_drag){
		g_drag->endDrag(Vec2i(-1000,-1000));
	}
}
void SceneTree::STTextCell::updateText(){
	cellStr=g_selNode->name;
}
