/* Seamless3d

Copyright (C) 2001-2017 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include <windows.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "resource.h"
#include "input.h"
#include "robot_aid.h"
#include "seamless3d_gui_render.h"

VerticalScrollBarSlider::VerticalScrollBarSlider(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, SceneTree& i_win, int refCountValue):
	Button(i_size, i_srcPos, i_tip, i_command, i_win, refCountValue),
	win(i_win)
{}
VerticalScrollBarTopButton::VerticalScrollBarTopButton(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, SceneTree& i_win, int refCountValue):
	Button(i_size, i_srcPos, i_tip, i_command, i_win, refCountValue),
	win(i_win)
{}
VerticalScrollBarBottomButton::VerticalScrollBarBottomButton(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, SceneTree& i_win, int refCountValue):
	Button(i_size, i_srcPos, i_tip, i_command, i_win, refCountValue),
	win(i_win)
{}
PageUp::PageUp(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, SceneTree& i_win, int refCountValue):
	Button(i_size, i_srcPos, i_tip, i_command, i_win, refCountValue),
	win(i_win)
{}
PageDown::PageDown(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, SceneTree& i_win, int refCountValue):
	Button(i_size, i_srcPos, i_tip, i_command, i_win, refCountValue),
	win(i_win)
{}



void VerticalScrollBarSlider::paint(){
	pasteSprite2( dstPos.x, dstPos.y, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray());
}
void VerticalScrollBarTopButton::paint(){
	pasteSprite2( dstPos.x, dstPos.y, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray());
}
void VerticalScrollBarBottomButton::paint(){
	pasteSprite2( dstPos.x, dstPos.y, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray());
}
void PageUp::paint(){
	//win.blitBlockFromHorizontalLine(dstPos, srcPos, Vec2i(9, win.size.y-dstPos.y-22));
	win.blitVerticalColorSweep(dstPos, srcPos-Vec2i(0,1), srcPos, Vec2i(6, win.size.y-dstPos.y-19));
}
void VerticalScrollBarSlider::action(){
	removeTypeFromList < Button>(g_callBackList);
	removeTypeFromList < DownButton>(g_callBackList);
	setDrag(this);
}
void VerticalScrollBarTopButton::action(){
	win.setScrolPos(-1);
}
void VerticalScrollBarBottomButton::action(){
	win.setScrolPos(1);
}
void PageUp::action(){
	win.setScrolPos(-win.numOfVisLines);
}
void PageDown::action(){
	win.setScrolPos(win.numOfVisLines);
}
void VerticalScrollBarSlider::setPixelPos(){
	const int top = win.topScroll.dstPos.y + win.topScroll.bSize.y;
	const int bottom = win.bottomScroll.dstPos.y - win.slider.bSize.y;
	const int relBottom = bottom - top;

	int relY = static_cast<float>(relBottom) / g_sceneTreeCollapsed.length * g_scene->scroll;
	dstPos = Vec2i(win.size.x - 20, top + relY);

}
void VerticalScrollBarSlider::doDrag(const Vec2i ptCursor){
	int diff=ptCursor.y-dragPos.y;

	const int top = win.topScroll.dstPos.y + win.topScroll.bSize.y;
	const int bottom = win.bottomScroll.dstPos.y - win.slider.bSize.y;
	const int relBottom = bottom - top;
	const int relDstPos = dstPos.y - top;


	if( diff < 0 && relDstPos + diff < 0){
		diff = -relDstPos;
	}
	if( diff > 0 && relDstPos + diff >= relBottom){
		diff = relBottom - relDstPos;
	}

	dstPos.y+=diff;
	dragPos.y+=diff;

	g_scene->scroll = static_cast<float>(relDstPos + diff) / relBottom * g_sceneTreeCollapsed.length;
	win.resizePageUpPageDown();
	win.keepLeagal();//why is this needed?
	win.paint();
#ifdef PROG_ROB
		g_upStr="new MouseWheel(";
		g_upStr+=InputEvent::sceneTreeScrolPos-g_scene->scroll;
		g_upStr+="),";
#endif
}
void VerticalScrollBarSlider::endDrag(const Vec2i ptCursor){
	g_drag = 0;
	ReleaseCapture();
	const Vec2i localCursor=ptCursor-win.pos;
	if( localCursor >= dstPos && localCursor < dstPos + bSize){
		open=true;//prevent tool tip opening
		g_callBackList+=this;
	}
}
